// Branchelementals.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2016 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the Branchelemental class 
  
#ifndef VINCIABRANCHELEMENTALS_H
#define VINCIABRANCHELEMENTALS_H

// PYTHIA 8 headers
#include "Pythia8/Event.h"
// Global VINCIA functionality
#include "Vincia/VinciaCommon.h"
// Specific VINCIA headers
#include "Vincia/TrialGeneratorsFSR.h"
#include "Vincia/TrialGeneratorsISR.h"

namespace Vincia {

//**************************************************************************
// Contains:

  class BranchelementalFSR;
  class BranchelementalISR;

//**************************************************************************

// The BranchelementalFSR class: contains a generic 2->3 branching, with 
// pointers to the relevant places in the event record and a few useful 
// functions. 
// The member function list() prints a brief record of the contents.
// The member function setTrial(bool) is used to (re)set whether this
// elemental already contains a saved trial branching, or whether a new 
// one should be generated.

  class BranchelementalFSR {
    
  public:
    
    // Constructors
    // Create empty branchelemental
    BranchelementalFSR() : system(0) {}
    // Create branchelemental with iOld1In,iOld2In as parents
    // and initialize the PDT pointers for the new ones
    BranchelementalFSR(Event& event, int iOld1In, int iOld2In)
    : system(0) {reset(event, iOld1In, iOld2In);};
    
    // Antenna Mass (negative if spacelike virtual)
    double mCalc() {
      sSav = m2(old1,old2);
      mSav = (sSav >= 0) ? sqrt(sSav) : -sqrt(-sSav);
      return mSav;
    }
    double m() const { return mSav; }
    double s() const { return sSav; }
    // CM energy squared of daughters (for cross checks)
    double sNew() const {return m2(new1,new2,new3);}
    // Dot products of daughters
    double s12() const {return m2(new1,new2) - new1.m2() - new2.m2();} 
    double s23() const {return m2(new2,new3) - new2.m2() - new3.m2();} 
    double s13() const {return m2(new1,new3) - new1.m2() - new3.m2();} 
    double sNew1() const {return new1.m2();} 
    double sNew2() const {return new2.m2();}
    double sNew3() const {return new3.m2();}

    // Reset and set trial true/false
    void reset(Event& event, int iOld1In, int iOld2In) {
      // Make local copy of input particles
      old1        = event[iOld1In];
      old2        = event[iOld2In];
      // Compute and store antenna invariant mass
      mCalc();      
      // Indices of parents and neighbors
      iOld1       = iOld1In;
      iOld2       = iOld2In;
      iL          = 0; 
      iR          = 0;
      // Trial Generators
      clearTrialGenerators();
      nVeto       = 0;
      nHull       = 0;
      nHadr       = 0;
      // Default antenna properties
      Qold        = 0.0;
      unOrdering  = false;
      new1=Particle(0,51,iOld1,iOld2,0,0,0,0,0.);
      new2=Particle(0,51,iOld1,iOld2,0,0,0,0,0.);
      new3=Particle(0,51,iOld1,iOld2,0,0,0,0,0.);      
      // Set pointers
      new1.setEvtPtr(&event);
      new2.setEvtPtr(&event);
      new3.setEvtPtr(&event);
    }

    // Function to reset all trial generators for this branchelemental
    void clearTrialGenerators() {
      trialGenPtrSav.resize(0);
      iAntPhysSav.resize(0);
      isSwappedSav.resize(0);
      hasSavedTrial.resize(0);
      scaleSav.resize(0);
      scaleOldSav.resize(0);
      zMinSav.resize(0);
      zMaxSav.resize(0);
      colFacSav.resize(0);
      alphaSav.resize(0);
      headRoomSav.resize(0);
      Q2reUnOrd.resize(0);
      Q2reOrd.resize(0);
      nVeto       = 0;
      nHull       = 0;
      nHadr       = 0;
    }

    // Add a trial generator to this branchelemental. Return name of generator.
    void addTrialGenerator(int iAntPhysIn, bool swapIn,  
                          TrialGeneratorFSR* trialGenPtrIn) {
      trialGenPtrSav.push_back(trialGenPtrIn);
      iAntPhysSav.push_back(iAntPhysIn);
      isSwappedSav.push_back(swapIn);
      hasSavedTrial.push_back(false);
      scaleSav.push_back(-1.0);
      scaleOldSav.push_back(-1.0);
      zMinSav.push_back(0.0);
      zMaxSav.push_back(0.0);
      colFacSav.push_back(0.0);
      alphaSav.push_back(0.0);
      headRoomSav.push_back(1.0);
      Q2reUnOrd.push_back(0.0);
      Q2reOrd.push_back(0.0);
    }
    
    // Set the restart scales
    void setQ2reUnOrd(double Q2gEmi, double Q2gSplit, int iTrial, bool doMin = false) {
      if (iTrial > -1) {
        double Q2 = Q2gEmi;
        if (iAntPhysSav[iTrial] == iQGsplit ||
            iAntPhysSav[iTrial] == iGGsplit) Q2 = Q2gSplit;
        Q2reUnOrd[iTrial] = ( doMin ? min(Q2,Q2reUnOrd[iTrial]) : Q2 );
        return;
      }
      for (int i=0; i<(int)trialGenPtrSav.size(); i++) {
        double Q2 = Q2gEmi;
        if (iAntPhysSav[i] == iQGsplit || iAntPhysSav[i] == iGGsplit)
          Q2 = Q2gSplit;
        Q2reUnOrd[i] = ( doMin ? min(Q2,Q2reUnOrd[i]) : Q2 );
      }
    }
    void setQ2reOrd(double Q2gEmi, double Q2gSplit, int iTrial, bool doMin = false) {
      if (iTrial > -1) {
        double Q2 = Q2gEmi;
        if (iAntPhysSav[iTrial] == iQGsplit ||
            iAntPhysSav[iTrial] == iGGsplit) Q2 = Q2gSplit;
        Q2reOrd[iTrial] = ( doMin ? min(Q2,Q2reOrd[iTrial]) : Q2 );
        return;
      }
      for (int i=0; i<(int)trialGenPtrSav.size(); i++) {
        double Q2 = Q2gEmi;
        if (iAntPhysSav[i] == iQGsplit || iAntPhysSav[i] == iGGsplit)
          Q2 = Q2gSplit;
        Q2reOrd[i] = ( doMin ? min(Q2,Q2reOrd[i]) : Q2 );
      }
    }
    
    // Function to return number of trial generators for this antenna
    int nTrialGenerators() const { return trialGenPtrSav.size(); }

    // Save a generated trial branching
    void saveTrial(int iTrial,double qOld,double qTrial,double zMin,double zMax,
                   double colFac,double alphaEff, double headRoom = 1.0) {
      hasSavedTrial[iTrial] = true;
      scaleSav[iTrial]      = qTrial;
      scaleOldSav[iTrial]   = qOld;
      zMinSav[iTrial]       = zMin;
      zMaxSav[iTrial]       = zMax;
      colFacSav[iTrial]     = colFac;
      alphaSav[iTrial]      = alphaEff;
      headRoomSav[iTrial]   = headRoom;
    }

    // Generate invariants for saved branching
    bool genTrialInvariants(double& y12, double& y23, int iTrial = -1) {
      // Automatically determine which trial function to use if -1 input
      int iGen = iTrial;
      if (iGen == -1) iGen = getTrialIndex();
      if (iGen <= -1) return false; 
      double z = trialGenPtrSav[iGen]->genZ(zMinSav[iGen],zMaxSav[iGen]);
      // Check physical phase space (note: this only checks massless hull)
      // (Use absolute z value since negative z values are used to indicate 
      //  swapped invariants for mD ordering)
      double yE = pow2(scaleSav[iGen]/mSav);
      if (abs(z) < trialGenPtrSav[iGen]->getZmin(yE) || 
          abs(z) > trialGenPtrSav[iGen]->getZmax(yE)) return false;
      // Convert to y12 , y23
      y12 = trialGenPtrSav[iGen]->getYij(yE,z);
      y23 = trialGenPtrSav[iGen]->getYjk(yE,z);
      return true;
    }

    // Get trial function index of winner
    int getTrialIndex() const {
      double qMax = 0.0;
      int iMax    = -1;
      for (int i = 0; i < int(scaleSav.size()); ++i) {
        if (hasSavedTrial[i]) {
          double qSav = scaleSav[i];
          if (qSav > qMax) {
            qMax = qSav;
            iMax = i;
          }
        }
      }
      return iMax;
    }
        
    // Check if a saved trial exists for a particular trialGenerator
    bool hasTrial(int iTrial) const {
      if (iTrial < int(hasSavedTrial.size())) return hasSavedTrial[iTrial];
      else return false;
    }

    // Get whether physical antenna is swapped 
    bool getIsSwapped(int iTrial=-1) const {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return isSwappedSav[iTrial];
    }
    
    // Get physical antenna function index of winner
    int getPhysIndex(int iTrial=-1) const {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return iAntPhysSav[iTrial];
    }
    
    // Get scale for a specific saved trial
    double getTrialScale(int iTrial) const {
      if (iTrial < int(scaleSav.size())) return scaleSav[iTrial];
      else return -1.0;
    }

    // Get scale of winner
    double getTrialScale() const {
      double qMax = 0.0;
      for (int i = 0; i < int(scaleSav.size()); ++i) {
        if (hasSavedTrial[i]) qMax = max(qMax,scaleSav[i]);
        else printOut("getTrialScale","Error! "
                      "not all trials have saved scales");
      }
      return qMax;
    }

    // Get color factor 
    double getColFac(int iTrial=-1) {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return colFacSav[iTrial];
    }
    
    // Get headroom factor
    double getHeadRoomFactor(int iTrial=-1) {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return headRoomSav[iTrial];
    }
    
    // Get alpha S
    double getAlphaTrial(int iTrial=-1) {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return alphaSav[iTrial];
    }
    
    // Flag to set if a saved trial should be ignored and a new one generated
    // Default value -1 : force all trials to renew
    void renewTrial(int iTrial = -1) {
      if (iTrial >= 0) hasSavedTrial[iTrial] = false;
      else {
        for (iTrial = 0; iTrial < int(hasSavedTrial.size()); ++iTrial)
          hasSavedTrial[iTrial] = false;
      }
    }

    // List function
    void list(ostream& os=cout, bool header=false, bool footer=false) const; 
    
    // Data storage members 
    int iOld1, iOld2;    
    Particle old1, old2;
    Particle new1, new2, new3;  
    // Nearest-Neighbour indices
    int iL,iR;
    // 
    int system,nVeto,nHull,nHadr;
    double Qold;
    // Ultimate ancestor (e.g., if part of decay system)
    int iAncestor;
    // Has trial branching, and whether strong ordering should be imposed
    bool unOrdering;  

    // Trial Generators and properties of saved trials
    vector<TrialGeneratorFSR*> trialGenPtrSav;
    vector<double> zMinSav, zMaxSav, colFacSav, alphaSav;
    vector<double> scaleSav, scaleOldSav, headRoomSav;
    vector<bool> hasSavedTrial, isSwappedSav;
    vector<int> iAntPhysSav;
    
    // Restart scales for ordered / unodered
    vector<double> Q2reUnOrd, Q2reOrd;

   private:

    // Saved antenna invariant mass value
    double mSav, sSav;

  };
  
//**************************************************************************

// The ISRBranchelemental class: 
// Generic 2->3 branching, similar to the Branchelemental class
// for the Time Shower. Added by NF to include some additional
// parameters.

  class BranchelementalISR {
    
  public:
    
    // Constructors
    // Create empty branchelemental
    BranchelementalISR() : system(0) {}
    // Create branchelemental with iOld1In,iOld2In as parents
    // and initialize the PDT pointers for the new ones
    BranchelementalISR(Event& event, int iOld1In, int iOld2In, bool isVal1In, bool isVal2In)
    : system(0) {
      reset(event, iOld1In, iOld2In, isVal1In, isVal2In);
    };
    
    // Antenna Mass (negative if spacelike virtual)
    double mCalc() {
      sSav = m2(old1,old2);
      mSav = (sSav >= 0) ? sqrt(sSav) : -sqrt(-sSav);
      return mSav;
    }
    double m() const { return mSav; }
    double s() const { return sSav; }
    // Dot products of daughters
    double s12() const { return m2(new1,new2) - new1.m2() - new2.m2(); }
    double s23() const { return m2(new2,new3) - new2.m2() - new3.m2(); }
    double s13() const { return m2(new1,new3) - new1.m2() - new3.m2(); }
    
    // Parent energies
    double e1() const { return old1.e(); }
    double e2() const { return old2.e(); }
    
    // Is this an II?
    bool isII() const { return ( !old1.isFinal() && !old2.isFinal() ); }
    // Is 1 a side A (p+) guy? (need to know for pdfs in IF)
    bool is1A() const { return ( old1.pz() > 0 ); }

    // Reset and set trial true/false
    void reset(Event& event, int iOld1In, int iOld2In, bool isVal1In, bool isVal2In) {
      // Make local copy of input particles
      old1        = event[iOld1In];
      old2        = event[iOld2In];
      bool isIIN  = ( !old1.isFinal() && !old2.isFinal() );
      // Make sure that for II 1 is the guy with p+ and 2 is the guy with p-
      //                    IF 1 is the initial gut and 2 is the final guy
      bool swap = false;
      if (isIIN) swap = (old1.pz() < 0.0);
      else swap = (old1.isFinal());
      if (swap) {
        // Update local copy of input particles
        old1        = event[iOld2In];
        old2        = event[iOld1In];
        // Valence?
        isVal1      = isVal2In;
        isVal2      = (isIIN ? isVal1In : false);
        // Indices of parents
        iOld1       = iOld2In;
        iOld2       = iOld1In;
      }
      else {
        // Valence?
        isVal1      = isVal1In;
        isVal2      = (isIIN ? isVal2In : false);
        // Indices of parents
        iOld1       = iOld1In;
        iOld2       = iOld2In;
      }
      // Compute and store antenna invariant mass
      mCalc();
      // Indices of neighbors
      iL          = 0; 
      iR          = 0;
      // Trial Generators
      clearTrialGenerators();
      nVeto       = 0;
      nHull       = 0;
      nHadr       = 0;
      // Default antenna properties
      unOrdering  = false;
      new1=Particle(0,-41,iOld1,iOld2,0,0,0,0,0.); // 41 = incoming on spacelike main branch
      new2=Particle(0,43,iOld1,iOld2,0,0,0,0,0.); // Emission 43 = outgoing produced by a branching
      new3=Particle(0,isII()?-41:44,iOld1,iOld2,0,0,0,0,0.); // 44 = outgoing shifted by a branching
      // Set pointers
      new1.setEvtPtr(&event);
      new2.setEvtPtr(&event);
      new3.setEvtPtr(&event);
    }

    // Function to reset all trial generators for this branchelemental
    void clearTrialGenerators() {
      trialGenPtrSav.resize(0);
      Q2reUnOrd.resize(0);
      Q2reOrd.resize(0);
      iAntPhysSav.resize(0);
      isSwappedSav.resize(0);
      hasSavedTrial.resize(0);
      scaleSav.resize(0);
      scaleOldSav.resize(0);
      zMinSav.resize(0);
      zMaxSav.resize(0);
      colFacSav.resize(0);
      alphaSav.resize(0);
      physPDFratioSav.resize(0);
      trialPDFratioSav.resize(0);
      trialPimpSav.resize(0);
      extraMassPDFfactorSav.resize(0);
      headRoomSav.resize(0);
      nShouldRescue.resize(0);
      nVeto       = 0;
      nHull       = 0;
      nHadr       = 0;
    }

    // Add a trial generator to this branchelemental. Return name of generator.
    void addTrialGenerator(int iAntPhysIn, bool swapIn,  
                          TrialGeneratorISR* trialGenPtrIn) {
      trialGenPtrSav.push_back(trialGenPtrIn);
      Q2reUnOrd.push_back(0.0);
      Q2reOrd.push_back(0.0);
      iAntPhysSav.push_back(iAntPhysIn);
      isSwappedSav.push_back(swapIn);
      hasSavedTrial.push_back(false);
      scaleSav.push_back(-1.0);
      scaleOldSav.push_back(-1.0);
      zMinSav.push_back(0.0);
      zMaxSav.push_back(0.0);
      colFacSav.push_back(0.0);
      alphaSav.push_back(0.0);
      physPDFratioSav.push_back(0.0);
      trialPDFratioSav.push_back(0.0);
      trialPimpSav.push_back(1.0);
      extraMassPDFfactorSav.push_back(0.0);
      headRoomSav.push_back(1.0);
      nShouldRescue.push_back(0);
    }
    
    // Reset the restart scale to the phase space limit
    void setQ2reUnOrdToPSlimit(double eA, double eB) {
      for (int i=0; i<(int)trialGenPtrSav.size(); i++) {
        Q2reUnOrd[i] = trialGenPtrSav[i]->getQ2max(s(), e1(),
                                                   is1A() ? eA : eB);
      }
    }
    void setQ2reOrdToPSlimit(double eA, double eB) {
      for (int i=0; i<(int)trialGenPtrSav.size(); i++) {
        Q2reOrd[i] = trialGenPtrSav[i]->getQ2max(s(), e1(),
                                                 is1A() ? eA : eB);
      }
    }
    
    // Set the restart scales
    void setQ2reUnOrd(double Q2In, int iTrial, bool doMin = false) {
      if (iTrial > -1) {
        Q2reUnOrd[iTrial] = ( doMin ? min(Q2In,Q2reUnOrd[iTrial]) : Q2In );
        return;
      }
      for (int i=0; i<(int)trialGenPtrSav.size(); i++)
        Q2reUnOrd[i] = ( doMin ? min(Q2In,Q2reUnOrd[i]) : Q2In );
    }
    void setQ2reOrd(double Q2In, int iTrial, bool doMin = false) {
      if (iTrial > -1) {
        Q2reOrd[iTrial] = ( doMin ? min(Q2reOrd[iTrial],Q2In) : Q2In );
        return;
      }
      for (int i=0; i<(int)trialGenPtrSav.size(); i++)
        Q2reOrd[i] = ( doMin ? min(Q2reOrd[i],Q2In) : Q2In );
    }
    
    // Add to and get rescue levels
    void addRescue(int iTrial) {
      nShouldRescue[iTrial]++;
    }
    int getNshouldRescue(int iTrial) {
      return nShouldRescue[iTrial];
    }
    void resetRescue() {
      for (int i=0; i<(int)nShouldRescue.size(); i++)
        nShouldRescue[i] = 0;
    }
    
    // Function to return number of trial generators for this antenna
    int nTrialGenerators() const { return trialGenPtrSav.size(); }

    // Save a generated trial branching
    void saveTrial(int iTrial,double qOld,double qTrial,double zMin,double zMax,
                   double colFac,double alphaEff, double pdfRatio,
                   double trialPimp, double extraMpdf, double headRoom = 1.0) {
      hasSavedTrial[iTrial]         = true;
      scaleSav[iTrial]              = qTrial;
      scaleOldSav[iTrial]           = qOld;
      zMinSav[iTrial]               = zMin;
      zMaxSav[iTrial]               = zMax;
      colFacSav[iTrial]             = colFac;
      alphaSav[iTrial]              = alphaEff;
      trialPDFratioSav[iTrial]      = pdfRatio;
      trialPimpSav[iTrial]          = trialPimp;
      extraMassPDFfactorSav[iTrial] = extraMpdf;
      headRoomSav[iTrial]           = headRoom;
    }
    
    // Add the physical pdf ratio
    void addPDF(int iTrial,double pdfRatio) {
      physPDFratioSav[iTrial] = pdfRatio;
    }

    // Generate invariants for saved branching
    bool genTrialInvariants(double& s1j, double& sj2, 
                            double eUsed, int iTrial = -1) {
      // Automatically determine which trial function to use if -1 input
      int iGen = iTrial;
      if (iGen == -1) iGen = getTrialIndex();
      if (iGen <= -1) return false; 
      double z = trialGenPtrSav[iGen]->genZ(zMinSav[iGen],zMaxSav[iGen]);
      // Check physical phase space (note: this only checks massless hull)
      // (Use absolute z value since negative z values are used to indicate 
      //  swapped invariants for mD ordering)
      double Q2E = pow2(scaleSav[iGen]);
      if ( abs(z) < trialGenPtrSav[iGen]->getZmin(Q2E, sSav, e1(), eUsed) || 
           abs(z) > trialGenPtrSav[iGen]->getZmax(Q2E, sSav, e1(), eUsed) ) 
        return false;
      // Convert to s1j, sj2
      s1j = trialGenPtrSav[iGen]->getS1j(Q2E,z,sSav);
      sj2 = trialGenPtrSav[iGen]->getSj2(Q2E,z,sSav);
      return true;
    }

    // Get trial function index of winner
    int getTrialIndex() const {
      double qMax = 0.0;
      int iMax    = -1;
      for (int i = 0; i < int(scaleSav.size()); ++i) {
        if (hasSavedTrial[i]) {
          double qSav = scaleSav[i];
          if (qSav > qMax) {
            qMax = qSav;
            iMax = i;
          }
        }
      }
      return iMax;
    }
        
    // Check if a saved trial exists for a particular trialGenerator
    bool hasTrial(int iTrial) const {
      if (iTrial < int(hasSavedTrial.size())) return hasSavedTrial[iTrial];
      else return false;
    }

    // Get whether physical antenna is swapped 
    bool getIsSwapped(int iTrial=-1) const {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return isSwappedSav[iTrial];
    }
    
    // Get physical antenna function index of winner
    int getPhysIndex(int iTrial=-1) const {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return iAntPhysSav[iTrial];
    }
    
    // Get scale for a specific saved trial
    double getTrialScale(int iTrial) const {
      if (iTrial < int(scaleSav.size())) return scaleSav[iTrial];
      else return -1.0;
    }

    // Get scale of winner
    double getTrialScale() const {
      double qMax = 0.0;
      for (int i = 0; i < int(scaleSav.size()); ++i) {
        if (hasSavedTrial[i]) qMax = max(qMax,scaleSav[i]);
        else printOut("getTrialScale","Error! "
                      "not all trials have saved scales");
      }
      return qMax;
    }

    // Get color factor 
    double getColFac(int iTrial=-1) {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return colFacSav[iTrial];
    }
    
    // Get headroom factor
    double getHeadRoomFactor(int iTrial=-1) {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return headRoomSav[iTrial];
    }
    
    // Get alpha S
    double getAlphaTrial(int iTrial=-1) {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return alphaSav[iTrial];
    }
    
    // Get pdf ratio
    double getPDFratioTrial(int iTrial=-1) {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return trialPDFratioSav[iTrial];
    }
    
    // Get pdf ratio
    double getPDFratioPhys(int iTrial=-1) {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return physPDFratioSav[iTrial];
    }
    
    // Get trial pimp
    double getPimpTrial(int iTrial=-1) {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return trialPimpSav[iTrial];
    }
    
    // Get the extra factor when getting rid of the heavy quarks
    double getExtraMassPDFfactor(int iTrial=-1) {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return extraMassPDFfactorSav[iTrial];
    }
    
    // Flag to set if a saved trial should be ignored and a new one generated
    // Default value -1 : force all trials to renew
    void renewTrial(int iTrial = -1) {
      if (iTrial >= 0) hasSavedTrial[iTrial] = false;
      else {
        for (iTrial = 0; iTrial < int(hasSavedTrial.size()); ++iTrial)
          hasSavedTrial[iTrial] = false;
      }
    }

    // List function
    void list(ostream& os=cout, bool header=false, bool footer=false) const; 
    
    // Data storage members
    int iOld1, iOld2;
    Particle old1, old2;
    bool isVal1, isVal2;
    Particle new1, new2, new3;
    // Color, not obvious, since for eg gg->H we have to II antennae
    int col;
    // Nearest-Neighbour indices
    int iL, iR;
    // System and counter for vetos
    int system, nVeto, nHull, nHadr;
    // Has trial branching, and whether strong ordering should be imposed
    bool unOrdering; 
    // Do we have to force a splitting (heavy quarks)
    bool forceSplittingSav;

    // Trial Generators and properties of saved trials
    vector<TrialGeneratorISR*> trialGenPtrSav;
    vector<double> zMinSav, zMaxSav, colFacSav, alphaSav;
    vector<double> physPDFratioSav, trialPDFratioSav;
    vector<double> extraMassPDFfactorSav;
    vector<double> scaleSav, scaleOldSav, headRoomSav, trialPimpSav;
    vector<bool> hasSavedTrial, isSwappedSav;
    vector<int> iAntPhysSav, nShouldRescue;
    // Note: isSwapped = true for II means we use side B trial generators
    //                            IF means is1A is false, 1 is a side B
    
    // Restart scales for ordered / unodered
    vector<double> Q2reUnOrd, Q2reOrd;

   private:

    // Saved antenna invariant mass value
    double mSav, sSav;

  };
  
}

#endif


