// TrialGeneratorsISR.h is a part of the VINCIA plug-in to the PYTHIA 8 
// generator.
// Copyright (C) 2016 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//                             http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the TrialGeneratorISR and 
// derived classes. Added by NF for ISR shower.

#ifndef VINCIATrialGeneratorsISR_H
#define VINCIATrialGeneratorsISR_H

#include "Pythia8/Pythia.h"
#include "Vincia/VinciaCommon.h"

namespace Vincia {

  // Instructions to create a new trial function:
  // See the classes below for examples and the VINCIA authors' compendium 
  // for existing definitions and derivations.
  // 1) Create a class inheriting from TrialGenerator
  // 2) Supply the basic trial function, aTrial()
  // 3) Choose a zeta definition for each predefined QE choice
  //       Encode the relevant zeta integrals, getIz()
  //       Encode the relevant zeta limits, getZmin() and getZmax()
  //       Encode the phase-space inversions, getYij() and getYjk()
  // 4) Encode the two getQ2 functions, for constant and running alphaS
  // 5) If needed, revise the init() function as well to suit your needs
  // 6) IMPORTANT! Add your definitions and derivations to the relevant 
  //    parts of the VINCIA authors' compendium, so that it is kept up to date.
  
  // Instructions to create a new evolution scale:
  // See the classes below for examples and the VINCIA authors' compendium
  // for existing definitions and derivations.
  // 1) For each existing TrialGenerator (i.e., each trial function that
  //    VINCIA already contains), select a zeta definition to use with 
  //    your new evolution scale.
  //       Encode the relevant zeta integrals, by adding to getIz()
  //       Encode the relevant zeta limits, by adding to getZmin() and getZmax()
  //       Encode the phase-space inversions, getYij() and getYjk()
  //    Note: the only function that it is really mandatory to add your 
  //    evolution scale to is the default soft-eikonal one encoded in the
  //    base class. The finite one may or may not be relevant, and the 
  //    additional collinear trial functions are only used for sector showers
  //    and for gluon splittings, which you may not be interested in modifying.
  //    The best is of course to ensure maximal compatibility by adding the
  //    relevant expressions to all trial functions.
  // 2) Add the solutions to your evolution integrals to the two getQ() 
  //    functions, for constant and running alphaS
  // 3) Revise the init() function to increase the number of recognized
  //    evolution scales
  // 4) IMPORTANT! Add your definitions and derivations to the relevant 
  //    parts of the VINCIA authors' compendium, so that it is kept up to date.

  // *****
  // Base class TrialGeneratorISR
  // Note: base class is coded for a soft-eikonal trial function

  class TrialGeneratorISR {
        
  public: 
       
    // -----------------------------------------------------------------
    // Constructor and pointer initialization (non-virtual)
    // -----------------------------------------------------------------

    // Constructor
    TrialGeneratorISR() : isInitPtr(false), isInit(false) {};
        
    // Init Pointers
    void initPtr(Rndm* rndmPtrIn, Settings* settingsPtrIn, Info* infoPtrIn);

    // -----------------------------------------------------------------
    // Virtual functions 
    // Defining the particular kind of trial generator
    // -----------------------------------------------------------------

    // Name of trial generator
    virtual string name() {
        return "TrialGeneratorISR";
    }
    // For statistics
    virtual int statInt() { return 0; }

    // Initialize
    virtual void init(double mcIn, double mbIn);

    // Trial antenna function
    virtual double aTrial(double saj, double sjb, double sAB);
    
    // Modified trial antenna function, used when combining accept
    // probabilities for gluon emission
    virtual double aTrialMod(double saj, double sjb, double sAB,
                             double eA, double eB, double ea, double eb);

    // Physical antenna function
    virtual double aPhys(double saj, double sjb, double sAB, int idA, int idB);
  
    // Evolution scale
    virtual double getQ2(double saj, double sjb, double sAB) { 
      return (evNormIIpT2Sav*saj*sjb/(saj+sjb+sAB)); 
    }
    virtual double getQ2max(double sAB, double, double) { 
      return (evNormIIpT2Sav*0.25*pow2(shhSav-sAB)/shhSav);
    }

    // Generate a new Q value, with first-order running alphaS
    virtual double genQ2(double q2old, double sAB, double zMin, double zMax, 
                         double colFac, double PDFratio, 
                         double b0, double kR, double Lambda,
                         double eA, double eB, double headroomFactor=1.0);

    // Generate a new Q value, with constant trial alphaS
    virtual double genQ2(double q2old, double sAB, double zMin, double zMax, 
                         double colFac, double alphaSvalue, double PDFratio, 
                         double eA, double eB, double headroomFactor=1.0);

    // Generate a new Q value, use in the very unordered region
    virtual double genQ2withQN(double q2old, double q2hat, double N, double sAB, 
                               double zMin, double zMax, 
                               double colFac, double alphaSvalue, double PDFratio, 
                               double eA, double eB, double headroomFactor=1.0);

    // Generate a new Q value, with running of the PDFs towards the mass threshold
    virtual double genQ2(double /*q2old*/, double /*sAB*/, double /*zMin*/, double /*zMax*/, 
                         double /*colFac*/, double /*alphaSvalue*/, double /*PDFratio*/, 
                         int /*idA*/, int /*idB*/, double /*eA*/, double /*eB*/, 
                         bool /*useMpdf*/, double /*headroomFactor*/) { return 0.0; }

    // Generate a new zeta value in [zMin,zMax]
    virtual double genZ(double zMin, double zMax);
    
    // The zeta integral
    virtual double getIz(double zMin, double zMax);

    // The zeta boundaries, for a given value of the evolution scale
    virtual double getZmin(double Qt2, double sAB, double eA, double eAused);
    virtual double getZmax(double Qt2, double sAB, double eA, double eAused);
    
    // Inverse transforms
    // To obtain saj and sjb from Qt2 and zeta
    virtual double getS1j(double Qt2, double zeta, double sAB);
    virtual double getSj2(double Qt2, double zeta, double sAB);
    
    // Trial PDF ratio
    virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                 int iSys, bool Rfunity,
                                 int idA, int idB, double eA, double eB, 
                                 double Qt2A, double Qt2B, bool useSaved = false);
    
    // Physical PDF ratio
    virtual double physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                int iSys, bool Rfunity,
                                int idA, int idB, double eA, double eB,
                                double ea, double eb, double Qt2A, double Qt2B);

    // -----------------------------------------------------------------
    // Non-Virtual functions
    // -----------------------------------------------------------------

    bool checkInit();
    
    int getGconvID() { return gConvIDSav; }
    
    int getNpdfOB() { return nPDFoutsideBound; }

  protected:

    // Pointers
    Rndm* rndmPtr;
    Settings* settingsPtr;
    Info* infoPtr;
    
    double TINYPDF;
    double TINYPDFtrial;
    
    // Use m or pT evolution for collinear singularities
    bool useMevolSav;

    // s hadron hadron
    double shhSav;
    
    // For conversion trial generators
    int gConvIDSav;
    int nGtoQISRSav;
    
    // Masses
    double mbSav;
    double mcSav;
    
    // Counter for PDF evaluation outside of boundaries
    int nPDFoutsideBound;
    int nReportPDFbound;
    
    // Should we use random xa in PDF trial?
    bool useRandomX;
    double trialPDFratioSav;

    // Normalization factors
    double evNormIIpT2Sav;
    double evNormMsplitISav;
    double evNormMconvISav;
    double evNormIFpT2Sav;
    double evNormMsplitFSav;
    
    // Exponent for x in the pdf ratios
    double alphaSplitSav; // II when evolved in mass
    double alphaConvSav;
    double betaQCollISav; // IF always evolved in pT
    double betaGColl1ISav;
    double betaGColl2ISav;
    double betaSplitISav; // IF for both
    double betaConvISav;

  private:

    // Status
    bool isInitPtr, isInit;

    // Verbose level
    int verbose;
  };


  //----------------------------------------------------------------------
  // Soft-eikonal trial function for II (clone of base class)
  //----------------------------------------------------------------------

  class TrialIISoft : public TrialGeneratorISR {

  public: 

    // Name of trial generator
    virtual string name() {
        return "TrialIISoft";
    }
    // For statistics
    virtual int statInt() { return 1; }

  };


  //----------------------------------------------------------------------
  // A collinear trial function for II (derived base class)
  //----------------------------------------------------------------------
  
  class TrialIIGCollA : public TrialGeneratorISR {

  public: 

   // Name of trial generator
    virtual string name() {
      return "TrialIIGCollA";
    }
    // For statistics
    virtual int statInt() { return 2; }

    // Trial antenna function
    virtual double aTrial(double saj, double sjb, double sAB);
    
    // Modified trial antenna function, used when combining accept
    // probabilities for gluon emission
    virtual double aTrialMod(double saj, double sjb, double sAB,
                             double eA, double eB, double ea, double eb);

    // Physical antenna function
    virtual double aPhys(double saj, double sjb, double sAB, int idA, int idB);
  
    // Evolution scale
    virtual double getQ2(double saj, double sjb, double sAB) { 
      return (evNormIIpT2Sav*saj*sjb/(saj+sjb+sAB)); 
    }
    virtual double getQ2max(double sAB, double, double) { 
      return (evNormIIpT2Sav*0.25*pow2(shhSav-sAB)/shhSav);
    }

    // Generate a new Q value, with first-order running alphaS
    virtual double genQ2(double q2old, double sAB, double zMin, double zMax, 
                         double colFac, double PDFratio, 
                         double b0, double kR, double Lambda,
                         double eA, double eB, double headroomFactor=1.0);

    // Generate a new Q value, with constant trial alphaS
    virtual double genQ2(double q2old, double sAB, double zMin, double zMax, 
                         double colFac, double alphaSvalue, double PDFratio, 
                         double eA, double eB, double headroomFactor=1.0);

    // Generate a new Q value, use in the very unordered region
    virtual double genQ2withQN(double q2old, double q2hat, double N, double sAB, 
                               double zMin, double zMax, 
                               double colFac, double alphaSvalue, double PDFratio, 
                               double eA, double eB, double headroomFactor=1.0);

    // Generate a new zeta value in [zMin,zMax]
    virtual double genZ(double zMin, double zMax);
        
    // The zeta integral
    virtual double getIz(double zMin, double zMax);

    // The zeta boundaries, for a given value of the evolution scale
    virtual double getZmin(double Qt2, double sAB, double eA, double eAused);
    virtual double getZmax(double Qt2, double sAB, double eA, double eAused);
        
    // Inverse transforms
    // To obtain saj and sjb from Qt2 and zeta
    virtual double getS1j(double Qt2, double zeta, double sAB);
    virtual double getSj2(double Qt2, double zeta, double sAB);    
    
    // Trial PDF ratio
    virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                 int iSys, bool Rfunity,
                                 int idA, int idB, double eA, double eB, 
                                 double Qt2A, double Qt2B, bool useSaved = false);
    
    // Physical PDF ratio
    virtual double physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                int iSys, bool Rfunity,
                                int idA, int idB, double eA, double eB,
                                double ea, double eb, double Qt2A, double Qt2B);

  };


  //----------------------------------------------------------------------
  // B collinear trial function for II (derived base class)
  //----------------------------------------------------------------------
  
  class TrialIIGCollB : public TrialIIGCollA {

  public: 

    // Name of trial generator
    virtual string name() {
        return "TrialIIGCollB";
    }

    // Trial antenna function
    virtual double aTrial(double saj, double sjb, double sAB) {
      return TrialIIGCollA::aTrial(sjb, saj, sAB);
    }
    
    // Modified trial antenna function, used when combining accept
    // probabilities for gluon emission
    virtual double aTrialMod(double saj, double sjb, double sAB,
                             double eA, double eB, double ea, double eb) {
      return TrialIIGCollA::aTrialMod(sjb, saj, sAB, eB, eA, eb, ea);
    }

    // Physical antenna function
    virtual double aPhys(double saj, double sjb, double sAB, int idA, int idB) {
      return TrialIIGCollA::aPhys(sjb, saj, sAB, idB, idA);
    }
        
    // Inverse transforms
    // To obtain saj and sjb from Qt2 and zeta
    virtual double getS1j(double Qt2, double zeta, double sAB) {
      return TrialIIGCollA::getSj2(Qt2, zeta, sAB);
    }
    virtual double getSj2(double Qt2, double zeta, double sAB) {
      return TrialIIGCollA::getS1j(Qt2, zeta, sAB);
    }
    
    // Trial PDF ratio
    virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                 int iSys, bool Rfunity,
                                 int idA, int idB, double eA, double eB, 
                                 double Qt2A, double Qt2B, bool useSaved = false) {
      return TrialIIGCollA::trialPDFratio(beamBPtr, beamAPtr, iSys, Rfunity,
                                          idB, idA, eB, eA, 
                                          Qt2B, Qt2A, useSaved);
    }
    
    // Physical PDF ratio
    virtual double physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                int iSys, bool Rfunity,
                                int idA, int idB, double eA, double eB,
                                double ea, double eb, double Qt2A, double Qt2B) {
      return TrialIIGCollA::physPDFratio(beamBPtr, beamAPtr, iSys, Rfunity,
                                         idB, idA, eB, eA, eb, ea, Qt2B, Qt2A);
    }

  };


  //----------------------------------------------------------------------
  // A splitting trial function for II (derived base class), q<-gqbar
  //----------------------------------------------------------------------
  
  class TrialIISplitA : public TrialGeneratorISR {

  public: 

   // Name of trial generator
    virtual string name() {
      return "TrialIISplitA";
    }
    // For statistics
    virtual int statInt() { return 3; }

    // Trial antenna function
    virtual double aTrial(double saj, double sjb, double sAB);
    
    // Modified trial antenna function, used when combining accept
    // probabilities for gluon emission
    virtual double aTrialMod(double, double, double, double, double, double, double) { 
      printf("%s::aTrialMod Not existing!\n", name().c_str()); 
      return 0.0; 
    }

    // Physical antenna function
    virtual double aPhys(double saj, double sjb, double sAB, int idA, int idB);
  
    // Evolution scale
    virtual double getQ2(double saj, double sjb, double sAB) { 
      return ( (useMevolSav) ? (evNormMsplitISav*saj)
                             : (evNormIIpT2Sav*saj*sjb/(saj+sjb+sAB)) );
    }
    virtual double getQ2max(double sAB, double, double) { 
      return ( (useMevolSav) ? (evNormMsplitISav*(shhSav-sAB)) 
                             : (evNormIIpT2Sav*0.25*pow2(shhSav-sAB)/shhSav) );
    }

    // Generate a new Q value, with first-order running alphaS
    virtual double genQ2(double q2old, double sAB, double zMin, double zMax, 
                         double colFac, double PDFratio, 
                         double b0, double kR, double Lambda,
                         double eA, double eB, double headroomFactor=1.0);

    // Generate a new Q value, with constant trial alphaS
    virtual double genQ2(double q2old, double sAB, double zMin, double zMax, 
                         double colFac, double alphaSvalue, double PDFratio, 
                         double eA, double eB, double headroomFactor=1.0);

    // Generate a new Q value, use in the very unordered region
    virtual double genQ2withQN(double q2old, double q2hat, double N, double sAB, 
                               double zMin, double zMax, 
                               double colFac, double alphaSvalue, double PDFratio, 
                               double eA, double eB, double headroomFactor=1.0);

    // Generate a new Q value, with running of the PDFs towards the mass threshold
    virtual double genQ2(double q2old, double sAB, double zMin, double zMax, 
                         double colFac, double alphaSvalue, double PDFratio, 
                         int idA, int idB, double eA, double eB, bool useMpdf,
                         double headroomFactor=1.0);

    // Generate a new zeta value in [zMin,zMax]
    virtual double genZ(double zMin, double zMax);
        
    // The zeta integral
    virtual double getIz(double zMin, double zMax);

    // The zeta boundaries, for a given value of the evolution scale
    virtual double getZmin(double Qt2, double sAB, double eA, double eAused);
    virtual double getZmax(double Qt2, double sAB, double eA, double eAused);
        
    // Inverse transforms
    // To obtain saj and sjb from Qt2 and zeta
    virtual double getS1j(double Qt2, double zeta, double sAB);
    virtual double getSj2(double Qt2, double zeta, double sAB);    
    
    // Trial PDF ratio
    virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                 int iSys, bool Rfunity,
                                 int idA, int idB, double eA, double eB, 
                                 double Qt2A, double Qt2B, bool useSaved = false);
    
    // Physical PDF ratio
    virtual double physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                int iSys, bool Rfunity,
                                int idA, int idB, double eA, double eB,
                                double ea, double eb, double Qt2A, double Qt2B);

  };


  //----------------------------------------------------------------------
  // B splitting trial function for II (derived base class), q<-gqbar
  //----------------------------------------------------------------------
  
  class TrialIISplitB : public TrialIISplitA {

  public: 

    // Name of trial generator
    virtual string name() {
        return "TrialIISplitB";
    }

    // Trial antenna function
    virtual double aTrial(double saj, double sjb, double sAB) {
      return TrialIISplitA::aTrial(sjb, saj, sAB);
    }
    
    // Modified trial antenna function, used when combining accept
    // probabilities for gluon emission
    virtual double aTrialMod(double, double, double, double, double, double, double) { 
      printf("%s::aTrialMod Not existing!\n", name().c_str()); 
      return 0.0; 
    }

    // Physical antenna function
    virtual double aPhys(double saj, double sjb, double sAB, int idA, int idB) {
      return TrialIISplitA::aPhys(sjb, saj, sAB, idB, idA);
    }

    // Evolution scale
    virtual double getQ2(double saj, double sjb, double sAB) {
      return TrialIISplitA::getQ2(sjb, saj, sAB);
    }

    // Generate a new Q value, with first-order running alphaS
    virtual double genQ2(double q2old, double sAB, double zMin, double zMax, 
                         double colFac, double PDFratio, 
                         double b0, double kR, double Lambda,
                         double eA, double eB, double headroomFactor=1.0) {
      return TrialIISplitA::genQ2(q2old, sAB, zMin, zMax, colFac, PDFratio,
                                  b0, kR, Lambda, eB, eA, headroomFactor);
    }

    // Generate a new Q value, with constant trial alphaS
    virtual double genQ2(double q2old, double sAB, double zMin, double zMax, 
                         double colFac, double alphaSvalue, double PDFratio, 
                         double eA, double eB, double headroomFactor=1.0) {
      return TrialIISplitA::genQ2(q2old, sAB, zMin, zMax, colFac, alphaSvalue,
                                  PDFratio, eB, eA, headroomFactor);
    }

    // Generate a new Q value, use in the very unordered region
    virtual double genQ2withQN(double q2old, double q2hat, double N, double sAB, 
                               double zMin, double zMax, 
                               double colFac, double alphaSvalue, double PDFratio, 
                               double eA, double eB, double headroomFactor=1.0) {
      return TrialIISplitA::genQ2withQN(q2old, q2hat, N, sAB, zMin, zMax, colFac,
                                        alphaSvalue, PDFratio, eB, eA, headroomFactor);
    }

    // Generate a new Q value, with running of the PDFs towards the mass threshold
    virtual double genQ2(double q2old, double sAB, double zMin, double zMax, 
                         double colFac, double alphaSvalue, double PDFratio, 
                         int idA, int idB, double eA, double eB, bool useMpdf,
                         double headroomFactor=1.0) {
      return TrialIISplitA::genQ2(q2old, sAB, zMin, zMax, colFac, alphaSvalue,
                                  PDFratio, idB, idA, eB, eA, useMpdf, headroomFactor);
    }
        
    // Inverse transforms
    // To obtain saj and sjb from Qt2 and zeta
    virtual double getS1j(double Qt2, double zeta, double sAB) {
      return TrialIISplitA::getSj2(Qt2, zeta, sAB);
    }
    virtual double getSj2(double Qt2, double zeta, double sAB) {
      return TrialIISplitA::getS1j(Qt2, zeta, sAB);
    }
    
    // Trial PDF ratio
    virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                 int iSys, bool Rfunity,
                                 int idA, int idB, double eA, double eB, 
                                 double Qt2A, double Qt2B, bool useSaved = false) {
      return TrialIISplitA::trialPDFratio(beamBPtr, beamAPtr, iSys, Rfunity,
                                          idB, idA, eB, eA, 
                                          Qt2B, Qt2A, useSaved);
    }
    
    // Physical PDF ratio
    virtual double physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                int iSys, bool Rfunity,
                                int idA, int idB, double eA, double eB,
                                double ea, double eb, double Qt2A, double Qt2B) {
      return TrialIISplitA::physPDFratio(beamBPtr, beamAPtr, iSys, Rfunity,
                                         idB, idA, eB, eA, eb, ea, Qt2B, Qt2A);
    }

  };


  //----------------------------------------------------------------------
  // A conversion trial function for II (derived base class), g<-qqbar
  //----------------------------------------------------------------------
  
  class TrialIIConvA : public TrialGeneratorISR {

  public: 

   // Name of trial generator
    virtual string name() {
      return "TrialIIConvA";
    }
    // For statistics
    virtual int statInt() { return 4; }

    // Trial antenna function
    virtual double aTrial(double saj, double sjb, double sAB);
    
    // Modified trial antenna function, used when combining accept
    // probabilities for gluon emission
    virtual double aTrialMod(double, double, double, double, double, double, double) { 
      printf("%s::aTrialMod Not existing!\n", name().c_str()); 
      return 0.0; 
    }

    // Physical antenna function
    virtual double aPhys(double saj, double sjb, double sAB, int idA, int idB);
    
    // Evolution scale
    virtual double getQ2(double saj, double sjb, double sAB) { 
      return ( (useMevolSav) ? (evNormMconvISav*saj) 
                             : (evNormIIpT2Sav*saj*sjb/(saj+sjb+sAB)) );
    }
    virtual double getQ2max(double sAB, double, double) { 
      return ( (useMevolSav) ? (evNormMconvISav*(shhSav-sAB)) 
                             : (evNormIIpT2Sav*0.25*pow2(shhSav-sAB)/shhSav) );
    }

    // Generate a new Q value, with first-order running alphaS
    virtual double genQ2(double q2old, double sAB, double zMin, double zMax, 
                         double colFac, double PDFratio, 
                         double b0, double kR, double Lambda,
                         double eA, double eB, double headroomFactor=1.0);

    // Generate a new Q value, with constant trial alphaS
    virtual double genQ2(double q2old, double sAB, double zMin, double zMax, 
                         double colFac, double alphaSvalue, double PDFratio, 
                         double eA, double eB, double headroomFactor=1.0);

    // Generate a new Q value, use in the very unordered region
    virtual double genQ2withQN(double q2old, double q2hat, double N, double sAB, 
                               double zMin, double zMax, 
                               double colFac, double alphaSvalue, double PDFratio, 
                               double eA, double eB, double headroomFactor=1.0);

    // Generate a new zeta value in [zMin,zMax]
    virtual double genZ(double zMin, double zMax);
        
    // The zeta integral
    virtual double getIz(double zMin, double zMax);

    // The zeta boundaries, for a given value of the evolution scale
    virtual double getZmin(double Qt2, double sAB, double eA, double eAused);
    virtual double getZmax(double Qt2, double sAB, double eA, double eAused);
        
    // Inverse transforms
    // To obtain saj and sjb from Qt2 and zeta
    virtual double getS1j(double Qt2, double zeta, double sAB);
    virtual double getSj2(double Qt2, double zeta, double sAB);    
    
    // Trial PDF ratio
    virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                 int iSys, bool Rfunity,
                                 int idA, int idB, double eA, double eB, 
                                 double Qt2A, double Qt2B, bool useSaved = false);
    
    // Physical PDF ratio
    virtual double physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                int iSys, bool Rfunity,
                                int idA, int idB, double eA, double eB,
                                double ea, double eb, double Qt2A, double Qt2B);

  };


  //----------------------------------------------------------------------
  // B conversion trial function for II (derived base class), g<-qqbar
  //----------------------------------------------------------------------
  
  class TrialIIConvB : public TrialIIConvA {

  public: 

    // Name of trial generator
    virtual string name() {
        return "TrialIIConvB";
    }

    // Trial antenna function
    virtual double aTrial(double saj, double sjb, double sAB) {
      return TrialIIConvA::aTrial(sjb, saj, sAB);
    }
    
    // Modified trial antenna function, used when combining accept
    // probabilities for gluon emission
    virtual double aTrialMod(double, double, double, double, double, double, double) { 
      printf("%s::aTrialMod Not existing!\n", name().c_str()); 
      return 0.0; 
    }

    // Physical antenna function
    virtual double aPhys(double saj, double sjb, double sAB, int idA, int idB) {
      return TrialIIConvA::aPhys(sjb, saj, sAB, idB, idA);
    }

    // Evolution scale
    virtual double getQ2(double saj, double sjb, double sAB) {
      return TrialIIConvA::getQ2(sjb, saj, sAB);
    }

    // Generate a new Q value, with first-order running alphaS
    virtual double genQ2(double q2old, double sAB, double zMin, double zMax, 
                         double colFac, double PDFratio, 
                         double b0, double kR, double Lambda,
                         double eA, double eB, double headroomFactor=1.0) {
      return TrialIIConvA::genQ2(q2old, sAB, zMin, zMax, colFac, PDFratio,
                                  b0, kR, Lambda, eB, eA, headroomFactor);
    }

    // Generate a new Q value, with constant trial alphaS
    virtual double genQ2(double q2old, double sAB, double zMin, double zMax, 
                         double colFac, double alphaSvalue, double PDFratio, 
                         double eA, double eB, double headroomFactor=1.0) {
      return TrialIIConvA::genQ2(q2old, sAB, zMin, zMax, colFac, alphaSvalue,
                                  PDFratio, eB, eA, headroomFactor);
    }

    // Generate a new Q value, use in the very unordered region
    virtual double genQ2withQN(double q2old, double q2hat, double N, double sAB, 
                               double zMin, double zMax, 
                               double colFac, double alphaSvalue, double PDFratio, 
                               double eA, double eB, double headroomFactor=1.0) {
      return TrialIIConvA::genQ2withQN(q2old, q2hat, N, sAB, zMin, zMax, colFac,
                                       alphaSvalue, PDFratio, eB, eA, headroomFactor);
    }
        
    // Inverse transforms
    // To obtain saj and sjb from Qt2 and zeta
    virtual double getS1j(double Qt2, double zeta, double sAB) {
      return TrialIIConvA::getSj2(Qt2, zeta, sAB);
    }
    virtual double getSj2(double Qt2, double zeta, double sAB) {
      return TrialIIConvA::getS1j(Qt2, zeta, sAB);
    }
    
    // Trial PDF ratio
    virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                 int iSys, bool Rfunity,
                                 int idA, int idB, double eA, double eB, 
                                 double Qt2A, double Qt2B, bool useSaved = false) {
      return TrialIIConvA::trialPDFratio(beamBPtr, beamAPtr, iSys, Rfunity,
                                         idB, idA, eB, eA, 
                                         Qt2B, Qt2A, useSaved);
    }
    
    // Physical PDF ratio
    virtual double physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                int iSys, bool Rfunity,
                                int idA, int idB, double eA, double eB,
                                double ea, double eb, double Qt2A, double Qt2B) {
      return TrialIIConvA::physPDFratio(beamBPtr, beamAPtr, iSys, Rfunity,
                                        idB, idA, eB, eA, eb, ea, Qt2B, Qt2A);
    }

  };


  //----------------------------------------------------------------------
  // Soft-eikonal trial function for IF (derived base class)
  //----------------------------------------------------------------------

  class TrialIFSoft : public TrialGeneratorISR {

  public: 

    // Name of trial generator
    virtual string name() {
      return "TrialIFSoft";
    }
    // For statistics
    virtual int statInt() { return 5; }

    // Trial antenna function
    virtual double aTrial(double saj, double sjk, double sAK);
    
    // Modified trial antenna function, used when combining accept
    // probabilities for gluon emission
    virtual double aTrialMod(double saj, double sjk, double sAK,
                             double eA, double eK, double ea, double ek);

    // Physical antenna function
    virtual double aPhys(double saj, double sjk, double sAK, int idA, int idK);
  
    // Evolution scale
    virtual double getQ2(double saj, double sjk, double sAK) { 
      return (evNormIFpT2Sav*saj*sjk/(sAK+sjk));
    }
    virtual double getQ2max(double sAK, double eA, double eAused) {
      double eAmax = ( (sqrt(shhSav)/2.0) - (eAused-eA) );
      return (evNormIFpT2Sav*sAK*(eAmax-eA)/eA);
    }

    // Generate a new Q value, with first-order running alphaS
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double PDFratio, 
                         double b0, double kR, double Lambda,
                         double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, with constant trial alphaS
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double alphaSvalue, double PDFratio, 
                         double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, use in the very unordered region
    virtual double genQ2withQN(double q2old, double q2hat, double N, double sAK, 
                               double zMin, double zMax, 
                               double colFac, double alphaSvalue, double PDFratio, 
                               double eA, double eK, double headroomFactor=1.0);

    // Generate a new zeta value in [zMin,zMax]
    virtual double genZ(double zMin, double zMax);
        
    // The zeta integral
    virtual double getIz(double zMin, double zMax);

    // The zeta boundaries, for a given value of the evolution scale
    virtual double getZmin(double Qt2, double sAK, double eA, double eAused);
    virtual double getZmax(double Qt2, double sAK, double eA, double eAused);
        
    // Inverse transforms
    // To obtain saj and sjk from Qt2 and zeta
    virtual double getS1j(double Qt2, double zeta, double sAK);
    virtual double getSj2(double Qt2, double zeta, double sAK);
    
    // Trial PDF ratio
    virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                 int iSys, bool Rfunity,
                                 int idA, int idK, double eA, double eK, 
                                 double Qt2A, double Qt2B, bool useSaved = false);
    
    // Physical PDF ratio
    virtual double physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                int iSys, bool Rfunity,
                                int idA, int idK, double eA, double eK,
                                double ea, double ek, double Qt2A, double Qt2B);

  };


  //----------------------------------------------------------------------
  // A quark collinear trial function for IF (derived base class)
  //----------------------------------------------------------------------
  
  class TrialIFQCollA : public TrialGeneratorISR {

  public: 

    // Name of trial generator
    virtual string name() {
      return "TrialIFQCollA";
    }
    // For statistics
    virtual int statInt() { return 6; }

    // Trial antenna function
    virtual double aTrial(double saj, double sjk, double sAK);
    
    // Modified trial antenna function, used when combining accept
    // probabilities for gluon emission
    virtual double aTrialMod(double saj, double sjk, double sAK,
                             double eA, double eK, double ea, double ek);

    // Physical antenna function
    virtual double aPhys(double saj, double sjk, double sAK, int idA, int idK);
  
    // Evolution scale
    virtual double getQ2(double saj, double sjk, double sAK) { 
      return (evNormIFpT2Sav*saj*sjk/(sAK+sjk));
    }
    virtual double getQ2max(double sAK, double eA, double eAused) {
      double eAmax = ( (sqrt(shhSav)/2.0) - (eAused-eA) );
      return (evNormIFpT2Sav*sAK*(eAmax-eA)/eA);
    }

    // Generate a new Q value, with first-order running alphaS
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double PDFratio, 
                         double b0, double kR, double Lambda,
                         double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, with constant trial alphaS
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double alphaSvalue, double PDFratio, 
                         double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, use in the very unordered region
    virtual double genQ2withQN(double q2old, double q2hat, double N, double sAK, 
                               double zMin, double zMax, 
                               double colFac, double alphaSvalue, double PDFratio, 
                               double eA, double eK, double headroomFactor=1.0);

    // Generate a new zeta value in [zMin,zMax]
    virtual double genZ(double zMin, double zMax);
        
    // The zeta integral
    virtual double getIz(double zMin, double zMax);

    // The zeta boundaries, for a given value of the evolution scale
    virtual double getZmin(double Qt2, double sAK, double eA, double eAused);
    virtual double getZmax(double Qt2, double sAK, double eA, double eAused);
        
    // Inverse transforms
    // To obtain saj and sjk from Qt2 and zeta
    virtual double getS1j(double Qt2, double zeta, double sAK);
    virtual double getSj2(double Qt2, double zeta, double sAK);
    
    // Trial PDF ratio
    virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                 int iSys, bool Rfunity,
                                 int idA, int idK, double eA, double eK, 
                                 double Qt2A, double Qt2B, bool useSaved = false);
    
    // Physical PDF ratio
    virtual double physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                int iSys, bool Rfunity,
                                int idA, int idK, double eA, double eK,
                                double ea, double ek, double Qt2A, double Qt2B);

  };


  //----------------------------------------------------------------------
  // K quark collinear trial function for IF (derived base class)
  //----------------------------------------------------------------------

  class TrialIFQCollK : public TrialGeneratorISR {

  public: 

    // Name of trial generator
    virtual string name() {
      return "TrialIFQCollK";
    }
    // For statistics
    virtual int statInt() { return 7; }

    // Trial antenna function
    virtual double aTrial(double saj, double sjk, double sAK);
    
    // Modified trial antenna function, used when combining accept
    // probabilities for gluon emission
    virtual double aTrialMod(double saj, double sjk, double sAK,
                             double eA, double eK, double ea, double ek);

    // Physical antenna function
    virtual double aPhys(double saj, double sjk, double sAK, int idA, int idK);
  
    // Evolution scale
    virtual double getQ2(double saj, double sjk, double sAK) { 
      return (evNormIFpT2Sav*saj*sjk/(sAK+sjk));
    }
    virtual double getQ2max(double sAK, double eA, double eAused) {
      double eAmax = ( (sqrt(shhSav)/2.0) - (eAused-eA) );
      return (evNormIFpT2Sav*sAK*(eAmax-eA)/eA);
    }
    
    // Generate a new Q value, with first-order running alphaS
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double PDFratio, 
                         double b0, double kR, double Lambda,
                         double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, with constant trial alphaS
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double alphaSvalue, double PDFratio, 
                         double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, use in the very unordered region
    virtual double genQ2withQN(double q2old, double q2hat, double N, double sAK, 
                               double zMin, double zMax, 
                               double colFac, double alphaSvalue, double PDFratio, 
                               double eA, double eK, double headroomFactor=1.0);

    // Generate a new zeta value in [zMin,zMax]
    virtual double genZ(double zMin, double zMax);
        
    // The zeta integral
    virtual double getIz(double zMin, double zMax);

    // The zeta boundaries, for a given value of the evolution scale
    virtual double getZmin(double Qt2, double sAK, double eA, double eAused);
    virtual double getZmax(double Qt2, double sAK, double eA, double eAused);
        
    // Inverse transforms
    // To obtain saj and sjk from Qt2 and zeta
    virtual double getS1j(double Qt2, double zeta, double sAK);
    virtual double getSj2(double Qt2, double zeta, double sAK);
    
    // Trial PDF ratio
    virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                 int iSys, bool Rfunity,
                                 int idA, int idK, double eA, double eK, 
                                 double Qt2A, double Qt2B, bool useSaved = false);
    
    // Physical PDF ratio
    virtual double physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                int iSys, bool Rfunity,
                                int idA, int idK, double eA, double eK,
                                double ea, double ek, double Qt2A, double Qt2B);

  };


  //----------------------------------------------------------------------
  // A1 gluon collinear trial function for IF (derived base class)
  //----------------------------------------------------------------------

  class TrialIFGCollA1 : public TrialGeneratorISR {

  public: 

    // Name of trial generator
    virtual string name() {
      return "TrialIFGCollA1";
    }
    // For statistics
    virtual int statInt() { return 8; }

    // Trial antenna function
    virtual double aTrial(double saj, double sjk, double sAK);
    
    // Modified trial antenna function, used when combining accept
    // probabilities for gluon emission
    virtual double aTrialMod(double saj, double sjk, double sAK,
                             double eA, double eK, double ea, double ek);

    // Physical antenna function
    virtual double aPhys(double saj, double sjk, double sAK, int idA, int idK);
  
    // Evolution scale
    virtual double getQ2(double saj, double sjk, double sAK) { 
      return (evNormIFpT2Sav*saj*sjk/(sAK+sjk));
    }
    virtual double getQ2max(double sAK, double eA, double eAused) {
      double eAmax = ( (sqrt(shhSav)/2.0) - (eAused-eA) );
      return (evNormIFpT2Sav*sAK*(eAmax-eA)/eA);
    }

    // Generate a new Q value, with first-order running alphaS
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double PDFratio, 
                         double b0, double kR, double Lambda,
                         double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, with constant trial alphaS
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double alphaSvalue, double PDFratio, 
                         double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, use in the very unordered region
    virtual double genQ2withQN(double q2old, double q2hat, double N, double sAK, 
                               double zMin, double zMax, 
                               double colFac, double alphaSvalue, double PDFratio, 
                               double eA, double eK, double headroomFactor=1.0);

    // Generate a new zeta value in [zMin,zMax]
    virtual double genZ(double zMin, double zMax);
        
    // The zeta integral
    virtual double getIz(double zMin, double zMax);

    // The zeta boundaries, for a given value of the evolution scale
    virtual double getZmin(double Qt2, double sAK, double eA, double eAused);
    virtual double getZmax(double Qt2, double sAK, double eA, double eAused);
        
    // Inverse transforms
    // To obtain saj and sjk from Qt2 and zeta
    virtual double getS1j(double Qt2, double zeta, double sAK);
    virtual double getSj2(double Qt2, double zeta, double sAK);
    
    // Trial PDF ratio
    virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                 int iSys, bool Rfunity,
                                 int idA, int idK, double eA, double eK, 
                                 double Qt2A, double Qt2B, bool useSaved = false);
    
    // Physical PDF ratio
    virtual double physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                int iSys, bool Rfunity,
                                int idA, int idK, double eA, double eK,
                                double ea, double ek, double Qt2A, double Qt2B);

  };


  //----------------------------------------------------------------------
  // A2 gluon collinear trial function for IF (derived base class)
  //----------------------------------------------------------------------

  class TrialIFGCollA2 : public TrialGeneratorISR {

  public: 

    // Name of trial generator
    virtual string name() {
      return "TrialIFGCollA2";
    }
    // For statistics
    virtual int statInt() { return 9; }

    // Trial antenna function
    virtual double aTrial(double saj, double sjk, double sAK);
    
    // Modified trial antenna function, used when combining accept
    // probabilities for gluon emission
    virtual double aTrialMod(double saj, double sjk, double sAK,
                             double eA, double eK, double ea, double ek);

    // Physical antenna function
    virtual double aPhys(double saj, double sjk, double sAK, int idA, int idK);
  
    // Evolution scale
    virtual double getQ2(double saj, double sjk, double sAK) { 
      return (evNormIFpT2Sav*saj*sjk/(sAK+sjk));
    }
    virtual double getQ2max(double sAK, double eA, double eAused) {
      double eAmax = ( (sqrt(shhSav)/2.0) - (eAused-eA) );
      return (evNormIFpT2Sav*sAK*(eAmax-eA)/eA);
    }

    // Generate a new Q value, with first-order running alphaS
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double PDFratio, 
                         double b0, double kR, double Lambda,
                         double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, with constant trial alphaS
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double alphaSvalue, double PDFratio, 
                         double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, use in the very unordered region
    virtual double genQ2withQN(double q2old, double q2hat, double N, double sAK, 
                               double zMin, double zMax, 
                               double colFac, double alphaSvalue, double PDFratio, 
                               double eA, double eK, double headroomFactor=1.0);

    // Generate a new zeta value in [zMin,zMax]
    virtual double genZ(double zMin, double zMax);
        
    // The zeta integral
    virtual double getIz(double zMin, double zMax);

    // The zeta boundaries, for a given value of the evolution scale
    virtual double getZmin(double Qt2, double sAK, double eA, double eAused);
    virtual double getZmax(double Qt2, double sAK, double eA, double eAused);
        
    // Inverse transforms
    // To obtain saj and sjk from Qt2 and zeta
    virtual double getS1j(double Qt2, double zeta, double sAK);
    virtual double getSj2(double Qt2, double zeta, double sAK);
    
    // Trial PDF ratio
    virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                 int iSys, bool Rfunity,
                                 int idA, int idK, double eA, double eK, 
                                 double Qt2A, double Qt2B, bool useSaved = false);
    
    // Physical PDF ratio
    virtual double physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                int iSys, bool Rfunity,
                                int idA, int idK, double eA, double eK,
                                double ea, double ek, double Qt2A, double Qt2B);

  };


  //----------------------------------------------------------------------
  // K gluon collinear trial function for IF (derived base class)
  //----------------------------------------------------------------------

  class TrialIFGCollK : public TrialGeneratorISR {

  public: 

    // Name of trial generator
    virtual string name() {
      return "TrialIFGCollK";
    }
    // For statistics
    virtual int statInt() { return 10; }

    // Trial antenna function
    virtual double aTrial(double saj, double sjk, double sAK);
    
    // Modified trial antenna function, used when combining accept
    // probabilities for gluon emission
    virtual double aTrialMod(double saj, double sjk, double sAK,
                             double eA, double eK, double ea, double ek);

    // Physical antenna function
    virtual double aPhys(double saj, double sjk, double sAK, int idA, int idK);
  
    // Evolution scale
    virtual double getQ2(double saj, double sjk, double sAK) { 
      return (evNormIFpT2Sav*saj*sjk/(sAK+sjk));
    }
    virtual double getQ2max(double sAK, double eA, double eAused) {
      double eAmax = ( (sqrt(shhSav)/2.0) - (eAused-eA) );
      return (evNormIFpT2Sav*sAK*(eAmax-eA)/eA);
    }

    // Generate a new Q value, with first-order running alphaS
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double PDFratio, 
                         double b0, double kR, double Lambda,
                         double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, with constant trial alphaS
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double alphaSvalue, double PDFratio, 
                         double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, use in the very unordered region
    virtual double genQ2withQN(double q2old, double q2hat, double N, double sAK, 
                               double zMin, double zMax, 
                               double colFac, double alphaSvalue, double PDFratio, 
                               double eA, double eK, double headroomFactor=1.0);

    // Generate a new zeta value in [zMin,zMax]
    virtual double genZ(double zMin, double zMax);
        
    // The zeta integral
    virtual double getIz(double zMin, double zMax);

    // The zeta boundaries, for a given value of the evolution scale
    virtual double getZmin(double Qt2, double sAK, double eA, double eAused);
    virtual double getZmax(double Qt2, double sAK, double eA, double eAused);
        
    // Inverse transforms
    // To obtain saj and sjk from Qt2 and zeta
    virtual double getS1j(double Qt2, double zeta, double sAK);
    virtual double getSj2(double Qt2, double zeta, double sAK);
    
    // Trial PDF ratio
    virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                 int iSys, bool Rfunity,
                                 int idA, int idK, double eA, double eK, 
                                 double Qt2A, double Qt2B, bool useSaved = false);
    
    // Physical PDF ratio
    virtual double physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                int iSys, bool Rfunity,
                                int idA, int idK, double eA, double eK,
                                double ea, double ek, double Qt2A, double Qt2B);

  };


  //----------------------------------------------------------------------
  // A splitting trial function for IF (derived base class), q<-gqbar
  //----------------------------------------------------------------------

  class TrialIFSplitA : public TrialGeneratorISR {

  public: 

    // Name of trial generator
    virtual string name() {
      return "TrialIFSplitA";
    }
    // For statistics
    virtual int statInt() { return 11; }

    // Trial antenna function
    virtual double aTrial(double saj, double sjk, double sAK);
    
    // Modified trial antenna function, used when combining accept
    // probabilities for gluon emission
    virtual double aTrialMod(double, double, double, double, double, double, double) { 
      printf("%s::aTrialMod Not existing!\n", name().c_str()); 
      return 0.0; 
    }

    // Physical antenna function
    virtual double aPhys(double saj, double sjk, double sAK, int idA, int idK);
  
    // Evolution scale
    virtual double getQ2(double saj, double sjk, double sAK) { 
      return ( (useMevolSav) ? (evNormMsplitISav*saj) : 
                               (evNormIFpT2Sav*saj*sjk/(sAK+sjk)) );
    }
    virtual double getQ2max(double sAK, double eA, double eAused) {
      double xA    = eA/(sqrt(shhSav)/2.0);
      double eAmax = ( (sqrt(shhSav)/2.0) - (eAused-eA) );
      return ( (useMevolSav) ? (evNormMsplitISav*sAK/xA) :
                               (evNormIFpT2Sav*sAK*(eAmax-eA)/eA) );
    }

    // Generate a new Q value, with first-order running alphaS
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double PDFratio, 
                         double b0, double kR, double Lambda,
                         double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, with constant trial alphaS
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double alphaSvalue, double PDFratio, 
                         double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, use in the very unordered region
    virtual double genQ2withQN(double q2old, double q2hat, double N, double sAK, 
                               double zMin, double zMax, 
                               double colFac, double alphaSvalue, double PDFratio, 
                               double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, with running of the PDFs towards the mass threshold
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double alphaSvalue, double PDFratio, 
                         int idA, int idK, double eA, double eK, bool useMpdf,
                         double headroomFactor=1.0);

    // Generate a new zeta value in [zMin,zMax]
    virtual double genZ(double zMin, double zMax);
        
    // The zeta integral
    virtual double getIz(double zMin, double zMax);

    // The zeta boundaries, for a given value of the evolution scale
    virtual double getZmin(double Qt2, double sAK, double eA, double eAused);
    virtual double getZmax(double Qt2, double sAK, double eA, double eAused);
        
    // Inverse transforms
    // To obtain saj and sjk from Qt2 and zeta
    virtual double getS1j(double Qt2, double zeta, double sAK);
    virtual double getSj2(double Qt2, double zeta, double sAK);
    
    // Trial PDF ratio
    virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                 int iSys, bool Rfunity,
                                 int idA, int idK, double eA, double eK, 
                                 double Qt2A, double Qt2B, bool useSaved = false);
    
    // Physical PDF ratio
    virtual double physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                int iSys, bool Rfunity,
                                int idA, int idK, double eA, double eK,
                                double ea, double ek, double Qt2A, double Qt2B);

  };


  //----------------------------------------------------------------------
  // K splitting trial function for IF (derived base class), g<-qqbar
  //----------------------------------------------------------------------

  class TrialIFSplitK : public TrialGeneratorISR {

  public: 

    // Name of trial generator
    virtual string name() {
      return "TrialIFSplitK";
    }
    // For statistics
    virtual int statInt() { return 12; }

    // Trial antenna function
    virtual double aTrial(double saj, double sjk, double sAK);
    
    // Modified trial antenna function, used when combining accept
    // probabilities for gluon emission
    virtual double aTrialMod(double, double, double, double, double, double, double) { 
      printf("%s::aTrialMod Not existing!\n", name().c_str()); 
      return 0.0; 
    }

    // Physical antenna function
    virtual double aPhys(double saj, double sjk, double sAK, int idA, int idK);
  
    // Evolution scale
    virtual double getQ2(double saj, double sjk, double sAK) { 
      return ( (useMevolSav) ? (evNormMsplitFSav*sjk) : 
                               (evNormIFpT2Sav*saj*sjk/(sAK+sjk)) );
    }
    virtual double getQ2max(double sAK, double eA, double eAused) {
      double xA    = eA/(sqrt(shhSav)/2.0);
      double eAmax = ( (sqrt(shhSav)/2.0) - (eAused-eA) );
      return ( (useMevolSav) ? (evNormMsplitFSav*sAK*(1.0-xA)/xA) :
                               (evNormIFpT2Sav*sAK*(eAmax-eA)/eA) );
    }

    // Generate a new Q value, with first-order running alphaS
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double PDFratio, 
                         double b0, double kR, double Lambda,
                         double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, with constant trial alphaS
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double alphaSvalue, double PDFratio, 
                         double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, use in the very unordered region
    virtual double genQ2withQN(double q2old, double q2hat, double N, double sAK, 
                               double zMin, double zMax, 
                               double colFac, double alphaSvalue, double PDFratio, 
                               double eA, double eK, double headroomFactor=1.0);

    // Generate a new zeta value in [zMin,zMax]
    virtual double genZ(double zMin, double zMax);
        
    // The zeta integral
    virtual double getIz(double zMin, double zMax);

    // The zeta boundaries, for a given value of the evolution scale
    virtual double getZmin(double Qt2, double sAK, double eA, double eAused);
    virtual double getZmax(double Qt2, double sAK, double eA, double eAused);
        
    // Inverse transforms
    // To obtain saj and sjk from Qt2 and zeta
    virtual double getS1j(double Qt2, double zeta, double sAK);
    virtual double getSj2(double Qt2, double zeta, double sAK);
    
    // Trial PDF ratio
    virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                 int iSys, bool Rfunity,
                                 int idA, int idK, double eA, double eK, 
                                 double Qt2A, double Qt2B, bool useSaved = false);
    
    // Physical PDF ratio
    virtual double physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                int iSys, bool Rfunity,
                                int idA, int idK, double eA, double eK,
                                double ea, double ek, double Qt2A, double Qt2B);

  };


  //----------------------------------------------------------------------
  // A conversion trial function for IF (derived base class), g<-qqbar
  //----------------------------------------------------------------------

  class TrialIFConvA : public TrialGeneratorISR {

  public: 

    // Name of trial generator
    virtual string name() {
      return "TrialIFConvA";
    }
    // For statistics
    virtual int statInt() { return 13; }

    // Trial antenna function
    virtual double aTrial(double saj, double sjk, double sAK);
    
    // Modified trial antenna function, used when combining accept
    // probabilities for gluon emission
    virtual double aTrialMod(double, double, double, double, double, double, double) { 
      printf("%s::aTrialMod Not existing!\n", name().c_str()); 
      return 0.0; 
    }

    // Physical antenna function
    virtual double aPhys(double saj, double sjk, double sAK, int idA, int idK);
  
    // Evolution scale
    virtual double getQ2(double saj, double sjk, double sAK) { 
      return ( (useMevolSav) ? (evNormMconvISav*saj) : 
                               (evNormIFpT2Sav*saj*sjk/(sAK+sjk)) );
    }
    virtual double getQ2max(double sAK, double eA, double eAused) {
      double xA    = eA/(sqrt(shhSav)/2.0);
      double eAmax = ( (sqrt(shhSav)/2.0) - (eAused-eA) );
      return ( (useMevolSav) ? (evNormMconvISav*sAK/xA) :
                               (evNormIFpT2Sav*sAK*(eAmax-eA)/eA) );
    }

    // Generate a new Q value, with first-order running alphaS
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double PDFratio, 
                         double b0, double kR, double Lambda,
                         double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, with constant trial alphaS
    virtual double genQ2(double q2old, double sAK, double zMin, double zMax, 
                         double colFac, double alphaSvalue, double PDFratio, 
                         double eA, double eK, double headroomFactor=1.0);

    // Generate a new Q value, use in the very unordered region
    virtual double genQ2withQN(double q2old, double q2hat, double N, double sAK, 
                               double zMin, double zMax, 
                               double colFac, double alphaSvalue, double PDFratio, 
                               double eA, double eK, double headroomFactor=1.0);

    // Generate a new zeta value in [zMin,zMax]
    virtual double genZ(double zMin, double zMax);
        
    // The zeta integral
    virtual double getIz(double zMin, double zMax);

    // The zeta boundaries, for a given value of the evolution scale
    virtual double getZmin(double Qt2, double sAK, double eA, double eAused);
    virtual double getZmax(double Qt2, double sAK, double eA, double eAused);
        
    // Inverse transforms
    // To obtain saj and sjk from Qt2 and zeta
    virtual double getS1j(double Qt2, double zeta, double sAK);
    virtual double getSj2(double Qt2, double zeta, double sAK);
    
    // Trial PDF ratio
    virtual double trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                 int iSys, bool Rfunity,
                                 int idA, int idK, double eA, double eK, 
                                 double Qt2A, double Qt2B,  bool useSaved = false);
    
    // Physical PDF ratio
    virtual double physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                int iSys, bool Rfunity,
                                int idA, int idK, double eA, double eK,
                                double ea, double ek, double Qt2A, double Qt2B);

  };

} // End namespace Vincia

#endif
