// Merging.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2016 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Merging function definitions for the TimeAntShower and SpaceAntShower 
// classes, added by NF.
// NOTE Only for Vincia:orderingMode = 1
//               Vincia:matchingLO = 0
//               Vincia:hardJets = off
//               Vincia:hardJetsQCD = off
//               Vincia:GluonSplittingCorrection = off
//      Set so in the ISR/FSR shower as well
// NOTE Assuming FSR pTnormalization = 4, mDnormalization = 2, all other = 1,
//      as defaults set in Vincia.h
// TODO shower starting scales correct?
// TODO for II we change both incoming guys and the FS, does that work properly?
// TODO Is the factor 1/2 in gluon emission correct?
// TODO why does Z sometimes disappear from event record?

// Include the Vincia headers 
#include "Vincia/TimeAntShower.h"
#include "Vincia/SpaceAntShower.h"

namespace Vincia {
  
  const bool PRINTNOW         = false;
  const bool PRINTDETAILSNOW  = false;
  
  //**********************************************************************
  // TimeAntShower
  //**********************************************************************
  
  // Return clustering kinematics - as needed for merging.
  Event TimeAntShower::clustered( const Event& eventOld, int iRa, 
                                  int iEm, int iRe, string name ) {
    if (!isTimelike(eventOld, iRa, iEm, iRe, name)) return Event();
    string vincia("Vincia:");
    if (name.find(vincia) == string::npos) return Event();
    
    Event event = eventOld;
    
    Particle radiator = event[iRa];
    Particle emission = event[iEm];
    Particle recoiler = event[iRe];
    
    // Clustering: kinematic part
    vector<int> idsVc;
    vector<Vec4> pVc;
    // Gluon emission or quark from gluon splitting (ij)k
    if (emission.id() == 21 || emission.id() > 0) {
      idsVc.push_back(radiator.id());
      idsVc.push_back(emission.id());
      idsVc.push_back(recoiler.id());
      pVc.push_back(radiator.p());
      pVc.push_back(emission.p());
      pVc.push_back(recoiler.p());
    }
    // Antiquark from gluon splitting i(jk)
    else {
      idsVc.push_back(recoiler.id());
      idsVc.push_back(emission.id());
      idsVc.push_back(radiator.id());
      pVc.push_back(recoiler.p());
      pVc.push_back(emission.p());
      pVc.push_back(radiator.p());
    }
    int ii = 0;
    int ij = 1;
    int ik = 2;
    vector<int> idClu; 
    vector<Vec4> pClu;
    int iant;
    bool swapAB;
    if (!vinCluPtr->getClu(idsVc, pVc, ii, ij, ik, idClu,
                           pClu, iant, swapAB)) return Event();
    
    // Id's and colours
    int idj   = emission.id();
    if (idj == 21) swapAB = true;
    int colj  = emission.col();
    int acolj = emission.acol();
    int idI   = idClu[swapAB ? 0 : 1]; // radiator
    int colI  = radiator.col();
    int acolI = radiator.acol();
    int idK   = idClu[swapAB ? 1 : 0]; // recoiler
    int colK  = recoiler.col();
    int acolK = recoiler.acol();
    // Gluon emission
    if (idj == 21) {
      if (colj == acolK) acolK = colI;
      else if (acolj == colK) colK = acolI;
    }
    // Gluon splitting
    else {
      idI = 21;
      if (colI == 0) colI = colj;
      if (acolI == 0) acolI = acolj;
    }
    
    // Change id, momentum, colour
    // Replace radiator and recoiler in event record
    event[iRa].id(idI);
    event[iRa].p(pClu[swapAB ? 0 : 1]);
    event[iRa].col(colI);
    event[iRa].acol(acolI);
    
    event[iRe].id(idK);
    event[iRe].p(pClu[swapAB ? 1 : 0]);
    event[iRe].col(colK);
    event[iRe].acol(acolK);
      
    // Remove emission from event record
    event.remove(iEm, iEm);
    // Remove emission from mother/daughter lists
    for (int i=0; i<event.size(); i++) {
      // Mother1
      if (event[i].mother1() > iEm) 
        event[i].mother1(event[i].mother1()-1);
      else if (event[i].mother1() == iEm) {
        if (event[i].mother2() == iEm) 
          event[i].mothers(0, 0);
        else if (event[i].mother2() > iEm)
          event[i].mothers(event[i].mother2()-1, 0);
        else
          event[i].mothers(event[i].mother2(), 0);
      }
      // Mother2
      if (event[i].mother2() > iEm) 
        event[i].mother2(event[i].mother2()-1);
      else if (event[i].mother2() == iEm) 
        event[i].mother2(0);
      // Daugther1
      if (event[i].daughter1() > iEm) 
        event[i].daughter1(event[i].daughter1()-1);
      else if (event[i].daughter1() == iEm) {
        if (event[i].daughter2() == iEm)
          event[i].daughters(0, 0);
        else if (event[i].daughter2() > iEm)
          event[i].daughters(iEm, event[i].daughter2()-1);
        else
          event[i].daughters(event[i].daughter2(), 0);
      }
      // Daugther2
      if (event[i].daughter2() > iEm) 
        event[i].daughter2(event[i].daughter2()-1);
      else if (event[i].daughter2() == iEm) {
        if (event[i].daughter1() < iEm-1)
          event[i].daughters(event[i].daughter1(), iEm-1);
        else 
          event[i].daughter2(0);
      }
    }
    
    // Check energy-momentum conservation
    if (verbose >= 3)  {
      // Start with incoming partons = beam daughters
      Vec4 pTot = -(event[event[1].daughter1()].p() + 
                    event[event[2].daughter1()].p());
      double eLab = abs(pTot.e());
      // Sum final state
      for (int i=0; i<event.size(); ++i)
        if (event[i].isFinal()) pTot += event[i].p();
      // Check energy-momentum/charge conservation.
      double epDev = abs( pTot.e()) + abs(pTot.px()) + abs(pTot.py())
                                    + abs(pTot.pz() );
      if (epDev > vinComPtr->epTolErr * eLab) {
        cout << "ERROR in SpaceAntShower::clustered for " << name
             << ": energy-momentum not conserved" << endl;
//         cout << " Unclustered event " << name << ":" << endl;
//         eventOld.list();
//         cout << " Clustered event for " << name << ":" << endl;
//         event.list();
      }
      else if (epDev > vinComPtr->epTolWarn * eLab) {
        cout << "WARNING in SpaceAntShower::clustered for " << name
             << ": energy-momentum not quite conserved" << endl;
      }
    }
    
    if (PRINTNOW) {
      cout << " Unclustered event for " << name << ":" << endl;
      eventOld.list();
      cout << " Clustered event for " << name << ":" << endl;
      event.list();
    }
    
    return event;
  }
  
  // Return the evolution variable.
  // Important note:
  //  - The first element of the return vector *must* be the value of the shower
  //    evolution variable corresponding to the branching defined by the integers.
  //  - The second element of the return vector *must* be the value of the shower
  //    evolution variable from which the shower would restart after the branching.
  //    (For many showers, these two values will be identical.)
  vector<double> TimeAntShower::getStateVariables ( const Event& event, int iRa, 
                                                    int iEm, int iRe, string name) {
    vector<double> evolVals;
    evolVals.resize(2);
    evolVals[0] = 0.0;
    evolVals[1] = 0.0;
    string vincia("Vincia:");
    if (name.find(vincia) == string::npos) return evolVals;
    
    double Q2evol = 0.0;
    // 1. Gluon emission: symmetric, ijk
    if (event[iEm].id() == 21)
      Q2evol = resolutionPtr->Q2E(event[iRa], event[iEm], event[iRe], 
                                  evolutionType);
    // 2. Gluon splitting: sjk
    else if (event[iEm].idAbs() < 7)
      Q2evol = resolutionPtr->Q2E(event[iRe], event[iEm], event[iRa], 7);
    
    double Q2restart = 0.0;
    // TODO Could be have MPI?
    // Strong ordering: Restart at evolution variable
    if (orderingMode == 1) Q2restart = Q2evol;
    
    if (PRINTNOW) {
      cout << " Q2evol    for " << name << " = " << Q2evol << endl;
      cout << " Q2restart for " << name << " = " << Q2restart << endl;
    }
    
    evolVals[0] = Q2evol;
    evolVals[1] = Q2restart;
    return evolVals;
  }
  
  // Check if attempted clustering is handled by timelike shower
  bool TimeAntShower::isTimelike( const Event& event, int iRa, int iEm, 
                                  int iRe, string ) {
    return ( event[iRa].isFinal() && event[iEm].isFinal() 
                                  && event[iRe].isFinal() );
  }
  
  // Return a string identifier of a splitting.
  string TimeAntShower::getSplittingName( const Event& event, int iRa, 
                                          int iEm, int iRe ) { 
    if (!isTimelike(event, iRa, iEm, iRe, "")) return "";
    
    string name = "";
    
    // TODO do I need to check if e.g. Em+Re was gluon?
    // and does it matter e.g. for FS Gsplit which quark has cc to rec?
    map<int,int> indexOfAcol;
    map<int,int> indexOfCol;
    if (event[iRa].col()  != 0) indexOfCol[ event[iRa].col() ] = iRa;
    if (event[iRa].acol() != 0) indexOfAcol[event[iRa].acol()] = iRa;
    if (event[iEm].col()  != 0) indexOfCol[ event[iEm].col() ] = iEm;
    if (event[iEm].acol() != 0) indexOfAcol[event[iEm].acol()] = iEm;
    if (event[iRe].col()  != 0) indexOfCol[ event[iRe].col() ] = iRe;
    if (event[iRe].acol() != 0) indexOfAcol[event[iRe].acol()] = iRe;
    
    // 1. Gluon emission
    // emission with colour connection to both partons
    if (event[iEm].id() == 21) {
      int iPEm1 = indexOfCol[event[iEm].acol()];
      int iPEm2 = indexOfAcol[event[iEm].col()];
      if ( (iPEm1 == iRa && iPEm2 == iRe) || (iPEm1 == iRe && iPEm2 == iRa) ) {
        int idRa = event[iRa].idAbs();
        int idRe = event[iRe].idAbs();
             if ( (idRa  <  7) && (idRe  <  7) ) name = "Vincia:QQemit";
        else if ( (idRa == 21) && (idRe == 21) ) name = "Vincia:GGemit";
        else if ( (idRa == 21) && (idRe  <  7) ) name = "Vincia:QGemit";
        else if ( (idRa  <  7) && (idRe == 21) ) name = "Vincia:QGemit";
      }
    }
    // 2. Gluon splitting
    // emission id = - radiator id, different colour and
    // one of them with colour connection to recoiler
    else if (event[iEm].idAbs() < 7) {
      if (event[iEm].id() == -event[iRa].id()) {
        int iPEm = ( event[iEm].id() > 0 ? indexOfAcol[event[iEm].col()]
                                         : indexOfCol[event[iEm].acol()] );
        int iPRa = ( event[iRa].id() > 0 ? indexOfAcol[event[iRa].col()]
                                         : indexOfCol[event[iRa].acol()] );
        if ( (iPEm != iRa) && (iPEm == iRe || iPRa == iRe) ) {
          int idRe = event[iRe].idAbs();
               if (idRe == 21) name = "Vincia:GGsplit";
          else if (idRe  <  7) name = "Vincia:QGsplit";
        }
      }
    }
    
    if (name != "") cout << " Splitting name = " << name << endl;
    else cout << " Did not find splitting" << endl;
    
    return name;
  }
  
  // Return the splitting probability.
  double TimeAntShower::getSplittingProb( const Event& event, int iRa, 
                                          int iEm, int iRe, string name ) { 
    if (!isTimelike(event, iRa, iEm, iRe, name)) return 0.0;
    string vincia("Vincia:");
    if (name.find(vincia) == string::npos) return 0.0;
      
    double splitProb = 0.0;
    
    // Find the right antenna
    for (int iant = iQQemit; iant <= iGGsplit; iant++) {
      if (name != getAnt(iant)->vinciaName()) continue;
      
      double sik    = (event[iRa].p()+event[iRe].p()).m2Calc();
      double sij    = (event[iRa].p()+event[iEm].p()).m2Calc();
      double sjk    = (event[iEm].p()+event[iRe].p()).m2Calc();
      double sIK    = sik + sij + sjk;
      
      // sij and sjk need to be swapped for QGemit if Ra is gluon
      if (iant == iQGemit && event[iRa].id() == 21) {
        int tmp = sjk;
        sjk = sij;
        sij = tmp;
      }
      
      // col*ant
      double ant    = getAnt(iant)->antennaFunction(sij/sIK,sjk/sIK)/sIK;
      double colFac = getAnt(iant)->getChargeFactor();
    
      // alphaS
      double mu2  = 0.0;
      double kMu2 = 0.0;
      if (event[iEm].id() == 21) {
        mu2  = sij*sjk/sIK;
        kMu2 = aSkMu2;
      } else {
        mu2  = sij;
        kMu2 = aSkMu2Split;
      }
      mu2 = max(mu2min, mu2freeze + mu2*kMu2);
      double alphaSnow = min(alphaSmax, alphaSptr->alphaS(mu2));
      
      // CMW
      if (event[iEm].id() == 21) {
        ant *= 0.5; // will get iRa <-> iRe combi
        if (CMWtype > 0) {
          double eik     = ( CMWtype == 1 ? 2.0*sIK/sij/sjk : 2.0*sik/sij/sjk );
          double aSratio = min(alphaSmax, alphaSCMWptr->alphaS(mu2)) / alphaSnow;
          ant += (eik*(aSratio-1.0));
        }
      }
      
      // TODO
      // Antenna Mass Correction Terms
      // HYPERJET
      
      if (PRINTDETAILSNOW) {
        cout << " alphaS = " << alphaSnow << endl; 
        cout << " col*ant = " << colFac*ant << endl;
      }
      
      splitProb = alphaSnow*colFac*ant;
      break;
    }
    
    if (PRINTNOW) {
      cout << " Splitting probability for " << name << " = "
           << scientific << splitProb << endl;
      cout.precision(3); 
      cout << fixed;
    }
    
    return splitProb;
  }
  
  //**********************************************************************
  // SpaceAntShower
  //**********************************************************************
  
  // Return clustering kinematics - as needed for merging.
  Event SpaceAntShower::clustered( const Event& eventOld, int iRa, 
                                   int iEm, int iRe, string name ) {
    if (!isSpacelike(eventOld, iRa, iEm, iRe, name)) return Event();
    string vincia("Vincia:");
    if (name.find(vincia) == string::npos) return Event();
    
    Event event = eventOld;
    
    Particle radiator = event[iRa];
    Particle emission = event[iEm];
    Particle recoiler = event[iRe];
    
    // Check if II or IF
    bool isII = ( !radiator.isFinal() && !recoiler.isFinal() );
    
    // II - start
    if (isII) {
      
      double sab = (radiator.p()+recoiler.p()).m2Calc();
      double saj = (radiator.p()+emission.p()).m2Calc();
      double sjb = (emission.p()+recoiler.p()).m2Calc();
      double sAB = sab - saj - sjb;
      if (sAB < 0.0) return Event();
      
      // Clustering: kinematic part
      Vec4 pA = radiator.p();
      Vec4 pB = recoiler.p();
      double rescaleFacA = sqrt(sAB/sab*(sAB+saj)/(sAB+sjb));
      double rescaleFacB = sqrt(sAB/sab*(sAB+sjb)/(sAB+saj));
      pA.rescale4(rescaleFacA);
      pB.rescale4(rescaleFacB);
      // Check energys -> is clustering possible?
      if ( (pA.e() < 0.0) || (pA.e() > radiator.e()) ) return Event();
      if ( (pB.e() < 0.0) || (pB.e() > recoiler.e()) ) return Event();
    
      // Id's and colours
      int idA   = radiator.id();
      int colA  = radiator.col();
      int acolA = radiator.acol();
      int idB   = recoiler.id();
      int colB  = recoiler.col();
      int acolB = recoiler.acol();
      int idj   = emission.id();
      int colj  = emission.col();
      int acolj = emission.acol();
      // Gluon emission
      if (idj == 21) {
        if (colj == colB) colB = acolA;
        else if (acolj == acolB) acolB = colA;
      }
      // Initial gluon splitting
      else if (name == "Vincia:QXsplitII") {
        idA = -idj;
        if (idA > 0) acolA = 0;
        else if (idA < 0) colA = 0;
      }
      // Initial gluon conversion
      else if (name == "Vincia:GXconvII") {
        idA = 21;
        if (colA == 0) colA = acolj;
        if (acolA == 0) acolA = colj;
      } else return Event();
      
      // Replace radiator and recoiler in event record
      event[iRa].id(idA);
      event[iRa].p(pA);
      event[iRa].col(colA);
      event[iRa].acol(acolA);
      
      event[iRe].id(idB);
      event[iRe].p(pB);
      event[iRe].col(colB);
      event[iRe].acol(acolB);
      
      // Particle vector for FS/IM recoilers
      // TODO this only works if there is only one system in the event
      vector<Particle> FSrecoilers; 
      FSrecoilers.resize(0);
      vector<Particle> IMrecoilers;
      IMrecoilers.resize(0);
      for (int i=0; i<event.size(); i++) {
        // Final state
        if (event[i].isFinal() && (i != iEm)) FSrecoilers.push_back(event[i]);
        // Intermediate state, if present
        else if (event[i].statusAbs() == 22) IMrecoilers.push_back(event[i]);
      }
      // 1) Boost in center-of-mass frame AB, from E,0,0,pz to E,0,0,0
      Vec4 pSum = pA + pB;
      // 2) Boost from E,0,0,0 to E',px',py',pz' (FSrecoilers before clustering in lab frame)
      Vec4 pRec(0.0, 0.0, 0.0, 0.0);
      for (int i=0; i<(int)FSrecoilers.size(); i++) pRec += FSrecoilers[i].p();
      // Now we perform the boost back
      for (int i=0; i<(int)FSrecoilers.size(); i++) {
        FSrecoilers[i].bstback(pRec);
        FSrecoilers[i].bst(pSum);
      }
      for (int i=0; i<(int)IMrecoilers.size(); i++) {
        IMrecoilers[i].bstback(pRec);
        IMrecoilers[i].bst(pSum);
      }
      
      // Change momentum in event record
      int countIM = 0;
      int countFS = 0;
      for (int i=0; i<event.size(); i++) {
        if (event[i].isFinal() && (i != iEm)) {
          event[i].p(FSrecoilers[countFS].p());
          countFS++;
        }
        if (event[i].statusAbs() == 22) {
          event[i].p(IMrecoilers[countIM].p());
          countIM++;
        }
      }
      
    }
    // II - end
    
    // IF - start
    else {
      
      Particle isa = (radiator.isFinal() ? recoiler : radiator);
      Particle fsk = (radiator.isFinal() ? radiator : recoiler);
      double sak = (isa.p()+fsk.p()).m2Calc();
      double saj = (isa.p()+emission.p()).m2Calc();
      double sjk = (emission.p()+fsk.p()).m2Calc();
      double sAK = sak + saj - sjk;
      if (sAK < 0.0) return Event();
    
      // Clustering: kinematic part
      Vec4 pA = isa.p();
      pA.rescale4(sAK/(sAK+sjk));
      Vec4 pK = pA-isa.p()+fsk.p()+emission.p();
      // Check energys -> is clustering possible?
      if ( (pA.e() < 0.0) || (pK.e() < 0.0) ) return Event();
    
      // Id's and colours
      int idA    = isa.id();
      int colA   = isa.col();
      int acolA  = isa.acol();
      int idK    = fsk.id();
      int colK   = fsk.col();
      int acolK  = fsk.acol();
      int idj    = emission.id();
      int colj   = emission.col();
      int acolj  = emission.acol();
      // Gluon emission
      if (idj == 21) {
        if (colj == acolK) acolK = acolA;
        else if (acolj == colK) colK = colA;
      }
      // Initial gluon splitting
      else if (name == "Vincia:QXsplitIF") {
        idA = -idj;
        if (idA > 0) acolA = 0;
        else if (idA < 0) colA = 0;
      }
      // Initial gluon conversion
      else if (name == "Vincia:GXconvIF") {
        idA = 21;
        if (colA == 0) colA = acolj;
        if (acolA == 0) acolA = colj;
      }
      // Final gluon splitting
      else if (name == "Vincia:XGsplitIF") {
        idK = 21;
        if (colK == 0) colK = colj;
        if (acolK == 0) acolK = acolj;
      } else return Event();
      
      // Replace radiator and recoiler in event record
      if (event[iRa].isFinal()) {
        event[iRe].id(idA);
        event[iRe].p(pA);
        event[iRe].col(colA);
        event[iRe].acol(acolA);
        
        event[iRa].id(idK);
        event[iRa].p(pK);
        event[iRa].col(colK);
        event[iRa].acol(acolK);
      } else {
        event[iRa].id(idA);
        event[iRa].p(pA);
        event[iRa].col(colA);
        event[iRa].acol(acolA);
        
        event[iRe].id(idK);
        event[iRe].p(pK);
        event[iRe].col(colK);
        event[iRe].acol(acolK);
      }
      
    }
    // IF - end
      
    // Remove emission from event record
    event.remove(iEm, iEm);
    // Remove emission from mother/daughter lists
    for (int i=0; i<event.size(); i++) {
      // Mother1
      if (event[i].mother1() > iEm) 
        event[i].mother1(event[i].mother1()-1);
      else if (event[i].mother1() == iEm) {
        if (event[i].mother2() == iEm) 
          event[i].mothers(0, 0);
        else if (event[i].mother2() > iEm)
          event[i].mothers(event[i].mother2()-1, 0);
        else
          event[i].mothers(event[i].mother2(), 0);
      }
      // Mother2
      if (event[i].mother2() > iEm) 
        event[i].mother2(event[i].mother2()-1);
      else if (event[i].mother2() == iEm) 
        event[i].mother2(0);
      // Daugther1
      if (event[i].daughter1() > iEm) 
        event[i].daughter1(event[i].daughter1()-1);
      else if (event[i].daughter1() == iEm) {
        if (event[i].daughter2() == iEm)
          event[i].daughters(0, 0);
        else if (event[i].daughter2() > iEm)
          event[i].daughters(iEm, event[i].daughter2()-1);
        else
          event[i].daughters(event[i].daughter2(), 0);
      }
      // Daugther2
      if (event[i].daughter2() > iEm) 
        event[i].daughter2(event[i].daughter2()-1);
      else if (event[i].daughter2() == iEm) {
        if (event[i].daughter1() < iEm-1)
          event[i].daughters(event[i].daughter1(), iEm-1);
        else 
          event[i].daughter2(0);
      }
    }
    
    // Check energy-momentum conservation
    if (verbose >= 3)  {
      // Start with incoming partons = beam daughters
      Vec4 pTot = -(event[event[1].daughter1()].p() + 
                    event[event[2].daughter1()].p());
      double eLab = abs(pTot.e());
      // Sum final state
      for (int i=0; i<event.size(); ++i)
        if (event[i].isFinal()) pTot += event[i].p();
      // Check energy-momentum/charge conservation.
      double epDev = abs( pTot.e()) + abs(pTot.px()) + abs(pTot.py())
                                    + abs(pTot.pz() );
      if (epDev > vinComPtr->epTolErr * eLab) {
        cout << "ERROR in SpaceAntShower::clustered for " << name
             << ": energy-momentum not conserved" << endl;
//         cout << " Unclustered event " << name << ":" << endl;
//         eventOld.list();
//         cout << " Clustered event for " << name << ":" << endl;
//         event.list();
      }
      else if (epDev > vinComPtr->epTolWarn * eLab) {
        cout << "WARNING in SpaceAntShower::clustered for " << name
             << ": energy-momentum not quite conserved" << endl;
      }
    }
    
    if (PRINTNOW) {
      cout << " Unclustered event for " << name << ":" << endl;
      eventOld.list();
      cout << " Clustered event for " << name << ":" << endl;
      event.list();
    }
    
    return event;
  }
  
  // Return the evolution variable.
  // Important note:
  //  - The first element of the return vector *must* be the value of the shower
  //    evolution variable corresponding to the branching defined by the integers.
  //  - The second element of the return vector *must* be the value of the shower
  //    evolution variable from which the shower would restart after the branching.
  //    (For many showers, these two values will be identical.)
  vector<double> SpaceAntShower::getStateVariables ( const Event& event, int iRa, 
                                                     int iEm, int iRe, string name) {
    vector<double> evolVals;
    evolVals.resize(2);
    evolVals[0] = 0.0;
    evolVals[1] = 0.0;
    string vincia("Vincia:");
    if (name.find(vincia) == string::npos) return evolVals;
    
    // Erase the Vincia: and II/IF part to minimize string checks
    name.erase(name.begin(), name.begin()+vincia.size());
    name.erase(name.end()-2, name.end());
    
    // Check if II or IF
    bool isII = ( !event[iRa].isFinal() && !event[iRe].isFinal() );
    
    double Q2evol = 0.0;
    // 1. Gluon emission: abj or akj
    if (event[iEm].id() == 21) {
      // 1.1 from II
      if (isII)
        Q2evol = resolutionPtr->Q2EII(event[iRa], event[iRe], event[iEm], 1);
      // 1.2 from IF side1
      else if (!event[iRa].isFinal())
        Q2evol = resolutionPtr->Q2EIF(event[iRa], event[iRe], event[iEm], 1);
      // 1.2 from IF side2
      else if (!event[iRe].isFinal())
        Q2evol = resolutionPtr->Q2EIF(event[iRe], event[iRa], event[iEm], 1);
    }
    // Non-emission
    else if (event[iEm].idAbs() < 7) {
      // 2. Initial gluon splitting: saj
      if (name == "QXsplit") {
        // 2.1 from II
        if (isII) Q2evol = resolutionPtr->Q2EII(event[iRa], event[iRe], event[iEm], 
                                                !evAllPT ? 2 : 1);
        // 2.2 form IF
        else      Q2evol = resolutionPtr->Q2EIF(event[iRa], event[iRe], event[iEm], 
                                                !evAllPT ? 2 : 1);
      }
      // 3. Gluon conversion: saj
      else if (name == "GXconv") {
        // 3.1 from II
        if (isII) Q2evol = resolutionPtr->Q2EII(event[iRa], event[iRe], event[iEm], 
                                                !evAllPT ? 2 : 1);
        // 3.2 form IF
        else      Q2evol = resolutionPtr->Q2EIF(event[iRa], event[iRe], event[iEm], 
                                                !evAllPT ? 3 : 1);
      }
      // 4. Fisal gluon splitting: sjk
      else if (name == "XGsplit") {
        Q2evol = resolutionPtr->Q2EIF(event[iRe], event[iRa], event[iEm], 
                                      !evAllPT ? 4 : 1);
      }
    }
    
    double Q2restart = 0.0;
    // TODO Could be have MPI?
    // Strong ordering: Restart at evolution variable
    if (orderingMode == 1) Q2restart = Q2evol;
    
    if (PRINTNOW) {
      cout << " Q2evol    for " << name << " = " << Q2evol << endl;
      cout << " Q2restart for " << name << " = " << Q2restart << endl;
    }
    
    evolVals[0] = Q2evol;
    evolVals[1] = Q2restart;
    return evolVals;
  }
  
  // Check if attempted clustering is handled by spacelike shower
  bool SpaceAntShower::isSpacelike( const Event& event, int iRa, int iEm, 
                                    int iRe, string ) {
    return !( event[iRa].isFinal() && event[iEm].isFinal() 
                                   && event[iRe].isFinal() );
  }
  
  // Return a string identifier of a splitting.
  string SpaceAntShower::getSplittingName( const Event& event, int iRa, 
                                           int iEm, int iRe ) { 
    if (!isSpacelike(event, iRa, iEm, iRe, "")) return "";
    
    string name = "";
    
    bool iRaIS = !event[iRa].isFinal();
    bool iReIS = !event[iRe].isFinal();
    bool isII  = (iRaIS && iReIS);
    int  idRa  = event[iRa].idAbs();
    bool iRaAQ = (event[iRa].id() < 0);
    int  idEm  = event[iEm].idAbs();
    bool iEmAQ = (event[iEm].id() < 0);
    int  idRe  = event[iRe].idAbs();
    
    // TODO do I need to check if e.g. Em+Re was gluon?
    // and does it matter e.g. for FS Gsplit which quark has cc to rec?
    map<int,int> indexOfAcol;
    map<int,int> indexOfCol;
    if (iRaIS) {
      if (event[iRa].acol() != 0) indexOfCol[ event[iRa].acol()] = iRa;
      if (event[iRa].col()  != 0) indexOfAcol[event[iRa].col() ] = iRa;
    } else {
      if (event[iRa].col()  != 0) indexOfCol[ event[iRa].col() ] = iRa;
      if (event[iRa].acol() != 0) indexOfAcol[event[iRa].acol()] = iRa;
    }
    if (event[iEm].col()  != 0) indexOfCol[ event[iEm].col() ] = iEm;
    if (event[iEm].acol() != 0) indexOfAcol[event[iEm].acol()] = iEm;
    if (iReIS) {
      if (event[iRe].acol() != 0) indexOfCol[ event[iRe].acol()] = iRe;
      if (event[iRe].col()  != 0) indexOfAcol[event[iRe].col() ] = iRe;
    } else {
      if (event[iRe].col()  != 0) indexOfCol[ event[iRe].col() ] = iRe;
      if (event[iRe].acol() != 0) indexOfAcol[event[iRe].acol()] = iRe;
    }
    
    // 1. Gluon emission
    // emission with colour connection to both partons
    if (idEm == 21) {
      int iPEm1 = indexOfCol[event[iEm].acol()];
      int iPEm2 = indexOfAcol[event[iEm].col()];
      if ((iPEm1 == iRa && iPEm2 == iRe) || (iPEm1 == iRe && iPEm2 == iRa)) {
        // 1.1 from II
        if (isII) {
               if ( (idRa  <  7) && (idRe  <  7) ) name = "Vincia:QQemitII";
          else if ( (idRa == 21) && (idRe == 21) ) name = "Vincia:GGemitII";
          else if ( (idRa == 21) && (idRe  <  7) ) name = "Vincia:GQemitII";
          else if ( (idRa  <  7) && (idRe == 21) ) name = "Vincia:GQemitII";
        }
        // 1.2 from IF
        else if (iRaIS || iReIS) {
          int idIS = (!iReIS ? idRa : idRe);
          int idFS = (!iReIS ? idRe : idRa);
               if ( (idIS  <  7) && (idFS  <  7) ) name = "Vincia:QQemitIF";
          else if ( (idIS == 21) && (idFS == 21) ) name = "Vincia:GGemitIF";
          else if ( (idIS == 21) && (idFS  <  7) ) name = "Vincia:GQemitIF";
          else if ( (idIS  <  7) && (idFS == 21) ) name = "Vincia:QGemitIF";
        }
      }
    }
    // Non-emission
    else if (idEm < 7) {
      // 2. Initial gluon splitting
      // emission with colour connection to initial gluon = radition, 
      // radiator with colour connection to recoiler
      if ( (idRa == 21) && iRaIS ) {
        int iPRa1 = indexOfAcol[event[iRa].acol()];
        int iPRa2 = indexOfCol[event[iRa].col()];
        if ( (iPRa1 == iEm && iPRa2 == iRe) || (iPRa2 == iEm && iPRa1 == iRe) ) {
          // 2.1 from II
          if (isII) name = "Vincia:QXsplitII";
          // 2.2 form IF
          else      name = "Vincia:QXsplitIF";
        }
      }
      // 3. Gluon conversion
      // initial quark = radiator with same id, different colour and one 
      // of them colour connection to recoiler
      else if ( (event[iEm].id() == event[iRa].id()) && iRaIS ) {
        int iPEm = (iEmAQ ? indexOfCol[event[iEm].acol()] : indexOfAcol[event[iEm].col()]);
        int iPRa = (iRaAQ ? indexOfAcol[event[iRa].acol()] : indexOfCol[event[iRa].col()]);
        if ( (iPEm != iRa) && (iPEm == iRe || iPRa == iRe) ) {
          // 2.1 from II
          if (isII) name = "Vincia:GXconvII";
          // 2.2 form IF
          else      name = "Vincia:GXconvIF";
        }
      }
      // 4. Fisal gluon splitting
      // fisal quark == radiator with -id, different colour and
      // one of them with colour connection to recoiler
      else if ( (event[iEm].id() == -event[iRa].id()) && !iRaIS && iReIS ) {
        int iPEm = (iEmAQ ? indexOfCol[event[iEm].acol()] : indexOfAcol[event[iEm].col()]);
        int iPRa = (iRaAQ ? indexOfCol[event[iRa].acol()] : indexOfAcol[event[iRa].col()]);
        if ( (iPEm != iRa) && (iPEm == iRe || iPRa == iRe) ) 
          name = "Vincia:XGsplitIF";
      }
    }
    
    if (PRINTNOW) {
      if (name != "") cout << " Splitting name = " << name << endl;
      else cout << " Did not find splitting" << endl;
    }
    
    return name;
  }
  
  // Return the splitting probability.
  double SpaceAntShower::getSplittingProb( const Event& event, int iRa, 
                                           int iEm, int iRe, string name ) { 
    if (!isSpacelike(event, iRa, iEm, iRe, name)) return 0.0;
    string vincia("Vincia:");
    if (name.find(vincia) == string::npos) return 0.0;
    
    double splitProb = 0.0;
    
    bool isII = ( !event[iRa].isFinal() && !event[iRe].isFinal() );
    
    if (isII) { for (int iant = iQQemitII; iant <= iGXconvII; iant++) {
      // Find the right antenna
      if (name != getAnt(iant)->vinciaName()) continue;
      
      // Swap iRa and iRe for GQemitII if Re is gluon
      if (iant == iGQemitII && event[iRe].id() == 21) {
        int tmp = iRa;
        iRa = iRe;
        iRe = tmp;
      }
      
      double sab = (event[iRa].p()+event[iRe].p()).m2Calc();
      double saj = (event[iRa].p()+event[iEm].p()).m2Calc();
      double sjb = (event[iEm].p()+event[iRe].p()).m2Calc();
      double sAB = (sab - saj - sjb);
      if (sAB < 0.0) return splitProb;
      
      // col*ant
      double ant    = getAnt(iant)->antennaFunction(saj,sjb,sAB);
      double colFac = getAnt(iant)->getChargeFactor();
      
      // alphaS
      double Q2e  = getStateVariables(event, iRa, iEm, iRe, name)[0];
      double kMu2 = 0.0;
      if (event[iEm].id() == 21)   kMu2 = aSkMu2;
      else if (iant == iQXsplitII) kMu2 = aSkMu2SplitI;
      else if (iant == iGXconvII)  kMu2 = aSkMu2Conv;
      double mu2       = max(mu2min, mu2freeze + Q2e*kMu2);
      double alphaSnow = min(alphaSmax, alphaSptr->alphaS(mu2));
      
      // CMW
      if (event[iEm].id() == 21) {
        ant *= 0.5; // will get iRa <-> iRe combi
        if (CMWtypeII > 0) {
          double eik     = ( CMWtypeII == 1 ? 2.0*sAB/saj/sjb : 2.0*sab/saj/sjb );
          double aSratio = min(alphaSmax, alphaSCMWptr->alphaS(mu2)) / alphaSnow;
          ant += (eik*(aSratio-1.0));
        }
      }
      
      // PDF ratio
      double PDFratio = 0.0;
      int iSystem     = 0;
      double TINYPDF  = pow(10,-10);
      Particle ISa    = event[iRa];
      Particle ISb    = event[iRe];
      double xa       = ISa.e()/(sqrt(shhSav)/2.0);
      double xb       = ISb.e()/(sqrt(shhSav)/2.0);
      double xA       = xa*sqrt(sAB/sab*(sAB+saj)/(sAB+sjb));
      double xB       = xb*sqrt(sAB/sab*(sAB+sjb)/(sAB+saj));
      int ida         = ( ISa.id() == 21 ? 0 : ISa.id() );
      int idb         = ( ISb.id() == 21 ? 0 : ISb.id() );
      int idA         = ida;
      if (iant == iQXsplitII) idA = -event[iEm].id();
      else if (iant == iGXconvII) idA = 0;
      int idB         = idb;
      BeamParticle* beamAPtrNow = (ISa.pz() > 0.0 ? beamAPtr : beamBPtr);
      BeamParticle* beamBPtrNow = (ISa.pz() > 0.0 ? beamBPtr : beamAPtr);
      double topA     = beamAPtrNow->xfISR(iSystem, ida, xa, Q2e);
      double bottomA  = beamAPtrNow->xfISR(iSystem, idA, xA, Q2e);
      double topB     = beamBPtrNow->xfISR(iSystem, idb, xb, Q2e);
      double bottomB  = beamBPtrNow->xfISR(iSystem, idB, xB, Q2e);
      if (abs(topA)    < TINYPDF) topA    = TINYPDF;
      if (abs(bottomA) < TINYPDF) bottomA = TINYPDF;
      if (abs(topB)    < TINYPDF) topB    = TINYPDF;
      if (abs(bottomB) < TINYPDF) bottomB = TINYPDF;
      PDFratio        = topA/bottomA*xA/xa * topB/bottomB*xB/xb;
      
      if (PRINTDETAILSNOW) {
        cout << " alphaS = " << alphaSnow << endl; 
        cout << " col*ant = " << colFac*ant << endl;
        cout << " PDF = " << PDFratio << endl;
      }
      
      splitProb = alphaSnow*colFac*ant*PDFratio;
    } }
    
    else { for (int iant = iQQemitIF; iant <= iXGsplitIF; iant++) {
      // Find the right antenna
      if (name != getAnt(iant)->vinciaName()) continue;
      
      // Swap iRa and iRe if iRe is initial
      if (!event[iRe].isFinal()) {
        int tmp = iRa;
        iRa = iRe;
        iRe = tmp;
      }
      
      double sak = (event[iRa].p()+event[iRe].p()).m2Calc();
      double saj = (event[iRa].p()+event[iEm].p()).m2Calc();
      double sjk = (event[iEm].p()+event[iRe].p()).m2Calc();
      double sAK = (sak + saj - sjk);
      if (sAK < 0.0) return splitProb;
      
      // col*ant
      double ant    = getAnt(iant)->antennaFunction(saj,sjk,sAK);
      double colFac = getAnt(iant)->getChargeFactor();
    
      // alphaS
      double Q2e  = 0.0;
      double kMu2 = 0.0;
      if (event[iEm].id() == 21) {
        Q2e  = saj*sjk/(sAK+sjk);
        kMu2 = aSkMu2;
      } else if (iant == iXGsplitIF) {
        if (!evAllPT) Q2e = sjk;
        else          Q2e = saj*sjk/(sAK+sjk);
        kMu2 = aSkMu2SplitF;
      } else if (iant == iQXsplitIF) {
        if (!evAllPT) Q2e = saj;
        else          Q2e = saj*sjk/(sAK+sjk);
        kMu2 = aSkMu2SplitI;
      } else if (iant == iGXconvIF) {
        if (!evAllPT) Q2e = saj;
        else          Q2e = saj*sjk/(sAK+sjk);
        kMu2 = aSkMu2Conv;
      }
      double mu2       = max(mu2min, mu2freeze + Q2e*kMu2);
      double alphaSnow = min(alphaSmax, alphaSptr->alphaS(mu2));
      
      // CMW
      if (event[iEm].id() == 21) {
        ant *= 0.5; // will get iRa <-> iRe combi
        if (CMWtypeIF > 0) {
          double eik     = ( CMWtypeIF == 1 ? 2.0*sAK/saj/sjk : 2.0*sak/saj/sjk );
          double aSratio = min(alphaSmax, alphaSCMWptr->alphaS(mu2)) / alphaSnow;
          ant += (eik*(aSratio-1.0));
        }
      }
      
      // PDF ratio
      double PDFratio = 0.0;
      int iSystem     = 0;
      double TINYPDF  = pow(10,-10);
      Particle ISa    = event[iRa];
      double xa       = ISa.e()/(sqrt(shhSav)/2.0);
      double xA       = xa*sAK/(sAK+sjk);
      int ida         = ( ISa.id() == 21 ? 0 : ISa.id() );
      int idA         = ida;
      if (iant == iQXsplitIF) idA = -event[iEm].id();
      else if (iant == iGXconvIF) idA = 0;
      BeamParticle* beamPtrNow = (ISa.pz() > 0.0 ? beamAPtr : beamBPtr);
      double top      = beamPtrNow->xfISR(iSystem, ida, xa, Q2e);
      double bottom   = beamPtrNow->xfISR(iSystem, idA, xA, Q2e);
      if (abs(top)    < TINYPDF) top    = TINYPDF;
      if (abs(bottom) < TINYPDF) bottom = TINYPDF;
      PDFratio        = top/bottom*xA/xa;
      
      if (PRINTDETAILSNOW) {
        cout << " alphaS = " << alphaSnow << endl; 
        cout << " col*ant = " << colFac*ant << endl;
        cout << " PDF = " << PDFratio << endl;
      }
      
      splitProb = alphaSnow*colFac*ant*PDFratio;
    } }
    
    if (PRINTNOW) {
      cout << " Splitting probability for " << name << " = "
           << scientific << splitProb << endl;
      cout.precision(3); 
      cout << fixed;
    }
    
    return splitProb;
  }

} // end namespace VINCIA
