// TrialGeneratorsISR.cc is a part of the VINCIA plug-in to the PYTHIA 8 
// generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2016 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for the 
// TrialGeneratorISR base class and derived classes

// Include the Vincia headers 
#include "Vincia/TrialGeneratorsISR.h"

namespace Vincia {
  
  // *****
  // Base class TrialGenerator
  // Note: base class is coded for a soft-eikonal trial function
 
  // Init Pointers
  void TrialGeneratorISR::initPtr(Rndm* rndmPtrIn, Settings* settingsPtrIn,
                                  Info* infoPtrIn) {
    rndmPtr         = rndmPtrIn;
    settingsPtr     = settingsPtrIn;
    infoPtr         = infoPtrIn;
    isInitPtr       = true;
  }
  
  // Initialize
  void TrialGeneratorISR::init(double mcIn, double mbIn) {
    // Check if pointers are initialized
    if ( !isInitPtr ) {
      printOut(name()+"::init()","Error! Pointers not initialized.");
      return;
    }
    TINYPDF          = pow(10,-10);
    TINYPDFtrial     = pow(10,-10);
    // Use m or pT evolution for collinear singularities
    useMevolSav      = !settingsPtr->flag("Vincia:evolveAllInPT");
    // s hadron hadron
    shhSav           = infoPtr->s();
    // For conversion trial generators
    gConvIDSav       = 0;
    // Number of active quark flavours
    nGtoQISRSav      = settingsPtr->mode("Vincia:nGluonToQuarkI");
    gConvIDSav       = 0;
    // Masses
    mbSav            = mbIn;
    mcSav            = mcIn;
    // Counter for PDF evaluation outside of boundaries
    nPDFoutsideBound = 0;
    nReportPDFbound  = 1;
    // Should we use random xa in PDF trial?
    //useRandomX       = true;
    useRandomX       = false;
    trialPDFratioSav = 1.0;
    // Normalization factors
    evNormIIpT2Sav   = settingsPtr->parm("Vincia:pTnormalizationII");
    evNormMsplitISav = settingsPtr->parm("Vincia:mSnormalizationI");
    evNormMconvISav  = settingsPtr->parm("Vincia:mCnormalizationI");
    evNormIFpT2Sav   = settingsPtr->parm("Vincia:pTnormalizationIF");
    evNormMsplitFSav = settingsPtr->parm("Vincia:mSnormalizationF");
    // Exponent for x in the pdf ratios
    alphaSplitSav  = (useMevolSav ? 1.0 : 0.0); // II, conversions, for evol in m
    alphaConvSav   = (useMevolSav ? 1.0 : 1.0);
    betaQCollISav  = 1.0; // IF, emission, always evolved in pT
    betaGColl1ISav = 1.0;
    betaGColl2ISav = 1.0;
    betaSplitISav  = 1.0; // IF, conversions, for evol in pT or m
    betaConvISav   = 1.0;
    // Verbose level
    verbose          = settingsPtr->mode("Vincia:Verbose");
    isInit           = true;
  }

  // -----------------------------------------------------------------
  // Virtual functions 
  // Defining the particular kind of trial generator
  // -----------------------------------------------------------------
    
  // Trial antenna function
  double TrialGeneratorISR::aTrial(double saj, double sjb, double sAB) {
    if (saj < 0. || sjb < 0.) return 0.;
    double sab = sAB+saj+sjb;
    return 2.0*sab*sab/saj/sjb/sAB;
  }
    
  // Modified trial antenna function, used when combining accept
  // probabilities for gluon emission
  double TrialGeneratorISR::aTrialMod(double saj, double sjb, double sAB,
                                      double, double, double, double) {
    if (saj < 0. || sjb < 0.) return 0.;
    double sab = sAB+saj+sjb;
    return 2.0*sab*sab/saj/sjb/sAB;
  }
    
  // Physical antenna function
  double TrialGeneratorISR::aPhys(double saj, double sjb, double sAB, int idA, int idB) {
    double sab = sAB+saj+sjb;
    if ((idA!=21) && (idB!=21)) return (2.0*sab*sAB/saj/sjb + saj/sjb + sjb/saj)/sAB;
    else if ((idA!=21) && (idB=21)) return (2.0*sab*sAB/saj/sjb + sjb/saj)/sAB;
    else if ((idA==21) && (idB!=21)) return (2.0*sab*sAB/saj/sjb + saj/sjb)/sAB;
    else return (2.0*sab*sAB/saj/sjb)/sAB;
    // Trial antenna contains also quark collinear singularity!
  }
  
  // Generate a new Q value, with first-order running alphaS
  double TrialGeneratorISR::genQ2(double q2old, double sAB, double zMin, double zMax, 
                                  double colFac, double PDFratio, 
                                  double b0, double kR, double Lambda,
                                  double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAB < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 2.0*M_PI*b0/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    double facLam = pow2(Lambda/kR) * evNormIIpT2Sav;
    return exp(pow(ran,comFac) * log(q2old/facLam)) * facLam;
  }
  
  // Generate a new Q value, with constant trial alphaS
  double TrialGeneratorISR::genQ2(double q2old, double sAB, double zMin, double zMax, 
                                  double colFac, double alphaSvalue, double PDFratio, 
                                  double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAB < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 2.0*M_PI/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    return q2old * pow(ran,comFac/alphaSvalue);
  }
  
  // Generate a new Q value, use in the very unordered region
  double TrialGeneratorISR::genQ2withQN(double q2old, double q2hat, double N, double sAB, 
                                        double zMin, double zMax, 
                                        double colFac, double alphaSvalue, double PDFratio, 
                                        double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAB < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 2.0*M_PI/Iz/colFac/alphaSvalue/PDFratio/headroomFactor;
    double expo   = ((N-2.0)/2.0);
    double ran    = rndmPtr->flat();
    return pow( pow(q2old,expo)/(1.0-pow(q2old/q2hat,expo)*
                                     log(ran)*comFac*expo), 1.0/expo);
  }
  
  // Generate a new zeta value in [zMin,zMax]
  double TrialGeneratorISR::genZ(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return -1.;
    double ran = rndmPtr->flat();
    double invZ = 1.0 + (1.0-zMin)/zMin*pow(zMin*(1.0-zMax)/zMax/(1.0-zMin),ran);
    return 1.0/invZ;
  }
    
  // The zeta integral
  double TrialGeneratorISR::getIz(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return 0.0;
    return log(zMax*(1.0-zMin)/zMin/(1.0-zMax)); 
  }
    
  // The zeta boundaries, for a given value of the evolution variable
  double TrialGeneratorISR::getZmin(double Qt2, double sAB, double, double) {
    if ( ( (shhSav-sAB)*(shhSav-sAB) - (4.0*Qt2/evNormIIpT2Sav*shhSav) )
         < TINYPDF ) return 0.5*(shhSav-sAB)/shhSav;
    double sajm = 0.5*( shhSav-sAB-sqrt((shhSav-sAB)*(shhSav-sAB)-
                                        (4.0*Qt2/evNormIIpT2Sav*shhSav)) );
    return sajm/shhSav;
  }
  double TrialGeneratorISR::getZmax(double Qt2, double sAB, double, double) {
    if ( ( (shhSav-sAB)*(shhSav-sAB) - (4.0*Qt2/evNormIIpT2Sav*shhSav) ) 
         < TINYPDF ) return 0.5*(shhSav-sAB)/shhSav;
    double sajp = 0.5*( shhSav-sAB+sqrt((shhSav-sAB)*(shhSav-sAB)-
                                        (4.0*Qt2/evNormIIpT2Sav*shhSav)) );
    return sajp/shhSav;
  }
    
  // Inverse transforms
  // To obtain saj and sjb from Qt2 and zeta
  double TrialGeneratorISR::getS1j(double Qt2, double zeta, double sAB) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getSj2(Qt2,-zeta,sAB);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getS1j() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double saj =  Qt2/evNormIIpT2Sav/zeta;
    return saj;
  }
  double TrialGeneratorISR::getSj2(double Qt2, double zeta, double sAB) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getS1j(Qt2,-zeta,sAB);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getSj2() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double sjb = (Qt2/evNormIIpT2Sav+sAB*zeta)/(1.0-zeta);
    return sjb;
  }
    
  // Trial PDF ratio
  double TrialGeneratorISR::trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                          int iSys, bool Rfunity,
                                          int idA, int idB, double eA, double eB, 
                                          double Qt2A, double Qt2B, bool useSaved) {
    if (useSaved) return trialPDFratioSav;
    double xA = eA/(sqrt(shhSav)/2.0);
    double xB = eB/(sqrt(shhSav)/2.0);
    if (idA==21) idA = 0;
    if (idB==21) idB = 0;
    double Rpdf = 1.0;
    if (Rfunity) {
      Rpdf = 1.0;
    } else if (useRandomX) {
      double xa     = xA*pow(0.95/xA,rndmPtr->flat());
      double xb     = xB*pow(0.95/xB,rndmPtr->flat());
      double top    = beamAPtr->xfISR(iSys, idA, xa, Qt2A);
      double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
      if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
      if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
      Rpdf          = top/xa/bottom*xA;
      top           = beamBPtr->xfISR(iSys, idB, xb, Qt2B);
      bottom        = beamBPtr->xfISR(iSys, idB, xB, Qt2B);
      if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
      if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
      Rpdf         *= top/xb/bottom*xB;
      Rpdf          = max(Rpdf,1.0);
    } else {
      Rpdf = 1.0;
    }
    trialPDFratioSav = Rpdf;
    return Rpdf;
  }
  
  // Physical PDF ratio
  double TrialGeneratorISR::physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                         int iSys, bool Rfunity,
                                         int idA, int idB, double eA, double eB, 
                                         double ea, double eb, double Qt2A, double Qt2B) {
    double xA = eA/(sqrt(shhSav)/2.0);
    double xa = ea/(sqrt(shhSav)/2.0);
    double xb = eb/(sqrt(shhSav)/2.0);
    double xB = eB/(sqrt(shhSav)/2.0);
    if (Rfunity) return xA*xB/xa/xb;
    if ( !beamAPtr->insideBounds(xa, Qt2A) || !beamAPtr->insideBounds(xA, Qt2A) ||
         !beamBPtr->insideBounds(xb, Qt2B) || !beamBPtr->insideBounds(xB, Qt2B) ) {
      nPDFoutsideBound++;
      if (nPDFoutsideBound <= nReportPDFbound)
        printf("%s::PDFratio evaluation outside boundaries \n"
               " Further warnings suppressed!", name().c_str());
    }
    double Rpdf = 1.0;
    if (idA==21) idA = 0;
    if (idB==21) idB = 0;
    double top    = beamAPtr->xfISR(iSys, idA, xa, Qt2A);
    double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
    if (abs(top)    < TINYPDF) top    = TINYPDF;
    if (abs(bottom) < TINYPDF) bottom = TINYPDF;
    Rpdf          = top/xa/bottom*xA;
    top           = beamBPtr->xfISR(iSys, idB, xb, Qt2B);
    bottom        = beamBPtr->xfISR(iSys, idB, xB, Qt2B);
    if (abs(top)    < TINYPDF) top    = TINYPDF;
    if (abs(bottom) < TINYPDF) bottom = TINYPDF;
    Rpdf         *= top/xb/bottom*xB;
    return Rpdf;  
  }
  
  bool TrialGeneratorISR::checkInit() {
    if (!isInit) {
      printf("%s::checkInit() Error! Not initialized \n", name().c_str());
      return false;        
    } else {
        return true;        
    }
  }
  
  
  //----------------------------------------------------------------------
  // A collinear trial function (derived base class)
  //----------------------------------------------------------------------
    
  // Trial antenna function
  double TrialIIGCollA::aTrial(double saj, double sjb, double sAB) {
    if (saj < 0. || sjb < 0.) return 0.;
    double sab = sAB+saj+sjb;
    return 2.0*sab*sab/sAB/sAB/saj;
  }
    
  // Modified trial antenna function, used when combining accept
  // probabilities for gluon emission
  double TrialIIGCollA::aTrialMod(double saj, double sjb, double sAB,
                                  double eA, double eB, double ea, double eb) {
    if (saj < 0. || sjb < 0.) return 0.;
    double xRatio = eA/ea*eB/eb;
    double sab = sAB+saj+sjb;
    return 2.0*sab*sab/sAB/sAB/saj*xRatio;
  }
    
  // Physical antenna function
  double TrialIIGCollA::aPhys(double saj, double sjb, double sAB, int, int) {
    if (saj < 0. || sjb < 0.) return 0.;
    double sab = sAB+saj+sjb;
    return (2.0*sjb/saj*sab/sAB+2.0*sjb/saj*sAB/(sab+saj))/sAB;
  }
  
  // Generate a new Q value, with first-order running alphaS
  double TrialIIGCollA::genQ2(double q2old, double sAB, double zMin, double zMax, 
                              double colFac, double PDFratio, 
                              double b0, double kR, double Lambda,
                              double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAB < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 2.0*M_PI*b0/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    double facLam = pow2(Lambda/kR) * evNormIIpT2Sav;
    return exp(pow(ran,comFac) * log(q2old/facLam)) * facLam;
  }
  
  // Generate a new Q value, with constant trial alphaS
  double TrialIIGCollA::genQ2(double q2old, double sAB, double zMin, double zMax, 
                              double colFac, double alphaSvalue, double PDFratio, 
                              double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAB < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 2.0*M_PI/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    return q2old * pow(ran,comFac/alphaSvalue);
  }
  
  // Generate a new Q value, use in the very unordered region
  double TrialIIGCollA::genQ2withQN(double q2old, double q2hat, double N, double sAB, 
                                    double zMin, double zMax, 
                                    double colFac, double alphaSvalue, double PDFratio, 
                                    double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAB < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 2.0*M_PI/Iz/colFac/alphaSvalue/PDFratio/headroomFactor;
    double expo   = ((N-2.0)/2.0);
    double ran    = rndmPtr->flat();
    return pow( pow(q2old,expo)/(1.0-pow(q2old/q2hat,expo)*
                                     log(ran)*comFac*expo), 1.0/expo);
  }
  
  // Generate a new zeta value in [zMin,zMax]
  double TrialIIGCollA::genZ(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return -1.;
    double ran = rndmPtr->flat();
    double z = -1.0 + (1.0+zMin)*pow((1.0+zMax)/(1.0+zMin),ran);
    return z;
  }
  
  // The zeta integral
  double TrialIIGCollA::getIz(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return 0.0;
    return log((1.0+zMax)/(1.0+zMin)); 
  }
    
  // The zeta boundaries, for a given value of the evolution variable
  double TrialIIGCollA::getZmin(double Qt2, double sAB, double, double) {
    if ( ( (shhSav-sAB)*(shhSav-sAB) - (4.0*Qt2/evNormIIpT2Sav*shhSav) ) 
         < TINYPDF ) return 0.5*(shhSav-sAB)/sAB;
    double sajm = 0.5*( shhSav-sAB-sqrt((shhSav-sAB)*(shhSav-sAB)-
                                        (4.0*Qt2/evNormIIpT2Sav*shhSav)) );
    return sajm/sAB;
  }
  double TrialIIGCollA::getZmax(double Qt2, double sAB, double, double) {
    if ( ( (shhSav-sAB)*(shhSav-sAB) - (4.0*Qt2/evNormIIpT2Sav*shhSav) ) 
         < TINYPDF ) return 0.5*(shhSav-sAB)/sAB;
    double sajp = 0.5*( shhSav-sAB+sqrt((shhSav-sAB)*(shhSav-sAB)-
                                        (4.0*Qt2/evNormIIpT2Sav*shhSav)) );
    return sajp/sAB;
  }
    
  // Inverse transforms
  // To obtain saj and sjb from Qt2 and zeta
  double TrialIIGCollA::getS1j(double Qt2, double zeta, double sAB) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getSj2(Qt2,-zeta,sAB);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getS1j() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double saj = Qt2/evNormIIpT2Sav*(1.0+zeta)/(zeta-Qt2/evNormIIpT2Sav/sAB);
    return saj;
  }
  double TrialIIGCollA::getSj2(double Qt2, double zeta, double sAB) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getS1j(Qt2,-zeta,sAB);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getSj2() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double sjb = zeta*sAB;
    return sjb;
  }
    
  // Trial PDF ratio
  double TrialIIGCollA::trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                      int iSys, bool Rfunity,
                                      int idA, int idB, double eA, double eB,
                                      double Qt2A, double Qt2B, bool useSaved) {
    if (useSaved) return trialPDFratioSav;
    double xA = eA/(sqrt(shhSav)/2.0);
    double xB = eB/(sqrt(shhSav)/2.0);
    if (idA==21) idA = 0;
    if (idB==21) idB = 0;
    double Rpdf = 1.0;
    if (Rfunity) {
      Rpdf = 1.0/xA * 1.0/xB;
    } else if (useRandomX) {
      double xa     = xA*pow(0.95/xA,rndmPtr->flat());
      double xb     = xB*pow(0.95/xB,rndmPtr->flat());
      double top    = beamAPtr->xfISR(iSys, idA, xa, Qt2A);
      double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
      if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
      if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
      Rpdf          = top/bottom;
      top           = beamBPtr->xfISR(iSys, idB, xb, Qt2B);
      bottom        = beamBPtr->xfISR(iSys, idB, xB, Qt2B);
      if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
      if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
      Rpdf         *= top/bottom;
      Rpdf          = max(Rpdf,1.0);
    } else {
      Rpdf = 1.0;
    }
    trialPDFratioSav = Rpdf;
    return Rpdf;
  }
  
  // Physical PDF ratio
  double TrialIIGCollA::physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                     int iSys, bool Rfunity,
                                     int idA, int idB, double eA, double eB, 
                                     double ea, double eb, double Qt2A, double Qt2B) {
    double xA = eA/(sqrt(shhSav)/2.0);
    double xB = eB/(sqrt(shhSav)/2.0);
    double xa = ea/(sqrt(shhSav)/2.0);
    double xb = eb/(sqrt(shhSav)/2.0);
    if (Rfunity) return 1.0;
    if ( !beamAPtr->insideBounds(xa, Qt2A) || !beamAPtr->insideBounds(xA, Qt2A) ||
         !beamBPtr->insideBounds(xb, Qt2B) || !beamBPtr->insideBounds(xB, Qt2B) ) {
      nPDFoutsideBound++;
      if (nPDFoutsideBound <= nReportPDFbound)
        printf("%s::PDFratio evaluation outside boundaries \n"
               " Further warnings suppressed!", name().c_str());
    }
    double Rpdf = 1.0;
    if (idA==21) idA = 0;
    if (idB==21) idB = 0;
    double top    = beamAPtr->xfISR(iSys, idA, xa, Qt2A);
    double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
    if (abs(top)    < TINYPDF) top    = TINYPDF;
    if (abs(bottom) < TINYPDF) bottom = TINYPDF;
    Rpdf          = top/bottom;
    top           = beamBPtr->xfISR(iSys, idB, xb, Qt2B);
    bottom        = beamBPtr->xfISR(iSys, idB, xB, Qt2B);
    if (abs(top)    < TINYPDF) top    = TINYPDF;
    if (abs(bottom) < TINYPDF) bottom = TINYPDF;
    Rpdf         *= top/bottom;
    return Rpdf;
  }
  
  
  //----------------------------------------------------------------------
  // A splitting trial function (derived base class), q->gqbar
  //----------------------------------------------------------------------
    
  // Trial antenna function
  double TrialIISplitA::aTrial(double saj, double sjb, double sAB) {
    if (saj < 0. || sjb < 0.) return 0.;
    double sab = sAB+saj+sjb;
    return sab/saj/sAB;
  }
  
  // Physical antenna function
  double TrialIISplitA::aPhys(double saj, double sjb, double sAB, int, int) {
    double sab = sAB+saj+sjb;
    return (-2.0*sjb*sAB/saj/(sab-saj)+sab/saj)/sAB;
  }
  
  // Generate a new Q value, with first-order running alphaS
  double TrialIISplitA::genQ2(double q2old, double sAB, double zMin, double zMax, 
                              double colFac, double PDFratio, 
                              double b0, double kR, double Lambda,
                              double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAB < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI*b0/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    double facLam = pow2(Lambda/kR) * ((useMevolSav) ? evNormMsplitISav : evNormIIpT2Sav);
    return exp(pow(ran,comFac) * log(q2old/facLam)) * facLam;
  }
  
  // Generate a new Q value, with constant trial alphaS
  double TrialIISplitA::genQ2(double q2old, double sAB, double zMin, double zMax, 
                              double colFac, double alphaSvalue, double PDFratio, 
                              double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAB < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    return q2old * pow(ran,comFac/alphaSvalue);
  }
  
  // Generate a new Q value, use in the very unordered region
  double TrialIISplitA::genQ2withQN(double q2old, double q2hat, double N, double sAB, 
                                    double zMin, double zMax, 
                                    double colFac, double alphaSvalue, double PDFratio, 
                                    double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAB < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI/Iz/colFac/alphaSvalue/PDFratio/headroomFactor;
    double expo   = ((N-2.0)/2.0);
    double ran    = rndmPtr->flat();
    return pow( pow(q2old,expo)/(1.0-pow(q2old/q2hat,expo)*
                                     log(ran)*comFac*expo), 1.0/expo);
  }
  
  // Generate a new Q value, with running of the PDFs towards the mass threshold
  double TrialIISplitA::genQ2(double q2old, double sAB, double zMin, double zMax, 
                              double colFac, double alphaSvalue, double PDFratio, 
                              int idA, int, double /*eA*/, double /*eB*/, bool, 
                              double headroomFactor) {
    // Use only if the user wants to get rid of c and b quarks
    // and use only in the right evolution window
    double mQ = (abs(idA) == 4 ? mcSav : mbSav);
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAB < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI/Iz/colFac/alphaSvalue/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    return ( exp(pow(ran,comFac) * 
             log(q2old/( useMevolSav ? evNormMsplitISav : evNormIIpT2Sav)/pow2(mQ))) * 
             pow2(mQ) );
  }
  
  // Generate a new zeta value in [zMin,zMax]
  double TrialIISplitA::genZ(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return -1.;
    double ran = rndmPtr->flat();
    if (useMevolSav) {
      if (alphaSplitSav==0.0) return zMax*pow(zMin/zMax,ran);
      return pow(ran*(pow(zMax,-alphaSplitSav)-pow(zMin,-alphaSplitSav))+
                 pow(zMin,-alphaSplitSav),1.0/(-alphaSplitSav));
    }
    return ( -1.0 + (1.0+zMin)*pow((1.0+zMax)/(1.0+zMin),ran) );
  }
  
  // The zeta integral
  double TrialIISplitA::getIz(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return 0.0;
    if (useMevolSav) {
      if (alphaSplitSav==0.0) return log(zMax/zMin);
      return (pow(zMax,-alphaSplitSav)-pow(zMin,-alphaSplitSav))/(-alphaSplitSav);
    }
    return log((1.0+zMax)/(1.0+zMin)); 
  }
    
  // The zeta boundaries, for a given value of the evolution variable
  double TrialIISplitA::getZmin(double Qt2, double sAB, double, double) {
    if (useMevolSav) return ((Qt2/evNormMsplitISav+sAB)/sAB);
    if ( ( (shhSav-sAB)*(shhSav-sAB) - (4.0*Qt2/evNormIIpT2Sav*shhSav) )
         < TINYPDF ) return 0.5*(shhSav-sAB)/sAB;
    double sajm = 0.5*( shhSav-sAB-sqrt((shhSav-sAB)*(shhSav-sAB)-
                                        (4.0*Qt2/evNormIIpT2Sav*shhSav)) );
    return sajm/sAB;
  }
  double TrialIISplitA::getZmax(double Qt2, double sAB, double, double) {
    if (useMevolSav) return (shhSav/sAB);
    if ( ( (shhSav-sAB)*(shhSav-sAB) - (4.0*Qt2/evNormIIpT2Sav*shhSav) )
         < TINYPDF ) return 0.5*(shhSav-sAB)/sAB;
    double sajp = 0.5*( shhSav-sAB+sqrt((shhSav-sAB)*(shhSav-sAB)-
                                        (4.0*Qt2/evNormIIpT2Sav*shhSav)) );
    return sajp/sAB;
  }
    
  // Inverse transforms
  // To obtain saj and sjb from Qt2 and zeta
  double TrialIISplitA::getS1j(double Qt2, double zeta, double sAB) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getSj2(Qt2,-zeta,sAB);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getS1j() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double saj = Qt2/evNormIIpT2Sav*(1.0+zeta)/(zeta-Qt2/evNormIIpT2Sav/sAB);
    if (useMevolSav) saj = Qt2/evNormMsplitISav;
    return saj;
  }
  double TrialIISplitA::getSj2(double Qt2, double zeta, double sAB) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getS1j(Qt2,-zeta,sAB);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getSj2() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double sjb = zeta*sAB;
    if (useMevolSav) sjb = sAB*(zeta-1)-Qt2/evNormMsplitISav;
    return sjb;
  }
    
  // Trial PDF ratio
  double TrialIISplitA::trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                      int iSys, bool Rfunity,
                                      int idA, int idB, double eA, double eB, 
                                      double Qt2A, double Qt2B, bool useSaved) {
    if (useSaved) return trialPDFratioSav;
    double xA = eA/(sqrt(shhSav)/2.0);
    double xB = eB/(sqrt(shhSav)/2.0);
    if (idA==21) idA = 0;
    if (idB==21) idB = 0;
    double Rpdf = 1.0;
    if (Rfunity) {
      Rpdf = pow(1.0/xA,alphaSplitSav) * pow(1.0/xB,alphaSplitSav);
    } else if (useRandomX) {
      double xa     = xA*pow(0.95/xA,rndmPtr->flat());
      double xb     = xB*pow(0.95/xB,rndmPtr->flat());
      double top    = beamAPtr->xfISR(iSys, 0,   xa, Qt2A);
      double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
      if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
      if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
      Rpdf          = top * pow(xa,alphaSplitSav-1.0) /
                      bottom / pow(xA,alphaSplitSav-1.0);
      top           = beamBPtr->xfISR(iSys, idB, xb, Qt2B);
      bottom        = beamBPtr->xfISR(iSys, idB, xB, Qt2B);
      if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
      if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
      Rpdf         *= top * pow(xb,alphaSplitSav-1.0) /
                      bottom / pow(xB,alphaSplitSav-1.0);
    } else {
      double top    = beamAPtr->xfISR(iSys, 0,   xA, Qt2A);
      double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
      if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
      if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
      Rpdf          = 1.2*top/bottom;
    }
    trialPDFratioSav = Rpdf;
    return Rpdf;
  }
  
  // Physical PDF ratio
  double TrialIISplitA::physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                     int iSys, bool Rfunity,
                                     int idA, int idB, double eA, double eB,
                                     double ea, double eb, double Qt2A, double Qt2B) {
    double xA = eA/(sqrt(shhSav)/2.0);
    double xB = eB/(sqrt(shhSav)/2.0);
    double xa = ea/(sqrt(shhSav)/2.0);
    double xb = eb/(sqrt(shhSav)/2.0);
    if (Rfunity) return pow(xa/xA,alphaSplitSav)*pow(xb/xB,alphaSplitSav)*xA*xB/xa/xb;
    if ( !beamAPtr->insideBounds(xa, Qt2A) || !beamAPtr->insideBounds(xA, Qt2A) ||
         !beamBPtr->insideBounds(xb, Qt2B) || !beamBPtr->insideBounds(xB, Qt2B) ) {
      nPDFoutsideBound++;
      if (nPDFoutsideBound <= nReportPDFbound)
        printf("%s::PDFratio evaluation outside boundaries \n"
               " Further warnings suppressed!", name().c_str());
    }
    double Rpdf = 1.0;
    if (idA==21) idA = 0;
    if (idB==21) idB = 0;
    double top    = beamAPtr->xfISR(iSys, 0,   xa, Qt2A);
    double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
    if (abs(top)    < TINYPDF) top    = TINYPDF;
    if (abs(bottom) < TINYPDF) bottom = TINYPDF;
    Rpdf          = top * pow(xa,alphaSplitSav-1.0) /
                    bottom / pow(xA,alphaSplitSav-1.0);
    top           = beamBPtr->xfISR(iSys, idB, xb, Qt2B);
    bottom        = beamBPtr->xfISR(iSys, idB, xB, Qt2B);
    if (abs(top)    < TINYPDF) top    = TINYPDF;
    if (abs(bottom) < TINYPDF) bottom = TINYPDF;
    Rpdf         *= top * pow(xb,alphaSplitSav-1.0) /
                    bottom / pow(xB,alphaSplitSav-1.0);
    return Rpdf; 
  }
  
  
  //----------------------------------------------------------------------
  // A conversion trial function (derived base class), g->qqbar
  //----------------------------------------------------------------------
    
  // Trial antenna function
  double TrialIIConvA::aTrial(double saj, double sjb, double sAB) {
    if (saj < 0. || sjb < 0.) return 0.;
    double sab = sAB+saj+sjb;
    return sab*sab/sAB/sAB/saj;
  }
    
  // Physical antenna function
  double TrialIIConvA::aPhys(double saj, double sjb, double sAB, int, int) {
    if (saj < 0. || sjb < 0.) return 0.;
    double sab = sAB+saj+sjb;
    return 1.0/2.0/saj*(sjb*sjb+sab*sab)/sAB/sAB;
  }
  
  // Generate a new Q value, with first-order running alphaS
  double TrialIIConvA::genQ2(double q2old, double sAB, double zMin, double zMax, 
                             double colFac, double PDFratio, 
                             double b0, double kR, double Lambda,
                             double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAB < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI*b0/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    double facLam = pow2(Lambda/kR) * ((useMevolSav) ? evNormMconvISav : evNormIIpT2Sav);
    return exp(pow(ran,comFac) * log(q2old/facLam)) * facLam;
  }
  
  // Generate a new Q value, with constant trial alphaS
  double TrialIIConvA::genQ2(double q2old, double sAB, double zMin, double zMax, 
                             double colFac, double alphaSvalue, double PDFratio, 
                             double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAB < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    return q2old * pow(ran,comFac/alphaSvalue);
  }
  
  // Generate a new Q value, use in the very unordered region
  double TrialIIConvA::genQ2withQN(double q2old, double q2hat, double N, double sAB, 
                                   double zMin, double zMax, 
                                   double colFac, double alphaSvalue, double PDFratio, 
                                   double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAB < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI/Iz/colFac/alphaSvalue/PDFratio/headroomFactor;
    double expo   = ((N-2.0)/2.0);
    double ran    = rndmPtr->flat();
    return pow( pow(q2old,expo)/(1.0-pow(q2old/q2hat,expo)*
                                     log(ran)*comFac*expo), 1.0/expo);
  }
  
  // Generate a new zeta value in [zMin,zMax]
  double TrialIIConvA::genZ(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return -1.;
    double ran = rndmPtr->flat();
    if (useMevolSav) {
      if (alphaConvSav==1.0) return zMax*pow(zMin/zMax,ran);
      return pow(ran*(pow(zMax,1.0-alphaConvSav)-pow(zMin,1.0-alphaConvSav))+
                 pow(zMin,1.0-alphaConvSav),1.0/(1.0-alphaConvSav));
    }
    return ( -1.0 + (1.0+zMin)*pow((1.0+zMax)/(1.0+zMin),ran) );
  }
  
  // The zeta integral
  double TrialIIConvA::getIz(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return 0.0;
    if (useMevolSav) {
      if (alphaConvSav==1.0) return log(zMax/zMin);
      return (pow(zMax,1.0-alphaConvSav)-pow(zMin,1.0-alphaConvSav))/(1.0-alphaConvSav);
    }
    return log((1.0+zMax)/(1.0+zMin)); 
  }
    
  // The zeta boundaries, for a given value of the evolution variable
  double TrialIIConvA::getZmin(double Qt2, double sAB, double, double) {
    if (useMevolSav) return ((Qt2/evNormMconvISav+sAB)/sAB);
    if ( ( (shhSav-sAB)*(shhSav-sAB) - (4.0*Qt2/evNormIIpT2Sav*shhSav) )
         < TINYPDF ) return 0.5*(shhSav-sAB)/sAB;
    double sajm = 0.5*( shhSav-sAB-sqrt((shhSav-sAB)*(shhSav-sAB)-
                                        (4.0*Qt2/evNormIIpT2Sav*shhSav)) );
    return sajm/sAB;
  }
  double TrialIIConvA::getZmax(double Qt2, double sAB, double, double) {
    if (useMevolSav) return (shhSav/sAB);
    if ( ( (shhSav-sAB)*(shhSav-sAB) - (4.0*Qt2/evNormIIpT2Sav*shhSav) )
         < TINYPDF ) return 0.5*(shhSav-sAB)/sAB;
    double sajp = 0.5*( shhSav-sAB+sqrt((shhSav-sAB)*(shhSav-sAB)-
                                        (4.0*Qt2/evNormIIpT2Sav*shhSav)) );
    return sajp/sAB;
  }
    
  // Inverse transforms
  // To obtain saj and sjb from Qt2 and zeta
  double TrialIIConvA::getS1j(double Qt2, double zeta, double sAB) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getSj2(Qt2,-zeta,sAB);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getS1j() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double saj = Qt2/evNormIIpT2Sav*(1.0+zeta)/(zeta-Qt2/evNormIIpT2Sav/sAB);
    if (useMevolSav) saj = Qt2/evNormMconvISav;
    return saj;
  }
  double TrialIIConvA::getSj2(double Qt2, double zeta, double sAB) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getS1j(Qt2,-zeta,sAB);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getSj2() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double sjb = zeta*sAB;
    if (useMevolSav) sjb = sAB*(zeta-1)-Qt2/evNormMconvISav;
    return sjb;
  }
    
  // Trial PDF ratio
  double TrialIIConvA::trialPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                     int iSys, bool Rfunity,
                                     int idA, int idB, double eA, double eB,
                                     double Qt2A, double Qt2B, bool useSaved) {
    if (useSaved) return trialPDFratioSav;
    double xA = eA/(sqrt(shhSav)/2.0);
    double xB = eB/(sqrt(shhSav)/2.0);
    if (idA==21) idA = 0;
    if (idB==21) idB = 0;
    double Rpdf = 1.0;
    if (Rfunity) {
      Rpdf = pow(1.0/xA,alphaConvSav) * pow(1.0/xB,alphaConvSav);
    } else {
      int nQuark = 5;
           if (Qt2A <= 4.0*mcSav*mcSav) nQuark = 3;
      else if (Qt2A <= 4.0*mbSav*mbSav) nQuark = 4;
      if (nQuark > nGtoQISRSav) nQuark = nGtoQISRSav;
      if (useRandomX) {
        double xa     = xA*pow(0.95/xA,rndmPtr->flat());
        double xb     = xB*pow(0.95/xB,rndmPtr->flat());
        double top    = 0.0;
        for (int i=1; i<=nQuark; i++)
          top += ( beamAPtr->xfISR(iSys, i, xa, Qt2A) * 
                   ((i<3) && (sqrt(Qt2A)<mbSav) ? 2.5 : 1.0)
                   + beamAPtr->xfISR(iSys, -i, xa, Qt2A) );
        double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
        if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
        if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
        Rpdf          = top * pow(xa,alphaConvSav-1.0) /
                        bottom / pow(xA,alphaConvSav-1.0);
        top           = beamBPtr->xfISR(iSys, idB, xb, Qt2B);
        bottom        = beamBPtr->xfISR(iSys, idB, xB, Qt2B);
        if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
        if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
        Rpdf         *= top * pow(xb,alphaConvSav-1.0) /
                        bottom / pow(xB,alphaConvSav-1.0);
      } else {
        double top = 0.0;
        for (int i=1; i<=nQuark; i++)
          top += ( beamAPtr->xfISR(iSys, i, xA, Qt2A) * 
                   ((i<3) && (sqrt(Qt2A)<mbSav) ? 2.5 : 1.0)
                   + beamAPtr->xfISR(iSys, -i, xA, Qt2A) );
        double bottom = beamAPtr->xfISR(iSys, 0, xA, Qt2A);
        if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
        if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
        Rpdf = top/bottom;
      }
    }
    trialPDFratioSav = Rpdf;
    return Rpdf;
  }
  
  
  // Physical PDF ratio
  double TrialIIConvA::physPDFratio(BeamParticle* beamAPtr, BeamParticle* beamBPtr, 
                                    int iSys, bool Rfunity,
                                    int idA, int idB, double eA, double eB,
                                    double ea, double eb, double Qt2A, double Qt2B) {
    double xA = eA/(sqrt(shhSav)/2.0);
    double xB = eB/(sqrt(shhSav)/2.0);
    double xa = ea/(sqrt(shhSav)/2.0);
    double xb = eb/(sqrt(shhSav)/2.0);
    int nQuark = 5;
         if (Qt2A <= 4.0*mcSav*mcSav) nQuark = 3;
    else if (Qt2A <= 4.0*mbSav*mbSav) nQuark = 4;
    if (nQuark > nGtoQISRSav) nQuark = nGtoQISRSav;
    if (Rfunity) {
      gConvIDSav = (rndmPtr->flat() > 0.5 ? -1.0 : 1.0) * 
                   ( int(rndmPtr->flat() * nQuark) + 1 );
      return pow(xa/xA,alphaConvSav)*pow(xb/xB,alphaConvSav)*xA*xB/xa/xb;
    }
    if ( !beamAPtr->insideBounds(xa, Qt2A) || !beamAPtr->insideBounds(xA, Qt2A) ||
         !beamBPtr->insideBounds(xb, Qt2B) || !beamBPtr->insideBounds(xB, Qt2B) ) {
      nPDFoutsideBound++;
      if (nPDFoutsideBound <= nReportPDFbound)
        printf("%s::PDFratio evaluation outside boundaries \n"
               " Further warnings suppressed!", name().c_str());
    }
    double Rpdf = 1.0;
    if (idA==21) idA = 0;
    if (idB==21) idB = 0;
    double top = 0.0;
    map<int,double> pdfValues;
    for (int i=1; i<=nQuark; i++) {
      pdfValues[i]  = beamAPtr->xfISR(iSys,  i, xa, Qt2A);
      pdfValues[-i] = beamAPtr->xfISR(iSys, -i, xa, Qt2A);
      top           = top + pdfValues[i] + pdfValues[-i];
    }
    double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
    if (abs(top)    < TINYPDF) top    = TINYPDF;
    if (abs(bottom) < TINYPDF) bottom = TINYPDF;
    Rpdf = top * pow(xa,alphaConvSav-1.0) /
           bottom / pow(xA,alphaConvSav-1.0);
    // Normalize pdfPtr to sum = unity
    vector<double> pdfNormed;
    vector<int> flavour;
    pdfNormed.push_back(0.0);
    flavour.push_back(0);
    double count = 0.0;
    for (map<int,double>::iterator it=pdfValues.begin(); it!=pdfValues.end(); ++it) {
      count = count + it->second/top;
      pdfNormed.push_back(count);
      flavour.push_back(it->first);
    }
    // Random number to decide for flavour
    double ran = rndmPtr->flat();
    for (int i=1; i<(int)pdfNormed.size(); i++)
      if ( (ran>=pdfNormed[i-1]) && (ran<pdfNormed[i]) )
        gConvIDSav = flavour[i];
    top    = beamBPtr->xfISR(iSys, idB, xb, Qt2B);
    bottom = beamBPtr->xfISR(iSys, idB, xB, Qt2B);
    if (abs(top)    < TINYPDF) top    = TINYPDF;
    if (abs(bottom) < TINYPDF) bottom = TINYPDF;
    Rpdf  *= top * pow(xb,alphaConvSav-1.0) / 
             bottom / pow(xB,alphaConvSav-1.0);
    return Rpdf;
  }
  
  
  //----------------------------------------------------------------------
  // Soft-eikonal trial function for IF (derived base class)
  //----------------------------------------------------------------------
    
  // Trial antenna function
  double TrialIFSoft::aTrial(double saj, double sjk, double sAK) {
    if (saj < 0. || sjk < 0.) return 0.;
    return (2.0*sAK+sjk)*(sAK+sjk)/saj/sjk/sAK;
  }
    
  // Modified trial antenna function, used when combining accept
  // probabilities for gluon emission
  double TrialIFSoft::aTrialMod(double saj, double sjk, double sAK,
                                double, double, double, double) {
    if (saj < 0. || sjk < 0.) return 0.;
    return (2.0*sAK+sjk)*(sAK+sjk)/saj/sjk/sAK;
  }
    
  // Physical antenna function
  double TrialIFSoft::aPhys(double saj, double sjk, double sAK, int, int) {
    if (saj < 0. || sjk < 0.) return 0.;
    double sak = sAK+sjk-saj;
    return 2.0*sak/sjk/saj;
  }
  
  // Generate a new Q value, with first-order running alphaS
  double TrialIFSoft::genQ2(double q2old, double sAK, double zMin, double zMax, 
                            double colFac, double PDFratio, 
                            double b0, double kR, double Lambda,
                            double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI*b0/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    double facLam = pow2(Lambda/kR) * evNormIFpT2Sav;
    return exp(pow(ran,comFac) * log(q2old/facLam)) * facLam;
  }
  
  // Generate a new Q value, with constant trial alphaS
  double TrialIFSoft::genQ2(double q2old, double sAK, double zMin, double zMax, 
                            double colFac, double alphaSvalue, double PDFratio, 
                            double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    return q2old * pow(ran,comFac/alphaSvalue);
  }
  
  // Generate a new Q value, use in the very unordered region
  double TrialIFSoft::genQ2withQN(double q2old, double q2hat, double N, double sAK, 
                                  double zMin, double zMax, 
                                  double colFac, double alphaSvalue, double PDFratio, 
                                  double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI/Iz/colFac/alphaSvalue/PDFratio/headroomFactor;
    double expo   = ((N-2.0)/2.0);
    double ran    = rndmPtr->flat();
    return pow( pow(q2old,expo)/(1.0-pow(q2old/q2hat,expo)*
                                     log(ran)*comFac*expo), 1.0/expo);
  }
  
  // Generate a new zeta value in [zMin,zMax]
  double TrialIFSoft::genZ(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return -1.;
    double ran = rndmPtr->flat();
    double z = zMax*pow(zMin/zMax,ran);
    return z;
  }
  
  // The zeta integral
  double TrialIFSoft::getIz(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return 0.0;
    return log(zMax/zMin); 
  }
    
  // The zeta boundaries, for a given value of the evolution variable
  double TrialIFSoft::getZmin(double Qt2, double sAK, double eA,
                              double eAused) {
    double xA     = eA/(sqrt(shhSav)/2.0);
    double eAmax  = ( (sqrt(shhSav)/2.0) - (eAused-eA) );
    double xAmax  = eAmax/(sqrt(shhSav)/2.0);
    double sjkmax = sAK*(xAmax-xA)/xA;
    return Qt2/evNormIFpT2Sav*(sAK+sjkmax)/sjkmax/sjkmax;
  }
  double TrialIFSoft::getZmax(double Qt2, double sAK, double, double) {
    return (Qt2/evNormIFpT2Sav+sAK)/Qt2*evNormIFpT2Sav;
  }
    
  // Inverse transforms
  // To obtain saj and sjk from Qt2 and zeta
  double TrialIFSoft::getS1j(double Qt2, double zeta, double sAK) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getSj2(Qt2,-zeta,sAK);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getS1j() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double saj = Qt2/evNormIFpT2Sav/2.0*
                 (1.0+sqrt(1.0+4.0*zeta*sAK/Qt2*evNormIFpT2Sav));
    return saj;
  }
  double TrialIFSoft::getSj2(double Qt2, double zeta, double sAK) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getS1j(Qt2,-zeta,sAK);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getSj2() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double sjk = Qt2/evNormIFpT2Sav/zeta/2.0*
                 (1.0+sqrt(1.0+4.0*zeta*sAK/Qt2*evNormIFpT2Sav));
    return sjk;
  }
    
  // Trial PDF ratio
  double TrialIFSoft::trialPDFratio(BeamParticle* beamAPtr, BeamParticle*, 
                                    int iSys, bool Rfunity,
                                    int idA, int, double eA, double, 
                                    double Qt2A, double, bool useSaved) {
    if (useSaved) return trialPDFratioSav;
    double xA = eA/(sqrt(shhSav)/2.0);
    if (idA==21) idA = 0;
    double Rpdf = 1.0;
    if (Rfunity) {
      Rpdf = 1.0;
    } else if (useRandomX) {
      double xa     = xA*pow(0.95/xA,rndmPtr->flat());
      double top    = beamAPtr->xfISR(iSys, idA, xa, Qt2A);
      double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
      if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
      if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
      Rpdf          = top/xa/bottom*xA;
      Rpdf          = max(Rpdf,1.0);
    } else {
      Rpdf = 1.0;
    }
    trialPDFratioSav = Rpdf;
    return Rpdf;
  }
  
  // Physical PDF ratio
  double TrialIFSoft::physPDFratio(BeamParticle* beamAPtr, BeamParticle*, 
                                   int iSys, bool Rfunity,
                                   int idA, int, double eA, double, 
                                   double ea, double, double Qt2A, double) {
    double xA = eA/(sqrt(shhSav)/2.0);
    double xa = ea/(sqrt(shhSav)/2.0);
    if (Rfunity) return xA/xa;
    if ( !beamAPtr->insideBounds(xa, Qt2A) || !beamAPtr->insideBounds(xA, Qt2A) ) {
      nPDFoutsideBound++;
      if (nPDFoutsideBound <= nReportPDFbound)
        printf("%s::PDFratio evaluation outside boundaries \n"
               " Further warnings suppressed!", name().c_str());
    }
    double Rpdf = 1.0;
    if (idA==21) idA = 0;
    double top    = beamAPtr->xfISR(iSys, idA, xa, Qt2A);
    double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
    if (abs(top)    < TINYPDF) top    = TINYPDF;
    if (abs(bottom) < TINYPDF) bottom = TINYPDF;
    Rpdf          = top/xa/bottom*xA;
    return Rpdf;  
  }
  
  
  //----------------------------------------------------------------------
  // A quark collinear trial function for IF (derived base class)
  //----------------------------------------------------------------------
    
  // Trial antenna function
  double TrialIFQCollA::aTrial(double saj, double sjk, double sAK) {
    if (saj < 0. || sjk < 0.) return 0.;
    return 1.0/sAK*(sjk+sAK)/saj;
  }
    
  // Modified trial antenna function, used when combining accept
  // probabilities for gluon emission
  double TrialIFQCollA::aTrialMod(double saj, double sjk, double sAK,
                                  double eA, double, double ea, double) {
    if (saj < 0. || sjk < 0.) return 0.;
    double xRatio = pow(eA/ea,betaQCollISav);
    return 1.0/sAK*(sjk+sAK)/saj*xRatio;
  }
    
  // Physical antenna function
  double TrialIFQCollA::aPhys(double saj, double sjk, double sAK, int, int) {
    if (saj < 0. || sjk < 0.) return 0.;
    return 1.0/sAK*sjk/saj;
  }
  
  // Generate a new Q value, with first-order running alphaS
  double TrialIFQCollA::genQ2(double q2old, double sAK, double zMin, double zMax, 
                              double colFac, double PDFratio, 
                              double b0, double kR, double Lambda,
                              double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI*b0/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    double facLam = pow2(Lambda/kR) * evNormIFpT2Sav;
    return exp(pow(ran,comFac) * log(q2old/facLam)) * facLam;
  }
  
  // Generate a new Q value, with constant trial alphaS
  double TrialIFQCollA::genQ2(double q2old, double sAK, double zMin, double zMax, 
                              double colFac, double alphaSvalue, double PDFratio, 
                              double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    return q2old * pow(ran,comFac/alphaSvalue);
  }
  
  // Generate a new Q value, use in the very unordered region
  double TrialIFQCollA::genQ2withQN(double q2old, double q2hat, double N, double sAK, 
                                    double zMin, double zMax, 
                                    double colFac, double alphaSvalue, double PDFratio, 
                                    double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI/Iz/colFac/alphaSvalue/PDFratio/headroomFactor;
    double expo   = ((N-2.0)/2.0);
    double ran    = rndmPtr->flat();
    return pow( pow(q2old,expo)/(1.0-pow(q2old/q2hat,expo)*
                                     log(ran)*comFac*expo), 1.0/expo);
  }
  
  // Generate a new zeta value in [zMin,zMax]
  double TrialIFQCollA::genZ(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return -1.;
    double ran = rndmPtr->flat();
    if (betaQCollISav==0.0) return zMax*pow(zMin/zMax,ran);
    return pow(ran*(pow(zMax,-betaQCollISav)-pow(zMin,-betaQCollISav))+
               pow(zMin,-betaQCollISav),1.0/(-betaQCollISav));
  }
  
  // The zeta integral
  double TrialIFQCollA::getIz(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return 0.0;
    if (betaQCollISav==0.0) return log(zMax/zMin);
    return (pow(zMax,-betaQCollISav)-pow(zMin,-betaQCollISav))/(-betaQCollISav);
  }
    
  // The zeta boundaries, for a given value of the evolution variable
  double TrialIFQCollA::getZmin(double Qt2, double sAK, double, double) {
    return (Qt2/evNormIFpT2Sav+sAK)/sAK;
  }
  double TrialIFQCollA::getZmax(double, double sAK, double eA,
                                double eAused) {
    double xA     = eA/(sqrt(shhSav)/2.0);
    double eAmax  = ( (sqrt(shhSav)/2.0) - (eAused-eA) );
    double xAmax  = eAmax/(sqrt(shhSav)/2.0);
    double sjkmax = sAK*(xAmax-xA)/xA;
    return (sjkmax+sAK)/sAK;
  }
  
  // Inverse transforms
  // To obtain saj and sjk from Qt2 and zeta
  double TrialIFQCollA::getS1j(double Qt2, double zeta, double sAK) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getSj2(Qt2,-zeta,sAK);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getS1j() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double saj = Qt2/evNormIFpT2Sav*zeta/(zeta-1.0);
    return saj;
  }
  double TrialIFQCollA::getSj2(double Qt2, double zeta, double sAK) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getS1j(Qt2,-zeta,sAK);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getSj2() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double sjk = sAK*(zeta-1.0);
    return sjk;
  }
    
  // Trial PDF ratio
  double TrialIFQCollA::trialPDFratio(BeamParticle* beamAPtr, BeamParticle*, 
                                      int iSys, bool Rfunity,
                                      int idA, int, double eA, double, 
                                      double Qt2A, double, bool useSaved) {
    if (useSaved) return trialPDFratioSav;
    double xA = eA/(sqrt(shhSav)/2.0);
    if (idA==21) idA = 0;
    double Rpdf = 1.0;
    if (Rfunity) {
      Rpdf = pow(1.0/xA,betaQCollISav);
    } else if (useRandomX) {
      double xa     = xA*pow(0.95/xA,rndmPtr->flat());
      double top    = beamAPtr->xfISR(iSys, idA, xa, Qt2A);
      double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
      if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
      if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
      Rpdf          = top * pow(xa,betaQCollISav-1.0) /
                      bottom / pow(xA,betaQCollISav-1.0);
      Rpdf          = max(Rpdf,1.0);
    } else {
      Rpdf = 1.0;
    }
    trialPDFratioSav = Rpdf;
    return Rpdf;
  }
  
  // Physical PDF ratio
  double TrialIFQCollA::physPDFratio(BeamParticle* beamAPtr, BeamParticle*, 
                                     int iSys, bool Rfunity,
                                     int idA, int, double eA, double, 
                                     double ea, double, double Qt2A, double) { 
    double xA = eA/(sqrt(shhSav)/2.0);
    double xa = ea/(sqrt(shhSav)/2.0);
    if (Rfunity) return pow(xa/xA,betaQCollISav)*xA/xa;
    if ( !beamAPtr->insideBounds(xa, Qt2A) || !beamAPtr->insideBounds(xA, Qt2A) ) {
      nPDFoutsideBound++;
      if (nPDFoutsideBound <= nReportPDFbound)
        printf("%s::PDFratio evaluation outside boundaries \n"
               " Further warnings suppressed!", name().c_str());
    }
    double Rpdf = 1.0;
    if (idA==21) idA = 0;
    double top    = beamAPtr->xfISR(iSys, idA, xa, Qt2A);
    double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
    if (abs(top)    < TINYPDF) top    = TINYPDF;
    if (abs(bottom) < TINYPDF) bottom = TINYPDF;
    Rpdf          = top * pow(xa,betaQCollISav-1.0) /
                    bottom / pow(xA,betaQCollISav-1.0);
    return Rpdf; 
  }
  
  
  //----------------------------------------------------------------------
  // K quark collinear trial function for IF (derived base class)
  //----------------------------------------------------------------------
    
  // Trial antenna function
  double TrialIFQCollK::aTrial(double saj, double sjk, double sAK) {
    if (saj < 0. || sjk < 0.) return 0.;
    return 1.0/sAK*(sAK+sjk)/sjk;
  }
    
  // Modified trial antenna function, used when combining accept
  // probabilities for gluon emission
  double TrialIFQCollK::aTrialMod(double saj, double sjk, double sAK,
                                  double, double, double, double) {
    if (saj < 0. || sjk < 0.) return 0.;
    return 1.0/sAK*(sAK+sjk)/sjk;
  }
    
  // Physical antenna function
  double TrialIFQCollK::aPhys(double saj, double sjk, double sAK, int, int) {
    if (saj < 0. || sjk < 0.) return 0.;
    return 1.0/sAK*saj/sjk;
  }
  
  // Generate a new Q value, with first-order running alphaS
  double TrialIFQCollK::genQ2(double q2old, double sAK, double zMin, double zMax, 
                              double colFac, double PDFratio, 
                              double b0, double kR, double Lambda,
                              double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI*b0/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    double facLam = pow2(Lambda/kR) * evNormIFpT2Sav;
    return exp(pow(ran,comFac) * log(q2old/facLam)) * facLam;
  }
  
  // Generate a new Q value, with constant trial alphaS
  double TrialIFQCollK::genQ2(double q2old, double sAK, double zMin, double zMax, 
                              double colFac, double alphaSvalue, double PDFratio, 
                              double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    return q2old * pow(ran,comFac/alphaSvalue);
  }
  
  // Generate a new Q value, use in the very unordered region
  double TrialIFQCollK::genQ2withQN(double q2old, double q2hat, double N, double sAK, 
                                    double zMin, double zMax, 
                                    double colFac, double alphaSvalue, double PDFratio, 
                                    double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI/Iz/colFac/alphaSvalue/PDFratio/headroomFactor;
    double expo   = ((N-2.0)/2.0);
    double ran    = rndmPtr->flat();
    return pow( pow(q2old,expo)/(1.0-pow(q2old/q2hat,expo)*
                                     log(ran)*comFac*expo), 1.0/expo);
  }
  
  // Generate a new zeta value in [zMin,zMax]
  double TrialIFQCollK::genZ(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return -1.;
    double ran = rndmPtr->flat();
    double z = ran*(zMin-zMax)+zMax;
    return z;
  }
  
  // The zeta integral
  double TrialIFQCollK::getIz(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return 0.0;
    return (zMax - zMin); 
  }
    
  // The zeta boundaries, for a given value of the evolution variable
  double TrialIFQCollK::getZmin(double Qt2, double sAK, double eA,
                                double eAused) {
    double xA     = eA/(sqrt(shhSav)/2.0);
    double eAmax  = ( (sqrt(shhSav)/2.0) - (eAused-eA) );
    double xAmax  = eAmax/(sqrt(shhSav)/2.0);
    double sjkmax = sAK*(xAmax-xA)/xA;
    return Qt2/evNormIFpT2Sav/sjkmax;
  }
  double TrialIFQCollK::getZmax(double, double, double, double) {
    return 1.0;
  }
  
  // Inverse transforms
  // To obtain saj and sjk from Qt2 and zeta
  double TrialIFQCollK::getS1j(double Qt2, double zeta, double sAK) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getSj2(Qt2,-zeta,sAK);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getS1j() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double saj = Qt2/evNormIFpT2Sav+zeta*sAK;
    return saj;
  }
  double TrialIFQCollK::getSj2(double Qt2, double zeta, double sAK) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getS1j(Qt2,-zeta,sAK);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getSj2() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double sjk = Qt2/evNormIFpT2Sav/zeta;
    return sjk;
  }
    
  // Trial PDF ratio
  double TrialIFQCollK::trialPDFratio(BeamParticle* beamAPtr, BeamParticle*, 
                                      int iSys, bool Rfunity,
                                      int idA, int, double eA, double, 
                                      double Qt2A, double, bool useSaved) {
    if (useSaved) return trialPDFratioSav;
    double xA = eA/(sqrt(shhSav)/2.0);
    if (idA==21) idA = 0;
    double Rpdf = 1.0;
    if (Rfunity) {
      Rpdf = 1.0;
    } else if (useRandomX) {
      double xa     = xA*pow(0.95/xA,rndmPtr->flat());
      double top    = beamAPtr->xfISR(iSys, idA, xa, Qt2A);
      double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
      if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
      if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
      Rpdf          = top/xa/bottom*xA;
      Rpdf          = max(Rpdf,1.2);
    } else {
      Rpdf = 1.2;
    }
    trialPDFratioSav = Rpdf;
    return Rpdf;
  }
  
  // Physical PDF ratio
  double TrialIFQCollK::physPDFratio(BeamParticle* beamAPtr, BeamParticle*, 
                                     int iSys, bool Rfunity,
                                     int idA, int, double eA, double, 
                                     double ea, double, double Qt2A, double) {
    double xA = eA/(sqrt(shhSav)/2.0);
    double xa = ea/(sqrt(shhSav)/2.0);
    if (Rfunity) return xA/xa;
    if ( !beamAPtr->insideBounds(xa, Qt2A) || !beamAPtr->insideBounds(xA, Qt2A) ) {
      nPDFoutsideBound++;
      if (nPDFoutsideBound <= nReportPDFbound)
        printf("%s::PDFratio evaluation outside boundaries \n"
               " Further warnings suppressed!", name().c_str());
    }
    double Rpdf = 1.0;
    if (idA==21) idA = 0;
    double top    = beamAPtr->xfISR(iSys, idA, xa, Qt2A);
    double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
    if (abs(top)    < TINYPDF) top    = TINYPDF;
    if (abs(bottom) < TINYPDF) bottom = TINYPDF;
    Rpdf          = top/xa/bottom*xA;
    return Rpdf;  
  }
  
  
  //----------------------------------------------------------------------
  // A1 gluon collinear trial function for IF (derived base class)
  //----------------------------------------------------------------------
    
  // Trial antenna function
  double TrialIFGCollA1::aTrial(double saj, double sjk, double sAK) {
    if (saj < 0. || sjk < 0.) return 0.;
    return 2.0/sAK/sAK/saj*(sAK+sjk)*(sAK+sjk);
  }
    
  // Modified trial antenna function, used when combining accept
  // probabilities for gluon emission
  double TrialIFGCollA1::aTrialMod(double saj, double sjk, double sAK,
                                   double eA, double, double ea, double) {
    if (saj < 0. || sjk < 0.) return 0.;
    double xRatio = pow(eA/ea,betaGColl1ISav);
    return 2.0/sAK/sAK/saj*(sAK+sjk)*(sAK+sjk)*xRatio;
  }
    
  // Physical antenna function
  double TrialIFGCollA1::aPhys(double saj, double sjk, double sAK, int, int) {
    if (saj < 0. || sjk < 0.) return 0.;
    double sak = sAK+sjk-saj;
    return 2.0/sAK*sjk/saj*sak/sAK;
  }
  
  // Generate a new Q value, with first-order running alphaS
  double TrialIFGCollA1::genQ2(double q2old, double sAK, double zMin, double zMax, 
                               double colFac, double PDFratio, 
                               double b0, double kR, double Lambda,
                               double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 2.0*M_PI*b0/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    double facLam = pow2(Lambda/kR) * evNormIFpT2Sav;
    return exp(pow(ran,comFac) * log(q2old/facLam)) * facLam;
  }
  
  // Generate a new Q value, with constant trial alphaS
  double TrialIFGCollA1::genQ2(double q2old, double sAK, double zMin, double zMax, 
                               double colFac, double alphaSvalue, double PDFratio, 
                               double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 2.0*M_PI/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    return q2old * pow(ran,comFac/alphaSvalue);
  }
  
  // Generate a new Q value, use in the very unordered region
  double TrialIFGCollA1::genQ2withQN(double q2old, double q2hat, double N, double sAK, 
                                     double zMin, double zMax, 
                                     double colFac, double alphaSvalue, double PDFratio, 
                                     double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 2.0*M_PI/Iz/colFac/alphaSvalue/PDFratio/headroomFactor;
    double expo   = ((N-2.0)/2.0);
    double ran    = rndmPtr->flat();
    return pow( pow(q2old,expo)/(1.0-pow(q2old/q2hat,expo)*
                                     log(ran)*comFac*expo), 1.0/expo);
  }
  
  // Generate a new zeta value in [zMin,zMax]
  double TrialIFGCollA1::genZ(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return -1.;
    double ran = rndmPtr->flat();
    if (betaGColl1ISav==1.0) return zMax*pow(zMin/zMax,ran);
    return pow(ran*(pow(zMax,1.0-betaGColl1ISav)-pow(zMin,1.0-betaGColl1ISav))+
               pow(zMin,1.0-betaGColl1ISav),1.0/(1.0-betaGColl1ISav));
  }
  
  // The zeta integral
  double TrialIFGCollA1::getIz(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return 0.0;
    if (betaGColl1ISav==1.0) return log(zMax/zMin);
    return (pow(zMax,1.0-betaGColl1ISav)-pow(zMin,1.0-betaGColl1ISav))/
           (1.0-betaGColl1ISav); 
  }
    
  // The zeta boundaries, for a given value of the evolution variable
  double TrialIFGCollA1::getZmin(double Qt2, double sAK, double, double) {
    return (Qt2/evNormIFpT2Sav+sAK)/sAK;
  }
  double TrialIFGCollA1::getZmax(double, double sAK, double eA,
                                 double eAused) {
    double xA     = eA/(sqrt(shhSav)/2.0);
    double eAmax  = ( (sqrt(shhSav)/2.0) - (eAused-eA) );
    double xAmax  = eAmax/(sqrt(shhSav)/2.0);
    double sjkmax = sAK*(xAmax-xA)/xA;
    return (sjkmax+sAK)/sAK;
  }
    
  // Inverse transforms
  // To obtain saj and sjk from Qt2 and zeta
  double TrialIFGCollA1::getS1j(double Qt2, double zeta, double sAK) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getSj2(Qt2,-zeta,sAK);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getS1j() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double saj = Qt2/evNormIFpT2Sav*zeta/(zeta-1.0);
    return saj;
  }
  double TrialIFGCollA1::getSj2(double Qt2, double zeta, double sAK) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getS1j(Qt2,-zeta,sAK);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getSj2() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double sjk = sAK*(zeta-1.0);
    return sjk;
  }
    
  // Trial PDF ratio
  double TrialIFGCollA1::trialPDFratio(BeamParticle* beamAPtr, BeamParticle*, 
                                       int iSys, bool Rfunity,
                                       int idA, int, double eA, double, 
                                       double Qt2A, double, bool useSaved) {
    if (useSaved) return trialPDFratioSav;
    double xA = eA/(sqrt(shhSav)/2.0);
    if (idA==21) idA = 0;
    double Rpdf = 1.0;
    if (Rfunity) {
      Rpdf = pow(1.0/xA,betaGColl1ISav);
    } else if (useRandomX) {
      double xa     = xA*pow(0.95/xA,rndmPtr->flat());
      double top    = beamAPtr->xfISR(iSys, idA, xa, Qt2A);
      double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
      if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
      if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
      Rpdf          = top * pow(xa,betaGColl1ISav-1.0) /
                      bottom / pow(xA,betaGColl1ISav-1.0);
      Rpdf          = max(Rpdf,1.0);
    } else {
      Rpdf = 1.0;
    }
    trialPDFratioSav = Rpdf;
    return Rpdf;
  }
  
  
  // Physical PDF ratio
  double TrialIFGCollA1::physPDFratio(BeamParticle* beamAPtr, BeamParticle*, 
                                      int iSys, bool Rfunity,
                                      int idA, int, double eA, double, 
                                      double ea, double, double Qt2A, double) { 
    double xA = eA/(sqrt(shhSav)/2.0);
    double xa = ea/(sqrt(shhSav)/2.0);
    if (Rfunity) return pow(xa/xA,betaGColl1ISav)*xA/xa;
    if ( !beamAPtr->insideBounds(xa, Qt2A) || !beamAPtr->insideBounds(xA, Qt2A) ) {
      nPDFoutsideBound++;
      if (nPDFoutsideBound <= nReportPDFbound)
        printf("%s::PDFratio evaluation outside boundaries \n"
               " Further warnings suppressed!", name().c_str());
    }
    double Rpdf = 1.0;
    if (idA==21) idA = 0;
    double top    = beamAPtr->xfISR(iSys, idA, xa, Qt2A);
    double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
    if (abs(top)    < TINYPDF) top    = TINYPDF;
    if (abs(bottom) < TINYPDF) bottom = TINYPDF;
    Rpdf          = top * pow(xa,betaGColl1ISav-1.0) /
                    bottom / pow(xA,betaGColl1ISav-1.0);
    Rpdf          = top/bottom;
    return Rpdf;  
  }
  
  
  //----------------------------------------------------------------------
  // A2 gluon collinear trial function for IF (derived base class)
  //----------------------------------------------------------------------
    
  // Trial antenna function
  double TrialIFGCollA2::aTrial(double saj, double sjk, double) {
    if (saj < 0. || sjk < 0.) return 0.;
    return 2.0/saj;
  }
    
  // Modified trial antenna function, used when combining accept
  // probabilities for gluon emission
  double TrialIFGCollA2::aTrialMod(double saj, double sjk, double,
                                   double eA, double, double ea, double) {
    if (saj < 0. || sjk < 0.) return 0.;
    double xRatio = pow(eA/ea,betaGColl2ISav);
    return 2.0/saj*xRatio;
  }
    
  // Physical antenna function
  double TrialIFGCollA2::aPhys(double saj, double sjk, double sAK, int, int) {
    if (saj < 0. || sjk < 0.) return 0.;
    return 2.0*sjk/saj/(sAK+sjk);
  }
  
  // Generate a new Q value, with first-order running alphaS
  double TrialIFGCollA2::genQ2(double q2old, double sAK, double zMin, double zMax, 
                               double colFac, double PDFratio, 
                               double b0, double kR, double Lambda,
                               double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 2.0*M_PI*b0/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    double facLam = pow2(Lambda/kR) * evNormIFpT2Sav;
    return exp(pow(ran,comFac) * log(q2old/facLam)) * facLam;
  }
  
  // Generate a new Q value, with constant trial alphaS
  double TrialIFGCollA2::genQ2(double q2old, double sAK, double zMin, double zMax, 
                               double colFac, double alphaSvalue, double PDFratio, 
                               double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 2.0*M_PI/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    return q2old * pow(ran,comFac/alphaSvalue);
  }
  
  // Generate a new Q value, use in the very unordered region
  double TrialIFGCollA2::genQ2withQN(double q2old, double q2hat, double N, double sAK, 
                                     double zMin, double zMax, 
                                     double colFac, double alphaSvalue, double PDFratio, 
                                     double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 2.0*M_PI/Iz/colFac/alphaSvalue/PDFratio/headroomFactor;
    double expo   = ((N-2.0)/2.0);
    double ran    = rndmPtr->flat();
    return pow( pow(q2old,expo)/(1.0-pow(q2old/q2hat,expo)*
                                     log(ran)*comFac*expo), 1.0/expo);
  }
  
  // Generate a new zeta value in [zMin,zMax]
  double TrialIFGCollA2::genZ(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return -1.;
    double ran = rndmPtr->flat();
    if (betaGColl2ISav==-1.0) return zMax*pow(zMin/zMax,ran);
    return pow(ran*(pow(zMax,-1.0-betaGColl2ISav)-pow(zMin,-1.0-betaGColl2ISav))+
               pow(zMin,-1.0-betaGColl2ISav),1.0/(-1.0-betaGColl2ISav));
  }
  
  // The zeta integral
  double TrialIFGCollA2::getIz(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return 0.0;
    if (betaGColl2ISav==-1.0) return log(zMax/zMin);
    return (pow(zMax,-1.0-betaGColl2ISav)-pow(zMin,-1.0-betaGColl2ISav))/
           (-1.0-betaGColl2ISav);
  }
    
  // The zeta boundaries, for a given value of the evolution variable
  double TrialIFGCollA2::getZmin(double Qt2, double sAK, double, double) {
    return (Qt2/evNormIFpT2Sav+sAK)/sAK;
  }
  double TrialIFGCollA2::getZmax(double, double sAK, double eA,
                                 double eAused) {
    double xA     = eA/(sqrt(shhSav)/2.0);
    double eAmax  = ( (sqrt(shhSav)/2.0) - (eAused-eA) );
    double xAmax  = eAmax/(sqrt(shhSav)/2.0);
    double sjkmax = sAK*(xAmax-xA)/xA;
    return (sjkmax+sAK)/sAK;
  }
    
  // Inverse transforms
  // To obtain saj and sjk from Qt2 and zeta
  double TrialIFGCollA2::getS1j(double Qt2, double zeta, double sAK) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getSj2(Qt2,-zeta,sAK);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getS1j() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double saj = Qt2/evNormIFpT2Sav*zeta/(zeta-1.0);
    return saj;
  }
  double TrialIFGCollA2::getSj2(double Qt2, double zeta, double sAK) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getS1j(Qt2,-zeta,sAK);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getSj2() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double sjk = sAK*(zeta-1.0);
    return sjk;
  }
    
  // Trial PDF ratio
  double TrialIFGCollA2::trialPDFratio(BeamParticle* beamAPtr, BeamParticle*, 
                                       int iSys, bool Rfunity,
                                       int idA, int, double eA, double, 
                                       double Qt2A, double, bool useSaved) {
    if (useSaved) return trialPDFratioSav;
    double xA = eA/(sqrt(shhSav)/2.0);
    if (idA==21) idA = 0;
    double Rpdf = 1.0;
    if (Rfunity) {
      Rpdf = pow(1.0/xA,betaGColl2ISav);
    } else if (useRandomX) {
      double xa     = xA*pow(0.95/xA,rndmPtr->flat());
      double top    = beamAPtr->xfISR(iSys, idA, xa, Qt2A);
      double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
      if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
      if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
      Rpdf          = top * pow(xa,betaGColl2ISav-1.0) /
                      bottom / pow(xA,betaGColl2ISav-1.0);
      Rpdf          = max(Rpdf,1.0);
    } else {
      Rpdf = 1.0;
    }
    trialPDFratioSav = Rpdf;
    return Rpdf;
  }
  
  
  // Physical PDF ratio
  double TrialIFGCollA2::physPDFratio(BeamParticle* beamAPtr, BeamParticle*, 
                                      int iSys, bool Rfunity,
                                      int idA, int, double eA, double, 
                                      double ea, double, double Qt2A, double) { 
    double xA = eA/(sqrt(shhSav)/2.0);
    double xa = ea/(sqrt(shhSav)/2.0);
    if (Rfunity) return pow(xa/xA,betaGColl2ISav)*xA/xa;
    if ( !beamAPtr->insideBounds(xa, Qt2A) || !beamAPtr->insideBounds(xA, Qt2A) ) {
      nPDFoutsideBound++;
      if (nPDFoutsideBound <= nReportPDFbound)
        printf("%s::PDFratio evaluation outside boundaries \n"
               " Further warnings suppressed!", name().c_str());
    }
    double Rpdf = 1.0;
    if (idA==21) idA = 0;
    double top    = beamAPtr->xfISR(iSys, idA, xa, Qt2A);
    double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
    if (abs(top)    < TINYPDF) top    = TINYPDF;
    if (abs(bottom) < TINYPDF) bottom = TINYPDF;
    Rpdf          = top * pow(xa,betaGColl2ISav-1.0) /
                    bottom / pow(xA,betaGColl2ISav-1.0);
    return Rpdf;   
  }
  
  
  //----------------------------------------------------------------------
  // K gluon collinear trial function for IF (derived base class)
  //----------------------------------------------------------------------
    
  // Trial antenna function
  double TrialIFGCollK::aTrial(double saj, double sjk, double sAK) {
    if (saj < 0. || sjk < 0.) return 0.;
    return 1.0/sAK*saj/sjk*(sAK+sjk)/sAK;
  }
    
  // Modified trial antenna function, used when combining accept
  // probabilities for gluon emission
  double TrialIFGCollK::aTrialMod(double saj, double sjk, double sAK,
                                  double eA, double, double ea, double) {
    if (saj < 0. || sjk < 0.) return 0.;
    double xRatio = eA/ea;
    return 1.0/sAK*saj/sjk*(sAK+sjk)/sAK*xRatio;
  }
    
  // Physical antenna function
  double TrialIFGCollK::aPhys(double saj, double sjk, double sAK, int, int) {
    if (saj < 0. || sjk < 0.) return 0.;
    double sak = sAK+sjk-saj;
    return 1.0/sAK*saj/sjk*sak/sAK;
  }
  
  // Generate a new Q value, with first-order running alphaS
  double TrialIFGCollK::genQ2(double q2old, double sAK, double zMin, double zMax, 
                              double colFac, double PDFratio, 
                              double b0, double kR, double Lambda,
                              double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI*b0/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    double facLam = pow2(Lambda/kR) * evNormIFpT2Sav;
    return exp(pow(ran,comFac) * log(q2old/facLam)) * facLam;
  }
  
  // Generate a new Q value, with constant trial alphaS
  double TrialIFGCollK::genQ2(double q2old, double sAK, double zMin, double zMax, 
                              double colFac, double alphaSvalue, double PDFratio, 
                              double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI/Iz/colFac/PDFratio/headroomFactor;
    double ran = rndmPtr->flat();
    return q2old * pow(ran,comFac/alphaSvalue);
  }
  
  // Generate a new Q value, use in the very unordered region
  double TrialIFGCollK::genQ2withQN(double q2old, double q2hat, double N, double sAK, 
                                    double zMin, double zMax, 
                                    double colFac, double alphaSvalue, double PDFratio, 
                                    double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI/Iz/colFac/alphaSvalue/PDFratio/headroomFactor;
    double expo   = ((N-2.0)/2.0);
    double ran    = rndmPtr->flat();
    return pow( pow(q2old,expo)/(1.0-pow(q2old/q2hat,expo)*
                                     log(ran)*comFac*expo), 1.0/expo);
  }
  
  // Generate a new zeta value in [zMin,zMax]
  double TrialIFGCollK::genZ(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return -1.;
    double ran = rndmPtr->flat();
    double z = sqrt(ran*(zMin*zMin-zMax*zMax)+zMax*zMax);
    return z;
  }
  
  // The zeta integral
  double TrialIFGCollK::getIz(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return 0.0;
    return 0.5*(zMax*zMax - zMin*zMin); 
  }
    
  // The zeta boundaries, for a given value of the evolution variable
  double TrialIFGCollK::getZmin(double Qt2, double sAK, double eA,
                                double eAused) {
    double xA     = eA/(sqrt(shhSav)/2.0);
    double eAmax  = ( (sqrt(shhSav)/2.0) - (eAused-eA) );
    double xAmax  = eAmax/(sqrt(shhSav)/2.0);
    double sjkmax = sAK*(xAmax-xA)/xA;
    return Qt2/evNormIFpT2Sav/sjkmax;
  }
  double TrialIFGCollK::getZmax(double, double, double, double) {
    return 1.0;
  }
  
  // Inverse transforms
  // To obtain saj and sjk from Qt2 and zeta
  double TrialIFGCollK::getS1j(double Qt2, double zeta, double sAK) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getSj2(Qt2,-zeta,sAK);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getS1j() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double saj = Qt2/evNormIFpT2Sav+zeta*sAK;
    return saj;
  }
  double TrialIFGCollK::getSj2(double Qt2, double zeta, double sAK) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getS1j(Qt2,-zeta,sAK);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getSj2() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double sjk = Qt2/evNormIFpT2Sav/zeta;
    return sjk;
  }
    
  // Trial PDF ratio
  double TrialIFGCollK::trialPDFratio(BeamParticle* beamAPtr, BeamParticle*, 
                                      int iSys, bool Rfunity,
                                      int idA, int, double eA, double, 
                                      double Qt2A, double, bool useSaved) {
    if (useSaved) return trialPDFratioSav;
    double xA = eA/(sqrt(shhSav)/2.0);
    if (idA==21) idA = 0;
    double Rpdf = 1.0;
    if (Rfunity) {
      Rpdf = 1.0;
    } else if (useRandomX) {
      double xa     = xA*pow(0.95/xA,rndmPtr->flat());
      double top    = beamAPtr->xfISR(iSys, idA, xa, Qt2A);
      double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
      if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
      if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
      Rpdf          = top/bottom;
      Rpdf          = max(Rpdf,1.0);
    } else {
      Rpdf = 1.0;
    }
    trialPDFratioSav = Rpdf;
    return Rpdf;
  }
  
  // Physical PDF ratio
  double TrialIFGCollK::physPDFratio(BeamParticle* beamAPtr, BeamParticle*, 
                                      int iSys, bool Rfunity,
                                      int idA, int, double eA, double, 
                                      double ea, double, double Qt2A, double) { 
    double xA = eA/(sqrt(shhSav)/2.0);
    double xa = ea/(sqrt(shhSav)/2.0);
    if (Rfunity) return xA/xa;
    if ( !beamAPtr->insideBounds(xa, Qt2A) || !beamAPtr->insideBounds(xA, Qt2A) ) {
      nPDFoutsideBound++;
      if (nPDFoutsideBound <= nReportPDFbound)
        printf("%s::PDFratio evaluation outside boundaries \n"
               " Further warnings suppressed!", name().c_str());
    }
    double Rpdf = 1.0;
    if (idA==21) idA = 0;
    double top    = beamAPtr->xfISR(iSys, idA, xa, Qt2A);
    double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
    if (abs(top)    < TINYPDF) top    = TINYPDF;
    if (abs(bottom) < TINYPDF) bottom = TINYPDF;
    Rpdf          = top/bottom;
    return Rpdf;  
  }
  
  
  //----------------------------------------------------------------------
  // A splitting trial function for IF (derived base class), q->gqbar
  //----------------------------------------------------------------------
    
  // Trial antenna function
  double TrialIFSplitA::aTrial(double saj, double sjk, double sAK) {
    if (saj < 0. || sjk < 0.) return 0.;
    return 2.0/sAK*(sAK+sjk)/saj;
  }
    
  // Physical antenna function
  double TrialIFSplitA::aPhys(double saj, double sjk, double sAK, int, int) {
    if (saj < 0. || sjk < 0.) return 0.;
    double sak = sAK+sjk-saj;
    return 1.0/sAK*sak/saj-2.0/sAK*sjk/saj*(sAK-saj)/(sAK+sjk);
  }
  
  // Generate a new Q value, with first-order running alphaS
  double TrialIFSplitA::genQ2(double q2old, double sAK, double zMin, double zMax, 
                              double colFac, double PDFratio, 
                              double b0, double kR, double Lambda,
                              double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 2.0*M_PI*b0/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    double facLam = pow2(Lambda/kR) * ((useMevolSav) ? evNormMsplitISav : evNormIFpT2Sav);
    return exp(pow(ran,comFac) * log(q2old/facLam)) * facLam;
  }
  
  // Generate a new Q value, with constant trial alphaS
  double TrialIFSplitA::genQ2(double q2old, double sAK, double zMin, double zMax, 
                              double colFac, double alphaSvalue, double PDFratio, 
                              double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 2.0*M_PI/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    return q2old * pow(ran,comFac/alphaSvalue);
  }
  
  // Generate a new Q value, use in the very unordered region
  double TrialIFSplitA::genQ2withQN(double q2old, double q2hat, double N, double sAK, 
                                    double zMin, double zMax, 
                                    double colFac, double alphaSvalue, double PDFratio, 
                                    double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 2.0*M_PI/Iz/colFac/alphaSvalue/PDFratio/headroomFactor;
    double expo   = ((N-2.0)/2.0);
    double ran    = rndmPtr->flat();
    return pow( pow(q2old,expo)/(1.0-pow(q2old/q2hat,expo)*
                                     log(ran)*comFac*expo), 1.0/expo);
  }
  
  // Generate a new Q value, with running of the PDFs towards the mass threshold
  double TrialIFSplitA::genQ2(double q2old, double sAK, double zMin, double zMax, 
                              double colFac, double alphaSvalue, double PDFratio, 
                              int idA, int, double /*eA*/, double /*eB*/, bool, 
                              double headroomFactor) {
    // Use only if the user wants to get rid of c and b quarks
    // and use only in the right evolution window
    double mQ = (abs(idA) == 4 ? mcSav : mbSav);
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 2.0*M_PI/Iz/colFac/alphaSvalue/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    return ( exp(pow(ran,comFac) * 
             log(q2old/( useMevolSav ? evNormMsplitISav : evNormIFpT2Sav)/pow2(mQ))) * 
             pow2(mQ) );
  }
  
  // Generate a new zeta value in [zMin,zMax]
  double TrialIFSplitA::genZ(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return -1.;
    double ran = rndmPtr->flat();
    if (betaSplitISav==0.0) return zMax*pow(zMin/zMax,ran);
    return pow(ran*(pow(zMax,-betaSplitISav)-pow(zMin,-betaSplitISav))+
               pow(zMin,-betaSplitISav),1.0/(-betaSplitISav));
  }
  
  // The zeta integral
  double TrialIFSplitA::getIz(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return 0.0;
    if (betaSplitISav==0.0) return log(zMax/zMin);
    return (pow(zMax,-betaSplitISav)-pow(zMin,-betaSplitISav))/(-betaSplitISav);
  }
    
  // The zeta boundaries, for a given value of the evolution variable
  double TrialIFSplitA::getZmin(double Qt2, double sAK, double, double) {
    if (useMevolSav) {
      if ((Qt2/evNormMsplitISav)<sAK) return 1.0;
      else return Qt2/evNormMsplitISav/sAK;
    }
    return (Qt2/evNormIFpT2Sav+sAK)/sAK;
  }
  double TrialIFSplitA::getZmax(double, double sAK, double eA,
                                double eAused) {
    double xA     = eA/(sqrt(shhSav)/2.0);
    double eAmax  = ( (sqrt(shhSav)/2.0) - (eAused-eA) );
    double xAmax  = eAmax/(sqrt(shhSav)/2.0);
    double sjkmax = sAK*(xAmax-xA)/xA;
    if (useMevolSav) return (sjkmax+sAK)/sAK;
    return (sjkmax+sAK)/sAK;
  }
    
  // Inverse transforms
  // To obtain saj and sjk from Qt2 and zeta
  double TrialIFSplitA::getS1j(double Qt2, double zeta, double sAK) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getSj2(Qt2,-zeta,sAK);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getS1j() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double saj = Qt2/evNormIFpT2Sav*zeta/(zeta-1.0);
    if (useMevolSav) saj = Qt2/evNormMsplitISav;
    return saj;
  }
  double TrialIFSplitA::getSj2(double Qt2, double zeta, double sAK) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getS1j(Qt2,-zeta,sAK);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getSj2() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double sjk = sAK*(zeta-1.0);
    if (useMevolSav) sjk = (zeta-1.0)*sAK;
    return sjk;
  }
    
  // Trial PDF ratio
  double TrialIFSplitA::trialPDFratio(BeamParticle* beamAPtr, BeamParticle*, 
                                      int iSys, bool Rfunity,
                                      int idA, int, double eA, double, 
                                      double Qt2A, double, bool useSaved) {
    if (useSaved) return trialPDFratioSav;
    double xA = eA/(sqrt(shhSav)/2.0);
    if (idA==21) idA = 0;
    double Rpdf = 1.0;
    if (Rfunity) {
      Rpdf = pow(1.0/xA,betaSplitISav);
    } else if (useRandomX) {
      double xa     = xA*pow(0.95/xA,rndmPtr->flat());
      double top    = beamAPtr->xfISR(iSys, 0,   xa, Qt2A);
      double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
      if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
      if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
      Rpdf          = top * pow(xa,betaSplitISav-1.0) /
                      bottom / pow(xA,betaSplitISav-1.0);
    } else {
      double top    = beamAPtr->xfISR(iSys, 0,   xA, Qt2A);
      double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
      if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
      if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
      Rpdf          = 1.6*top/bottom;
    }
    trialPDFratioSav = Rpdf;
    return Rpdf;
  }
  
  // Physical PDF ratio
  double TrialIFSplitA::physPDFratio(BeamParticle* beamAPtr, BeamParticle*, 
                                     int iSys, bool Rfunity,
                                     int idA, int, double eA, double, 
                                     double ea, double, double Qt2A, double) { 
    double xA = eA/(sqrt(shhSav)/2.0);
    double xa = ea/(sqrt(shhSav)/2.0);
    if (Rfunity) return pow(xa/xA,betaSplitISav)*xA/xa;
    if ( !beamAPtr->insideBounds(xa, Qt2A) || !beamAPtr->insideBounds(xA, Qt2A) ) {
      nPDFoutsideBound++;
      if (nPDFoutsideBound <= nReportPDFbound)
        printf("%s::PDFratio evaluation outside boundaries \n"
               " Further warnings suppressed!", name().c_str());
    }
    double Rpdf = 1.0;
    if (idA==21) idA = 0;
    double top    = beamAPtr->xfISR(iSys, 0,   xa, Qt2A);
    double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
    if (abs(top)    < TINYPDF) top    = TINYPDF;
    if (abs(bottom) < TINYPDF) bottom = TINYPDF;
    Rpdf          = top * pow(xa,betaSplitISav-1.0) /
                    bottom / pow(xA,betaSplitISav-1.0);
    Rpdf          = top/bottom;
    return Rpdf; 
  }
  
  
  //----------------------------------------------------------------------
  // K splitting trial function for IF (derived base class), g->qqbar
  //----------------------------------------------------------------------
    
  // Trial antenna function
  double TrialIFSplitK::aTrial(double saj, double sjk, double sAK) {
    if (saj < 0. || sjk < 0.) return 0.;
    return 1.0/2.0/sjk*(sAK+sjk)*(sAK+sjk)/sAK/sAK;
  }
    
  // Physical antenna function
  double TrialIFSplitK::aPhys(double saj, double sjk, double sAK, int, int) {
    if (saj < 0. || sjk < 0.) return 0.;
    double sak = sAK+sjk-saj;
    return 1.0/2.0/sjk*(sak*sak+saj*saj)/sAK/sAK;
  }
  
  // Generate a new Q value, with first-order running alphaS
  double TrialIFSplitK::genQ2(double q2old, double sAK, double zMin, double zMax, 
                              double colFac, double PDFratio, 
                              double b0, double kR, double Lambda,
                              double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 8.0*M_PI*b0/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    double facLam = pow2(Lambda/kR) * ((useMevolSav) ? evNormMsplitFSav : evNormIFpT2Sav);
    return exp(pow(ran,comFac) * log(q2old/facLam)) * facLam;
  }
  
  // Generate a new Q value, with constant trial alphaS
  double TrialIFSplitK::genQ2(double q2old, double sAK, double zMin, double zMax, 
                              double colFac, double alphaSvalue, double PDFratio, 
                              double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 8.0*M_PI/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    return q2old * pow(ran,comFac/alphaSvalue);
  }
  
  // Generate a new Q value, use in the very unordered region
  double TrialIFSplitK::genQ2withQN(double q2old, double q2hat, double N, double sAK, 
                                    double zMin, double zMax, 
                                    double colFac, double alphaSvalue, double PDFratio, 
                                    double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 8.0*M_PI/Iz/colFac/alphaSvalue/PDFratio/headroomFactor;
    double expo   = ((N-2.0)/2.0);
    double ran    = rndmPtr->flat();
    return pow( pow(q2old,expo)/(1.0-pow(q2old/q2hat,expo)*
                                     log(ran)*comFac*expo), 1.0/expo);
  }
  
  // Generate a new zeta value in [zMin,zMax]
  double TrialIFSplitK::genZ(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return -1.;
    double ran = rndmPtr->flat();
    double z = ran*(zMin-zMax)+zMax;
    return z;
  }
  
  // The zeta integral
  double TrialIFSplitK::getIz(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return 0.0;
    return (zMax - zMin); 
  }
    
  // The zeta boundaries, for a given value of the evolution variable
  double TrialIFSplitK::getZmin(double Qt2, double sAK, double eA,
                                double eAused) {
    if (useMevolSav) return 0.0;
    double xA     = eA/(sqrt(shhSav)/2.0);
    double eAmax  = ( (sqrt(shhSav)/2.0) - (eAused-eA) );
    double xAmax  = eAmax/(sqrt(shhSav)/2.0);
    double sjkmax = sAK*(xAmax-xA)/xA;
    return Qt2/evNormIFpT2Sav/sjkmax;
  }
  double TrialIFSplitK::getZmax(double, double, double, double) {
    return 1.0;
  }
    
  // Inverse transforms
  // To obtain saj and sjk from Qt2 and zeta
  double TrialIFSplitK::getS1j(double Qt2, double zeta, double sAK) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getSj2(Qt2,-zeta,sAK);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getS1j() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double saj = Qt2/evNormIFpT2Sav+zeta*sAK;
    if (useMevolSav) saj = zeta*(sAK+Qt2/evNormMsplitFSav);
    return saj;
  }
  double TrialIFSplitK::getSj2(double Qt2, double zeta, double sAK) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getS1j(Qt2,-zeta,sAK);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getSj2() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double sjk = Qt2/evNormIFpT2Sav/zeta;
    if (useMevolSav) sjk = Qt2/evNormMsplitFSav;
    return sjk;
  }
    
  // Trial PDF ratio
  double TrialIFSplitK::trialPDFratio(BeamParticle* beamAPtr, BeamParticle*, 
                                      int iSys, bool Rfunity,
                                      int idA, int, double eA, double, 
                                      double Qt2A, double, bool useSaved) {
    if (useSaved) return trialPDFratioSav;
    double xA = eA/(sqrt(shhSav)/2.0);
    if (idA==21) idA = 0;
    double Rpdf = 1.0;
    if (Rfunity) {
      Rpdf = 1.0;
    } else if (useRandomX) {
      double xa     = xA*pow(0.95/xA,rndmPtr->flat());
      double top    = beamAPtr->xfISR(iSys, idA, xa, Qt2A);
      double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
      if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
      if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
      Rpdf          = top/bottom;
    } else {
      Rpdf = 1.8;
    }
    trialPDFratioSav = Rpdf;
    return Rpdf;
  }
  
  
  // Physical PDF ratio
  double TrialIFSplitK::physPDFratio(BeamParticle* beamAPtr, BeamParticle*, 
                                     int iSys, bool Rfunity,
                                     int idA, int, double eA, double, 
                                     double ea, double, double Qt2A, double) {
    double xA = eA/(sqrt(shhSav)/2.0);
    double xa = ea/(sqrt(shhSav)/2.0);
    if (Rfunity) return xA/xa;
    if ( !beamAPtr->insideBounds(xa, Qt2A) || !beamAPtr->insideBounds(xA, Qt2A) ) {
      nPDFoutsideBound++;
      if (nPDFoutsideBound <= nReportPDFbound)
        printf("%s::PDFratio evaluation outside boundaries \n"
               " Further warnings suppressed!", name().c_str());
    }
    double Rpdf = 1.0;
    if (idA==21) idA = 0;
    double top    = beamAPtr->xfISR(iSys, idA, xa, Qt2A);
    double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
    if (abs(top)    < TINYPDF) top    = TINYPDF;
    if (abs(bottom) < TINYPDF) bottom = TINYPDF;
    Rpdf          = top/bottom;
    return Rpdf;  
  }
  
  
  //----------------------------------------------------------------------
  // A conversion trial function for IF (derived base class), g->qqbar
  //----------------------------------------------------------------------
    
  // Trial antenna function
  double TrialIFConvA::aTrial(double saj, double sjk, double sAK) {
    if (saj < 0. || sjk < 0.) return 0.;
    return 1.0/saj*(sAK+sjk)*(sAK+sjk)/sAK/sAK;
  }
    
  // Physical antenna function
  double TrialIFConvA::aPhys(double saj, double sjk, double sAK, int, int) {
    if (saj < 0. || sjk < 0.) return 0.;
    double sak = sAK+sjk-saj;
    return 1.0/2.0/saj*(sak*sak+sjk*sjk)/sAK/sAK;
  }
  
  // Generate a new Q value, with first-order running alphaS
  double TrialIFConvA::genQ2(double q2old, double sAK, double zMin, double zMax, 
                             double colFac, double PDFratio, 
                             double b0, double kR, double Lambda,
                             double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI*b0/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    double facLam = pow2(Lambda/kR) * ((useMevolSav) ? evNormMconvISav : evNormIFpT2Sav);
    return exp(pow(ran,comFac) * log(q2old/facLam)) * facLam;
  }
  
  // Generate a new Q value, with constant trial alphaS
  double TrialIFConvA::genQ2(double q2old, double sAK, double zMin, double zMax, 
                             double colFac, double alphaSvalue, double PDFratio, 
                             double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI/Iz/colFac/PDFratio/headroomFactor;
    double ran    = rndmPtr->flat();
    return q2old * pow(ran,comFac/alphaSvalue);
  }
  
  // Generate a new Q value, use in the very unordered region
  double TrialIFConvA::genQ2withQN(double q2old, double q2hat, double N, double sAK, 
                                   double zMin, double zMax, 
                                   double colFac, double alphaSvalue, double PDFratio, 
                                   double /*eA*/, double /*eB*/, double headroomFactor) {
    //double xA = eA/(sqrt(shhSav)/2.0);
    //double xB = eB/(sqrt(shhSav)/2.0);
    // Sanity checks
    if (!checkInit()) return 0.0;
    if (sAK < 0. || q2old < 0.) return 0.0;
    // Generate new trial scale
    double Iz     = getIz(zMin,zMax);
    double comFac = 4.0*M_PI/Iz/colFac/alphaSvalue/PDFratio/headroomFactor;
    double expo   = ((N-2.0)/2.0);
    double ran    = rndmPtr->flat();
    return pow( pow(q2old,expo)/(1.0-pow(q2old/q2hat,expo)*
                                     log(ran)*comFac*expo), 1.0/expo);
  }
  
  // Generate a new zeta value in [zMin,zMax]
  double TrialIFConvA::genZ(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return -1.;
    double ran = rndmPtr->flat();
    if (betaConvISav==1.0) return zMax*pow(zMin/zMax,ran);
    return pow(ran*(pow(zMax,1.0-betaConvISav)-pow(zMin,1.0-betaConvISav))+
               pow(zMin,1.0-betaConvISav),1.0/(1.0-betaConvISav));
  }
  
  // The zeta integral
  double TrialIFConvA::getIz(double zMin, double zMax) {
    if (zMin > zMax || zMin < 0.) return 0.0;
    if (betaConvISav==1.0) return log(zMax/zMin);
    return (pow(zMax,1.0-betaConvISav)-pow(zMin,1.0-betaConvISav))/(1.0-betaConvISav); 
  }
    
  // The zeta boundaries, for a given value of the evolution variable
  double TrialIFConvA::getZmin(double Qt2, double sAK, double, double) {
    if (useMevolSav) {
      if ((Qt2/evNormMconvISav)<sAK) return 1.0;
      else return Qt2/evNormMconvISav/sAK;
    }
    return (Qt2/evNormIFpT2Sav+sAK)/sAK;
  }
  double TrialIFConvA::getZmax(double, double sAK, double eA,
                               double eAused) {
    double xA     = eA/(sqrt(shhSav)/2.0);
    double eAmax  = ( (sqrt(shhSav)/2.0) - (eAused-eA) );
    double xAmax  = eAmax/(sqrt(shhSav)/2.0);
    double sjkmax = sAK*(xAmax-xA)/xA;
    return (sjkmax+sAK)/sAK;
  }
    
  // Inverse transforms
  // To obtain saj and sjk from Qt2 and zeta
  double TrialIFConvA::getS1j(double Qt2, double zeta, double sAK) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getSj2(Qt2,-zeta,sAK);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getS1j() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double saj = Qt2/evNormIFpT2Sav*zeta/(zeta-1.0);
    if (useMevolSav) saj = Qt2/evNormMconvISav;
    return saj;
  }
  double TrialIFConvA::getSj2(double Qt2, double zeta, double sAK) {
    // If zeta < 0, swap invariants
    if (zeta < 0) return getS1j(Qt2,-zeta,sAK);
    // Sanity check
    if (Qt2 < 0. || zeta <= 0.) {
      printf("%s::getSj2() Error! Out of range \n", name().c_str());
      return 0.;
    }
    double sjk = sAK*(zeta-1.0);
    if (useMevolSav) sjk = (zeta-1.0)*sAK;
    return sjk;
  }
    
  // Trial PDF ratio
  double TrialIFConvA::trialPDFratio(BeamParticle* beamAPtr, BeamParticle*, 
                                     int iSys, bool Rfunity,
                                     int idA, int, double eA, double, 
                                     double Qt2A, double, bool useSaved) {
    if (useSaved) return trialPDFratioSav;
    double xA = eA/(sqrt(shhSav)/2.0);
    if (idA==21) idA = 0;
    double Rpdf = 1.0;
    if (Rfunity) {
      Rpdf = pow(1.0/xA,betaConvISav);
    } else {
      int nQuark = 5;
           if (Qt2A <= 4.0*mcSav*mcSav) nQuark = 3;
      else if (Qt2A <= 4.0*mbSav*mbSav) nQuark = 4;
      if (nQuark > nGtoQISRSav) nQuark = nGtoQISRSav;
      double xa  = (useRandomX ? (xA*pow(0.95/xA,rndmPtr->flat())) : xA);
      double top = 0.0;
      for (int i=1; i<=nQuark; i++) 
        top += (beamAPtr->xfISR(iSys, i, xa, Qt2A) + beamAPtr->xfISR(iSys, -i, xa, Qt2A));
      double bottom = beamAPtr->xfISR(iSys, 0, xA, Qt2A);
      if (abs(top)    < TINYPDFtrial) top    = TINYPDFtrial;
      if (abs(bottom) < TINYPDFtrial) bottom = TINYPDFtrial;
      Rpdf = top * pow(xa,betaConvISav-1.0) /
             bottom / pow(xA,betaConvISav-1.0);
    }
    trialPDFratioSav = Rpdf;
    return Rpdf;
  }
  
  
  // Physical PDF ratio
  double TrialIFConvA::physPDFratio(BeamParticle* beamAPtr, BeamParticle*, 
                                    int iSys, bool Rfunity,
                                    int idA, int, double eA, double, 
                                    double ea, double, double Qt2A, double) {
    double xA = eA/(sqrt(shhSav)/2.0);
    double xa = ea/(sqrt(shhSav)/2.0);
    int nQuark = 5;
         if (Qt2A <= 4.0*mcSav*mcSav) nQuark = 3;
    else if (Qt2A <= 4.0*mbSav*mbSav) nQuark = 4;
    if (nQuark > nGtoQISRSav) nQuark = nGtoQISRSav;
    if (Rfunity) {
      gConvIDSav = (rndmPtr->flat() > 0.5 ? -1.0 : 1.0) * 
                   ( int(rndmPtr->flat() * nQuark) + 1 );
      return pow(xa/xA,betaConvISav)*xA/xa;
    }
    if ( !beamAPtr->insideBounds(xa, Qt2A) || !beamAPtr->insideBounds(xA, Qt2A) ) {
      nPDFoutsideBound++;
      if (nPDFoutsideBound <= nReportPDFbound)
        printf("%s::PDFratio evaluation outside boundaries \n"
               " Further warnings suppressed!", name().c_str());
    }
    double Rpdf = 1.0;
    if (idA==21) idA = 0;
    double top = 0.0;
    map<int,double> pdfValues;
    for (int i=1; i<=nQuark; i++) {
      pdfValues[i]  = beamAPtr->xfISR(iSys,  i, xa, Qt2A);
      pdfValues[-i] = beamAPtr->xfISR(iSys, -i, xa, Qt2A);
      top           = top + pdfValues[i] + pdfValues[-i];
    }
    double bottom = beamAPtr->xfISR(iSys, idA, xA, Qt2A);
    if (abs(top)    < TINYPDF) top    = TINYPDF;
    if (abs(bottom) < TINYPDF) bottom = TINYPDF;
    Rpdf = top * pow(xa,betaConvISav-1.0) /
           bottom / pow(xA,betaConvISav-1.0);
    // Normalize pdf to sum = unity
    vector<double> pdfNormed;
    vector<int> flavour;
    pdfNormed.push_back(0.0);
    flavour.push_back(0);
    double count = 0.0;
    for (map<int,double>::iterator it=pdfValues.begin(); it!=pdfValues.end(); ++it) {
      count = count + it->second/top;
      pdfNormed.push_back(count);
      flavour.push_back(it->first);
    }
    // Random number to decide for flavour
    double ran = rndmPtr->flat();
    for (int i=1; i<(int)pdfNormed.size(); i++)
      if ( (ran>=pdfNormed[i-1]) && (ran<pdfNormed[i]) )
        gConvIDSav = flavour[i];
    return Rpdf;
  }
  
}
