// VinciaCommon.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2016 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for VinciaCommon.

#include "Vincia/VinciaCommon.h"

namespace Vincia {

//**********************************************************************

// VinciaCommon

//--------------------------------------------------------------------

// init()

bool VinciaCommon::init() {

  // Check initPtr
  if (!isInitPtr) {
    if (verbose >= 1) 
      printOut("VinciaCommon::init","Error! pointers not initialized"); 
    return false;
  }
  
  // Verbose Level
  verbose            = settingsPtr->mode("Vincia:verbose");
  
  // Checks
  epTolErr           = settingsPtr->parm("Check:epTolErr");
  epTolWarn          = settingsPtr->parm("Check:epTolWarn");
  mTolErr            = settingsPtr->parm("Check:mTolErr");
  mTolWarn           = settingsPtr->parm("Check:mTolWarn");
  
  // Counters
  nUnkownPDG    = 0;
  nIncorrectCol = 0;
  nNAN          = 0;
  nVertex       = 0;
  nChargeCons   = 0;
  nMotDau       = 0;
  nUnmatchedMass.resize(2);
  nEPcons.resize(2);
  for (int i=0; i<2; i++) {
    nUnmatchedMass[i] = 0;
    nEPcons[i]        = 0;
  }
  
  // Quark masses
  mt                 = particleDataPtr->m0(6);
  if (mt < TINY) mt  = 171.0;
  mb                 = min(mt,particleDataPtr->m0(5));
  if (mb < TINY) mb  = min(mt,4.8);
  mc                 = min(mb,particleDataPtr->m0(4));
  if (mc < TINY) mc  = min(mb,1.5);
  ms                 = min(mc,particleDataPtr->m0(3));
  if (ms < TINY) ms  = min(mc,0.1);
  
  
  // user alphaS, with and without CMW
  double alphaSvalue  = settingsPtr->parm("Vincia:alphaSvalue");
  int    alphaSorder  = settingsPtr->mode("Vincia:alphaSorder");
  int    alphaSnfmax  = settingsPtr->mode("Vincia:alphaSnfmax");
  bool doCMW = true;
  if ( (settingsPtr->mode("Vincia:CMWtypeFF") == 0) &&
       (settingsPtr->mode("Vincia:CMWtypeIF") == 0) &&
       (settingsPtr->mode("Vincia:CMWtypeII") == 0) ) doCMW = false;
  alphaStrong.init(    alphaSvalue, alphaSorder, alphaSnfmax, false);
  alphaStrongCMW.init( alphaSvalue, alphaSorder, alphaSnfmax, doCMW);
  
  // default alphaS, with and without CMW
  alphaSvalue  = settingsPtr->parmDefault("Vincia:alphaSvalue");
  alphaSorder  = settingsPtr->modeDefault("Vincia:alphaSorder");
  alphaSnfmax  = settingsPtr->modeDefault("Vincia:alphaSnfmax");
  doCMW = true;
  if ( (settingsPtr->modeDefault("Vincia:CMWtypeFF") == 0) &&
       (settingsPtr->modeDefault("Vincia:CMWtypeIF") == 0) &&
       (settingsPtr->modeDefault("Vincia:CMWtypeII") == 0) ) doCMW = false;
  alphaStrongDef.init(    alphaSvalue, alphaSorder, alphaSnfmax, false);
  alphaStrongDefCMW.init( alphaSvalue, alphaSorder, alphaSnfmax, doCMW);
  
  // Return
  isInit = true;
  return true;
}
  
// Function to find the lowest meson mass for a parton pair
// treating gluons as down quarks. Used to determine hadronisation
// boundaries consistent with assumption of string length > 1 meson.
double VinciaCommon::mHadMin(const int id1in, const int id2in) {
  int id1 = abs(id1in);
  if (id1 == 21 || id1 <= 2) id1 = 1;
  int id2 = abs(id2in);
  if (id2 == 21 || id1 <= 2) id2 = 1;
  int idMes = max(id1,id2)*100 + min(id1,id2)*10 + 1;
  // Special for ssbar, use eta rather than eta'.
  if (idMes == 331) idMes = 221;
  return particleDataPtr->m0(idMes);
}

// Function to check the event after each branching,
// mostly copied from Pythia8.
void VinciaCommon::showerChecks(Event& event, bool ISR) {
  
  // Only for verbose >= 4
  if (verbose < 4) return;

  // Incoming partons = beam daughters counted with negative momentum and charge.
  Vec4 pSum;
  double chargeSum = 0.0;
  for (int i = 0; i < event.size(); ++i) if (!event[i].isFinal()) {
    if ( (event[i].mother1() == 1) || (event[i].mother1() == 2) ) {
      pSum      -= event[i].p();
      chargeSum -= event[i].charge();
    }
  }
  double eLab = abs(pSum.e());

  // Loop over particles in the event.
  for (int i = 0; i < event.size(); ++i) {

    // Look for any unrecognized particle codes.
    int id = event[i].id();
    if (id == 0 || !particleDataPtr->isParticle(id)) {
      nUnkownPDG++;
      if (nUnkownPDG == 1)
        cout << "ERROR in Vincia::ShowerChecks" << (ISR ? "(ISR)" : "(FSR)") 
             << ": unknown particle code"
             << ", i = " << i << ", id = " << id << endl;
    }

    // Check that colour assignments are the expected ones.
    else {
      int colType = event[i].colType();
      int col     = event[i].col();
      int acol    = event[i].acol();
      if (    (colType ==  0 && (col  > 0 || acol  > 0))
           || (colType ==  1 && (col <= 0 || acol  > 0))
           || (colType == -1 && (col  > 0 || acol <= 0))
           || (colType ==  2 && (col <= 0 || acol <= 0)) ) {
        nIncorrectCol++;
        if (nIncorrectCol == 1)
          cout << "ERROR in Vincia::ShowerChecks" << (ISR ? "(ISR)" : "(FSR)") 
               << ": incorrect colours"
               << ", i = " << i << ", id = " << id << " cols = " << col
               << " " << acol << endl;
      }
    }

    // Look for particles with mismatched or not-a-number energy/momentum/mass.
    if (    abs(event[i].px()) >= 0.0 && abs(event[i].py()) >= 0.0
         && abs(event[i].pz()) >= 0.0 && abs(event[i].e())  >= 0.0
         && abs(event[i].m())  >= 0.0 ) {
      double errMass  = abs(event[i].mCalc() - event[i].m()) / max( 1.0, event[i].e());
      if (errMass > mTolErr) {
        nUnmatchedMass[0]++;
        if (nUnmatchedMass[0] == 1)
          cout << "ERROR in Vincia::ShowerChecks" << (ISR ? "(ISR)" : "(FSR)") 
               << ": unmatched particle energy/momentum/mass"
               << ", i = " << i << ", id = " << id << endl;
      } else if (errMass > mTolWarn) {
        nUnmatchedMass[1]++;
        if (nUnmatchedMass[1] == 1)
          cout << "WARNING in Vincia::ShowerChecks" << (ISR ? "(ISR)" : "(FSR)") 
               << ": not quite matched particle energy/momentum/mass"
               << ", i = " << i << ", id = " << id << endl;
      }
    } else {
      nNAN++;
      if (nNAN == 1)
        cout << "ERROR in Vincia::ShowerChecks" << (ISR ? "(ISR)" : "(FSR)") 
             << ": not-a-number energy/momentum/mass"
             << ", i = " << i << ", id = " << id << endl;
    }

    // Look for particles with not-a-number vertex/lifetime.
    if (    abs(event[i].xProd()) >= 0.0 && abs(event[i].yProd()) >= 0.0
         && abs(event[i].zProd()) >= 0.0 && abs(event[i].tProd()) >= 0.0
         && abs(event[i].tau())   >= 0.0 ) ;
    else {
      nVertex++;
      if (nVertex == 1)
        cout << "ERROR in Vincia::ShowerChecks" << (ISR ? "(ISR)" : "(FSR)") 
             << ": not-a-number vertex/lifetime"
             << ", i = " << i << ", id = " << id << endl;
    }

    // Add final-state four-momentum and charge.
    if (event[i].isFinal()) {
      pSum      += event[i].p();
      chargeSum += event[i].charge();
    }

  // End of particle loop.
  }

  // Check energy-momentum/charge conservation.
  double epDev = abs( pSum.e()) + abs(pSum.px()) + abs(pSum.py())
                                + abs(pSum.pz() );
  if (epDev > epTolErr * eLab) {
    nEPcons[0]++;
    if (nEPcons[0] == 1)
      cout << "ERROR in Vincia::ShowerChecks" << (ISR ? "(ISR)" : "(FSR)") 
           << ": energy-momentum not conserved" << endl;
  }
  else if (epDev > epTolWarn * eLab) {
    nEPcons[1]++;
    if (nEPcons[1] == 1)
      cout << "WARNING in Vincia::ShowerChecks" << (ISR ? "(ISR)" : "(FSR)") 
           << ": energy-momentum not quite conserved" << endl;
  }
  if (abs(chargeSum) > 0.1) {
    nChargeCons++;
    if (nChargeCons == 1)
      cout << "ERROR in Vincia::ShowerChecks" << (ISR ? "(ISR)" : "(FSR)") 
           << ": charge not conserved" << endl;
  }

  // Check that mother and daughter information match for each particle.
  vector<int> noMot;
  vector<int> noDau;
  vector< pair<int,int> > noMotDau;

  // Loop through the event and check that there are beam particles.
  bool hasBeams = false;
  for (int i = 0; i < event.size(); ++i) {
    int status = event[i].status();
    if (abs(status) == 12) hasBeams = true;

    // Check that mother and daughter lists not empty where not expected to.
    vector<int> mList = event[i].motherList();
    vector<int> dList = event[i].daughterList();
    if (mList.size() == 0 && abs(status) != 11 && abs(status) != 12)
      noMot.push_back(i);
    if (dList.size() == 0 && status < 0 && status != -11)
      noDau.push_back(i);

    // Check that the particle appears in the daughters list of each mother.
    for (int j = 0; j < int(mList.size()); ++j) {
      if (    event[mList[j]].daughter1() <= i
           && event[mList[j]].daughter2() >= i ) continue;
      vector<int> dmList = event[mList[j]].daughterList();
      bool foundMatch = false;
      for (int k = 0; k < int(dmList.size()); ++k)
      if (dmList[k] == i) {
        foundMatch = true;
        break;
      }
      if (!hasBeams && mList.size() == 1 && mList[0] == 0) foundMatch = true;
      if (!foundMatch) {
        bool oldPair = false;
        for (int k = 0; k < int(noMotDau.size()); ++k)
        if (noMotDau[k].first == mList[j] && noMotDau[k].second == i) {
          oldPair = true;
          break;
        }
        if (!oldPair) noMotDau.push_back( make_pair( mList[j], i) );
      }
    }

    // Check that the particle appears in the mothers list of each daughter.
    for (int j = 0; j < int(dList.size()); ++j) {
      if (    event[dList[j]].statusAbs() > 80
           && event[dList[j]].statusAbs() < 90
           && event[dList[j]].mother1() <= i
           && event[dList[j]].mother2() >= i ) continue;
      vector<int> mdList = event[dList[j]].motherList();
      bool foundMatch = false;
      for (int k = 0; k < int(mdList.size()); ++k)
      if (mdList[k] == i) {
        foundMatch = true;
        break;
      }
      if (!foundMatch) {
        bool oldPair = false;
        for (int k = 0; k < int(noMotDau.size()); ++k)
        if (noMotDau[k].first == i && noMotDau[k].second == dList[j]) {
          oldPair = true;
          break;
        }
        if (!oldPair) noMotDau.push_back( make_pair( i, dList[j]) );
      }
    }
  // End loop through the event.
  }

  // Warn if any errors were found.
  if (noMot.size() > 0 || noDau.size() > 0 || noMotDau.size() > 0) {
    nMotDau++;
    if (nMotDau == 1) {
      cout << "ERROR in Vincia::ShowerChecks" << (ISR ? "(ISR)" : "(FSR)") 
           << ": mismatch in daughter and mother lists" << endl;
      // Print some more info.
      if (noMot.size() > 0) {
        cout << " missing mothers for particles ";
        for (int i = 0; i < int(noMot.size()); ++i) cout << noMot[i] << " ";
        cout << endl;
      }
      if (noDau.size() > 0) {
        cout << " missing daughters for particles ";
        for (int i = 0; i < int(noDau.size()); ++i) cout << noDau[i] << " ";
        cout << endl;
      }
      if (noMotDau.size() > 0) {
        cout << " inconsistent history for (mother,daughter) pairs ";
        for (int i = 0; i < int(noMotDau.size()); ++i)
          cout << "(" << noMotDau[i].first << "," << noMotDau[i].second << ") ";
        cout << endl;
      }
    }
  }

  // Done.
}
  
//**********************************************************************

// VINCIA Auxiliary helper functions

string num2str(int i, int width) {
      ostringstream tmp;
      if (width <= 1) tmp<<i;
      else if (abs(i) < pow(10.0,width-1) ||
               ( i > 0 && i < pow(10.0,width) )) {
        tmp<<fixed<<setw(width)<<i;
      }
      else {
        string ab="k";
        double r=i;
        if (abs(i) < 1e5) {
          r/=1e3;
        } else if (abs(i) < 1e8) {
          r/=1e6;
          ab="M";
        } else if (abs(i) < 1e11) {
          r/=1e9;
          ab="G";
        } else if (abs(i) < 1e14) {
          ab="T";
          r/=1e12;
        }
        tmp<<fixed<<setw(width-1)<<(r > 10 ? setprecision(width-4) : setprecision(width-3))
           <<r<<ab;
      }
      return tmp.str();
}
string num2str(double r,int width) {
      ostringstream tmp;
      if (width <= 0) tmp<<r;
      else if (r == 0.0 || (abs(r) > 0.1 && abs(r)<pow(10.,max(width-3,1)))
               || width <= 8) { 
        tmp<<fixed<<setw(max(width,3))<<setprecision(min(3,max(1,width-2)))<<r; 
      } else { 
        tmp<<scientific<<setprecision(max(2,width-7))<<setw(max(9,width))<<r; 
      }
      return tmp.str();
}
  string bool2str(bool b, int width) {
    string tmp="";
    if (b) tmp="on";
    else tmp="off";
    int nPad = width-tmp.length();
    for (int i=1; i<=nPad; ++i) tmp = " "+tmp;        
    return tmp;
}
void printErr(string place, string message) {
  cout.setf(ios::internal);
  cerr<<" ("<<place<<") "<<message<<"\n"; 
}
void printOut(string place, string message) {
  cout.setf(ios::internal);
  cout<<" ("<<(place+") ")<<message<<"\n"; 
}

//*********

// Natural cubic spline with vanishing 2nd order derivatives at boundaries.
// Given vectors x_i and y_i=f(x_i), the vector of 2nd derivs, y2 is returned.

void makeSpline(vector<double>& x, vector<double>& y, vector<double>& y2)
{
  int npoints(x.size());
  vector<double> u(npoints);
  y2[0]=u[0]=0.0;
  //Decomposition loop of tridiagonal alg. y2 and u used for tmp storage of 
  //decomposed factors.
  for (int i=1;i<npoints-1;i++ ) {
    double sig=(x[i]-x[i-1])/(x[i+1]-x[i-1]);
    double p=sig*y2[i-1]+2.0;
    y2[i]=(sig-1.0)/p;
    u[i]=(y[i+1]-y[i])/(x[i+1]-x[i])-(y[i]-y[i-1])/(x[i]-x[i-1]);
    u[i]=(6.0*u[i]/(x[i+1]-x[i-1])-sig*u[i-1])/p;
  }
  y2[npoints-1]=0.0;
  for (int k=npoints-2;k>=0;k--) y2[k]=y2[k]*y2[k+1]+u[k];  
  return;
}

//Evaulation of spline, with values x and y, and splined 2nd der y2
double getSpline(const double x, vector<double>& xa, 
                 vector<double>& ya, vector<double>& y2a) {
  int npoints(xa.size());
  int k;
  double h,b,a;
  int klo=0;
  int khi=npoints-1;
  while (khi-klo > 1) {
    k=(khi+klo)>>1;
    if ( xa[k] > x ) khi=k;
    else klo=k;
  }
  h = xa[khi] - xa[klo];
  if ( h==0 ) cerr << " (getSpline:) Error: divide by zero\n" 
                   <<" khi ="<<khi<<" klo ="<<klo<<" : "
                   <<xa[khi]<<" - "<<xa[klo]<<endl;
  a = ( xa[khi] - x ) / h;
  b = ( x - xa[klo] ) / h;
  double y=a*ya[klo]+b*ya[khi]+((a*a*a-a)*y2a[klo]+(b*b*b-b)*y2a[khi])*(h*h)/6.0;
  return y;
}

//**************************************************************************

// External auxiliaries: extra four-products
  double m(const Vec4& v) {double s=m2(v); return (s >= 0)? sqrt(s):-sqrt(-s);}

double m2(const Vec4& v) {
  return pow2(v.e())-pow2(v.px())-pow2(v.py())-pow2(v.pz());}

double m2(const Vec4& v1, const Vec4& v2, const Vec4& v3) {
  double m2 = pow2(v1.e() + v2.e() + v3.e()) 
    - pow2(v1.px() + v2.px() + v3.px())
    - pow2(v1.py() + v2.py() + v3.py()) 
    - pow2(v1.pz() + v2.pz() + v3.pz());
  return m2; }  

double m2(const Vec4& v1, const Vec4& v2, 
                  const Vec4& v3, const Vec4& v4) {
  double m2 = pow2(v1.e() + v2.e() + v3.e() + v4.e()) 
    - pow2(v1.px() + v2.px() + v3.px() + v4.px())
    - pow2(v1.py() + v2.py() + v3.py() + v4.py()) 
    - pow2(v1.pz() + v2.pz() + v3.pz() + v4.pz());
  return m2; }  

double m2(const Particle& p1, const Particle& p2, const Particle& p3) {
  return m2(p1.p(),p2.p(),p3.p());}

double dot4(const Vec4& v1, const Vec4& v2) {
  return v1.e()*v2.e() - dot3(v1,v2);
}

double dot4(const Particle& p1, const Particle& p2) {
  return dot4(p1.p(), p2.p());
}

// ******

// Convert Vec4 to primitive double array, for passing to Fortran.

void Vec4toD4(const Vec4& v, double* p) {
  p[0]=v.px();
  p[1]=v.py();
  p[2]=v.pz();
  p[3]=v.e();
}

// MadGraph format, with energy first

void Vec4toMG(const Vec4& v, double* p) {
  p[0]=v.e();
  p[1]=v.px();
  p[2]=v.py();
  p[3]=v.pz();
}
  
void Vec4toMG(const vector<Vec4>& v, int i0, int i1, int i2,
              double p[4][4]) {
  p[0][0]=v[i0].e();
  p[0][1]=v[i0].px();
  p[0][2]=v[i0].py();
  p[0][3]=v[i0].pz();
  p[1][0]=v[i1].e();
  p[1][1]=v[i1].px();
  p[1][2]=v[i1].py();
  p[1][3]=v[i1].pz();
  p[2][0]=v[i2].e();
  p[2][1]=v[i2].px();
  p[2][2]=v[i2].py();
  p[2][3]=v[i2].pz();
  p[3][0]=v[3].e();
  p[3][1]=v[3].px();
  p[3][2]=v[3].py();
  p[3][3]=v[3].pz();
}
  
void Vec4toMG(const vector<Vec4>& v, int i0, int i1, int i2, int i3, 
              double p[5][4]) {
  p[0][0]=v[i0].e();
  p[0][1]=v[i0].px();
  p[0][2]=v[i0].py();
  p[0][3]=v[i0].pz();
  p[1][0]=v[i1].e();
  p[1][1]=v[i1].px();
  p[1][2]=v[i1].py();
  p[1][3]=v[i1].pz();
  p[2][0]=v[i2].e();
  p[2][1]=v[i2].px();
  p[2][2]=v[i2].py();
  p[2][3]=v[i2].pz();
  p[3][0]=v[i3].e();
  p[3][1]=v[i3].px();
  p[3][2]=v[i3].py();
  p[3][3]=v[i3].pz();
  p[4][0]=v[4].e();
  p[4][1]=v[4].px();
  p[4][2]=v[4].py();
  p[4][3]=v[4].pz();
}
  
void Vec4toMG(const vector<Vec4>& v, int i0, int i1, int i2, int i3, int i4, 
              double p[6][4]) {
  p[0][0]=v[i0].e();
  p[0][1]=v[i0].px();
  p[0][2]=v[i0].py();
  p[0][3]=v[i0].pz();
  p[1][0]=v[i1].e();
  p[1][1]=v[i1].px();
  p[1][2]=v[i1].py();
  p[1][3]=v[i1].pz();
  p[2][0]=v[i2].e();
  p[2][1]=v[i2].px();
  p[2][2]=v[i2].py();
  p[2][3]=v[i2].pz();
  p[3][0]=v[i3].e();
  p[3][1]=v[i3].px();
  p[3][2]=v[i3].py();
  p[3][3]=v[i3].pz();
  p[4][0]=v[i4].e();
  p[4][1]=v[i4].px();
  p[4][2]=v[i4].py();
  p[4][3]=v[i4].pz();
  p[5][0]=v[5].e();
  p[5][1]=v[5].px();
  p[5][2]=v[5].py();
  p[5][3]=v[5].pz();
}
  
void Vec4toMGQCD(const vector<Vec4>& v, int i0, int i1, int i2, int i3, 
                 double p[4][4]) {
  p[0][0]=v[i0].e();
  p[0][1]=v[i0].px();
  p[0][2]=v[i0].py();
  p[0][3]=v[i0].pz();
  p[1][0]=v[i1].e();
  p[1][1]=v[i1].px();
  p[1][2]=v[i1].py();
  p[1][3]=v[i1].pz();
  p[2][0]=v[i2].e();
  p[2][1]=v[i2].px();
  p[2][2]=v[i2].py();
  p[2][3]=v[i2].pz();
  p[3][0]=v[i3].e();
  p[3][1]=v[i3].px();
  p[3][2]=v[i3].py();
  p[3][3]=v[i3].pz();
}
  
void Vec4toMGQCD(const vector<Vec4>& v, int i0, int i1, int i2, int i3,  int i4, 
                 double p[5][4]) {
  p[0][0]=v[i0].e();
  p[0][1]=v[i0].px();
  p[0][2]=v[i0].py();
  p[0][3]=v[i0].pz();
  p[1][0]=v[i1].e();
  p[1][1]=v[i1].px();
  p[1][2]=v[i1].py();
  p[1][3]=v[i1].pz();
  p[2][0]=v[i2].e();
  p[2][1]=v[i2].px();
  p[2][2]=v[i2].py();
  p[2][3]=v[i2].pz();
  p[3][0]=v[i3].e();
  p[3][1]=v[i3].px();
  p[3][2]=v[i3].py();
  p[3][3]=v[i3].pz();
  p[4][0]=v[i4].e();
  p[4][1]=v[i4].px();
  p[4][2]=v[i4].py();
  p[4][3]=v[i4].pz();
}
  
void Vec4toMGQCD(const vector<Vec4>& v, int i0, int i1, int i2, int i3,  int i4,  int i5, 
                 double p[6][4]) {
  p[0][0]=v[i0].e();
  p[0][1]=v[i0].px();
  p[0][2]=v[i0].py();
  p[0][3]=v[i0].pz();
  p[1][0]=v[i1].e();
  p[1][1]=v[i1].px();
  p[1][2]=v[i1].py();
  p[1][3]=v[i1].pz();
  p[2][0]=v[i2].e();
  p[2][1]=v[i2].px();
  p[2][2]=v[i2].py();
  p[2][3]=v[i2].pz();
  p[3][0]=v[i3].e();
  p[3][1]=v[i3].px();
  p[3][2]=v[i3].py();
  p[3][3]=v[i3].pz();
  p[4][0]=v[i4].e();
  p[4][1]=v[i4].px();
  p[4][2]=v[i4].py();
  p[4][3]=v[i4].pz();
  p[5][0]=v[i5].e();
  p[5][1]=v[i5].px();
  p[5][2]=v[i5].py();
  p[5][3]=v[i5].pz();
}


//***********************************************************************

//*********

// Find all daughters of iAnc in event. Return a list with first color 
// daughter first, remaining partons unordered.

int findDaughters(Event& event,int iAnc,vector<int>& iDau) {

  // Sanity check
  if (iAnc <= 0 || iAnc >= event.size() ) return 0;
  iDau.resize(0);

  // Trace first daughter to the end.
  int iDa1=event[iAnc].daughter1();
  while (event[iDa1].status() < 0 && iDa1 > 0) iDa1=event[iDa1].daughter1();

  // Sanity check
  if (iDa1 == 0) return 0;

  // Store
  iDau.push_back(iDa1);

  // Add remaining partons in no particular order first
  for (int i=1;i<=event.size();i++) {
    if (i != iDa1 && event[i].isFinal() && isAncestor(event, i,iAnc)) {      
      iDau.push_back(i);
    }
  }

  // Return number of final-state daughters
  return iDau.size();

}

//*********

// Order a list of partons in color / branching sequence
// following the first parton in the list

int colorSort(Event& event,vector<int>& iList) {

  // Shorthand for final-state parton multiplicities
  int nPartons=iList.size();
  if (nPartons == 0) return 1;

  // Line up partons to follow first parton in color/branching order
  int iA=0;
  
  while (iA < nPartons-1) {

    int cType = event[iList[iA]].colType();

    // if iA is quark or gluon, find anti-color partner
    if (cType == 1 || cType == 2) {
      for (int iB=iA+1;iB<nPartons;iB++) {
        if (event[iList[iB]].acol() == event[iList[iA]].col()) {
          int iTmp(iList[iA+1]);
          iList[iA+1]=iList[iB];
          iList[iB]=iTmp;
          iA++;
        }
      }
    } 

    // if iA is antiquark, next is any quark of same flavor 
    else {
      int idA = event[iList[iA]].id();
      for (int iB=iA+1;iB<nPartons;iB++) {
        if (event[iList[iB]].id() == - idA) {
          int iTmp(iList[iA+1]);
          iList[iA+1]=iList[iB];
          iList[iB]=iTmp;
          iA++;
        }
      }
      
    }

  }

  // Normal return
  return 0;

}

//*********

// Check whether a given particle is an arbitrarily-steps-removed
// mother to another.
// (adapted for antenna showers, from Event::isAncestor from Pythia 8)
// (should probably add loop that stores additional mothers, so that we can 
// check whether all mothers are the same?)

bool isAncestor(Event& event, int i, int iAncestor) {

  // Begin loop to trace upwards from the daughter.
  int iUp = i;
  for ( ; ; ) {

    // If positive match then done.
    if (iUp == iAncestor) return true;

    // If out of range then failed to find match.
    if (iUp <= 0 || iUp > event.size()) return false;

    // Else go to mother1
    iUp = event[iUp].mother1(); 
    continue;

  }

  // End of loop. Should never reach beyond here.
  return false;

}

bool compVecSizeHelper(vector<int> a, vector<int> b) {
  return (a.size() < b.size());
}

//**************************************************************************

// Math support auxiliaries

double Li2(const double x,const double kmax,const double xerr)
{
  if (x<0.0) return 0.5*Li2(x*x)-Li2(-x);
  if (x<=0.5)
    {
      double sum  = x;
      double term = x;
      for(int k=2; k<kmax; k++) 
        {
          double rk = (k-1.0)/k;
          term *= x;
          term *= rk*rk;
          sum += term;
          if (fabs(term/sum) < xerr) return sum;
        }
      cout<<"Maximum number of iterations exceeded in Li2"<<endl;
      return sum;
    }
  if (x<1.0) 
    return M_PI*M_PI/6.0-Li2(1.0-x)-log(x)*log(1.0-x);
  if (x==1.0) 
    return M_PI*M_PI/6.0;
  if (x<=1.01)
    {
      const double eps = x - 1.0;
      const double lne = log(eps);
      const double c0 = M_PI*M_PI/6.0;
      const double c1 =   1.0 - lne;
      const double c2 = -(1.0 - 2.0*lne)/4.0;
      const double c3 =  (1.0 - 3.0*lne)/9.0;
      const double c4 = -(1.0 - 4.0*lne)/16.0;
      const double c5 =  (1.0 - 5.0*lne)/25.0;
      const double c6 = -(1.0 - 6.0*lne)/36.0;
      const double c7 =  (1.0 - 7.0*lne)/49.0;
      const double c8 = -(1.0 - 8.0*lne)/64.0;
      return c0+eps*(c1+eps*(c2+eps*(c3+eps*(c4+eps*(c5+eps*(c6+eps*(c7+eps*c8)))))));
    }
  double log_x=log(x);
  if (x<=2.0) 
    return M_PI*M_PI/6.0+Li2(1.0 - 1.0/x)-log_x*(log(1.0-1.0/x)+0.5*log_x);
  return M_PI*M_PI/3.0-Li2(1.0/x)-0.5*log_x*log_x;
}

int factorial(const int n) {
  int fac=1;
  for (int i=2;i<=n;i++) fac *= i;
  return fac;
}

int binomial(const int n, const int m) {
  if (m < 0 || m > n) return 0;
  else if (m == n || m == 0) return 1;
  else if (m == 1 || m == n-1) return n;
  else return factorial(n)/factorial(m)/factorial(n-m);
}

// MR 21.12.2010: Gram determinant
// the inviariants used in the argument are 2*pi*pj
double gDet(
  double s12tilde, double s23tilde, double s13tilde, 
  double m1, double m2, double m3
  ) {
  return        s12tilde*s23tilde*s13tilde/4
                    -pow2(s12tilde)*pow2(m3)/4
                    -pow2(s13tilde)*pow2(m2)/4
                    -pow2(s23tilde)*pow2(m1)/4
                    +pow2(m1)*pow2(m2)*pow2(m3);
}
// version with y13 omitted for normalised coordinates
double gDet(
  double y12tilde, double y23tilde, 
  double mu1, double mu2,double mu3
  ) {
  double y13tilde = 
    1-y12tilde-y23tilde-pow2(mu1)-pow2(mu2)-pow2(mu3);
  return gDet(y12tilde, y23tilde, y13tilde, mu1, mu2, mu3);
}
double gDet(Vec4 p1, Vec4 p2, Vec4 p3) {
  return gDet(
    2*p1*p2, 2*p2*p3, 2*p1*p3, p1.mCalc(), p2.mCalc(), p3.mCalc() );
}


double y23tildeMin(double y12tilde, double mu1, double mu2, 
                   double mu3) {
  return        -(-2*pow2(mu2) + 2*pow2(mu1)*
                  pow2(mu2) + 2*pow4(mu2) + 
                  2*pow2(mu2)*pow2(mu3) - y12tilde + 
                  pow2(mu1)*y12tilde + 3*pow2(mu2)*y12tilde + 
                  pow2(mu3)*y12tilde + pow2(y12tilde) + 
                  sqrt(1 - pow2(mu1) - pow2(mu2) - 
                       2*mu3 + pow2(mu3) - y12tilde)*
                  sqrt(1 - pow2(mu1) - pow2(mu2) + 
                       2*mu3 + pow2(mu3) - y12tilde)*
                  sqrt(-4*pow2(mu1)*pow2(mu2) + 
                       pow2(y12tilde)))/
    (2.*(pow2(mu1) + pow2(mu2) + y12tilde));
}

double y23tildeMax(double y12tilde, double mu1, double mu2, 
                   double mu3) {
  return        -(-2*pow2(mu2) + 2*pow2(mu1)*
                  pow2(mu2) + 2*pow4(mu2) + 
                  2*pow2(mu2)*pow2(mu3) - y12tilde + 
                  pow2(mu1)*y12tilde + 3*pow2(mu2)*y12tilde + 
                  pow2(mu3)*y12tilde + pow2(y12tilde) - 
                  sqrt(1 - pow2(mu1) - pow2(mu2) - 
                       2*mu3 + pow2(mu3) - y12tilde)*
                  sqrt(1 - pow2(mu1) - pow2(mu2) + 
                       2*mu3 + pow2(mu3) - y12tilde)*
                  sqrt(-4*pow2(mu1)*pow2(mu2) + 
                       pow2(y12tilde)))/
    (2.*(pow2(mu1) + pow2(mu2) + y12tilde));
}


//*********

// The log of the Gamma function for positive x
// (Adapted from Numerical Recipes in Fortran, 1994)

double lnGamma(const double xx) 
{
  static double cof[6]={76.18009172947146,
                        -86.50532032941677,
                        24.01409824083091,
                        -1.231739572450155,
                        0.1208650973866179e-2,
                        -0.5395239384953e-5};
  static double stp = 2.5066282746310005;
  if (xx <= 0) {
    printErr("lnGamma","cannot be called with x <= 0");
    return 0.0;
  }
  double y=xx;
  double tmp=xx+5.5;
  tmp = (xx+0.5)*log(tmp)-tmp;
  double ser=1.000000000190015;
  for (int j=0; j<6; j++) {
    y   += 1.0;
    ser += cof[j]/y;
  }
  return tmp + log(stp*ser/xx);
}

double Gamma(const double xx) 
{ 
  if (xx >= 0) {
    return exp(lnGamma(xx)); 
  } else {
    printErr("Gamma","Gamma(x<0) not implemented");
    return 0.0;
  }
}
  
//*********

// Solve the integral of fun1 over the interval [xlo,xhi]
// by adaptive gaussian quadrature (flat preweighting).
// Adapted from the CERNLIB DGAUSS routine by K.S. Kolbig. Error
// estimated by comparing 8-point and 16-point results.

double integrateGauss(double (*fun)(double), double xlo, double xhi, double tol)  
{
  //8-point unweighted
  static double x8[4]={0.96028985649753623,
                       0.79666647741362674,
                       0.52553240991632899, 
                       0.18343464249564980};
  static double w8[4]={0.10122853629037626,
                       0.22238103445337447,
                       0.31370664587788729,
                       0.36268378337836198};
  //16-point unweighted
  static double x16[8]={0.98940093499164993, 
                        0.94457502307323258, 
                        0.86563120238783174, 
                        0.75540440835500303, 
                        0.61787624440264375, 
                        0.45801677765722739,
                        0.28160355077925891, 
                        0.09501250983763744};
  static double w16[8]={0.027152459411754095,
                        0.062253523938647893,
                        0.095158511682492785,
                        0.12462897125553387,
                        0.14959598881657673,
                        0.16915651939500254,
                        0.18260341504492359,
                        0.18945061045506850};
  //boundary checks
  if (xlo == xhi) return 0.0;
  if ( xlo > xhi ) {
    cerr<<" (integrateGauss:) -> xhi < xlo"<<endl;
    return 0.0;
  }
  //initialize
  double sum = 0.0;
  double c = 0.001/abs(xhi-xlo);
  double zlo = xlo;
  double zhi = xhi;
    
  bool nextbin = true;
  
  while ( nextbin ) {
    
    double zmi = 0.5*(zhi+zlo); // midpoint
    double zmr = 0.5*(zhi-zlo); // midpoint, relative to zlo
    
    //calculate 8-point and 16-point quadratures
    double s8=0.0;
    for (int i=0;i<4;i++) {
      double dz = zmr * x8[i];
      s8 += w8[i]*(fun(zmi+dz) + fun(zmi-dz));
    }
    s8 *= zmr;
    double s16=0.0;
    for (int i=0;i<8;i++) {
      double dz = zmr * x16[i];
      s16 += w16[i]*(fun(zmi+dz) + fun(zmi-dz)); 
    }
    s16 *= zmr;
    if (abs(s16-s8) < tol*(1+abs(s16))) { 
      //precision in this bin OK, add to cumulative and go to next
      nextbin=true;
      sum += s16;
      //next bin: LO = end of current, HI = end of integration region.
      zlo=zhi;
      zhi=xhi;
      if ( zlo == zhi ) nextbin = false;
    } else {
      //precision in this bin not OK, subdivide.
      if (1.0 + c*abs(zmr) == 1.0) {
        cerr << " (integrateGauss:) too high accuracy required"<<endl;
        sum = 0.0 ;
        break;
      }
      zhi=zmi;
      nextbin=true;
    }
  }
  return sum;
}

  TXiFunctor::TXiFunctor(vector<double> mIn, vector<double> energiesIn) {
    // if the size of the two vectors is not equal, initialise empty
    if (mIn.size() != energiesIn.size()) {
      m = vector<double>(0);
      energies = vector<double>(0);
    }
    else {
      m = mIn;
      energies = energiesIn;
    }
  }
  
  double TXiFunctor::operator() (double xi) {
    double retval = 0.;
    for (vector<double>::size_type i=0; i<m.size(); i++) {
      retval += sqrt( pow2(m[i]) + pow2(xi)*pow2(energies[i]) );
    }
    return retval;
  }
  

//*********

// zbrent 
// solve fun(x)-r = 0 for x 
// tol is (probably?) an absolute tolerance parameter ...

double zbrent(double (*fun)(double), double r, double x1, double x2, double tol)
{
  int iter;
  double a=x1,b=x2,c=x2,d=x2-x1,e=x2-x1,min1,min2;
  double fa=fun(a)-r,fb=fun(b)-r,fc,p,q,r1,s,tol1,xm;
  double REALTINY=min(TINY,1e-12);
  tol = max(tol,REALTINY);

  // Check if there is a single zero in range
  if ( fa*fb > 0 ) { return 0.0; }

  // Start search
  fc=fb;
  for (iter=1;iter<max(1000,int(1.0/sqrt(tol)));iter++) {
    if ((fb > 0.0 && fc > 0.0) || (fb < 0.0 && fc < 0.0)) {
      c=a; 
      fc=fa;
      e=d=b-a;
    }
    if (abs(fc) < abs(fb)) {
      a=b;
      b=c;
      c=a;
      fa=fb;
      fb=fc;
      fc=fa;
    }
    tol1=2.0*REALTINY*fabs(b)+0.5*tol; 
    xm=0.5*(c-b);
    if (fabs(xm) <= tol1 || fb == 0.0) return b;
    if (fabs(e) >= tol1 && fabs(fa) > fabs(fb)) {
      s=fb/fa;
      if (a == c) {
        p=2.0*xm*s;
        q=1.0-s;
      } else {
        q=fa/fc;
        r1=fb/fc;
        p=s*(2.0*xm*q*(q-r1)-(b-a)*(r1-1.0));
        q=(q-1.0)*(r1-1.0)*(s-1.0);
      }
      if (p > 0.0) q = -q;
      p=fabs(p);
      min1=3.0*xm*q-fabs(tol1*q);
      min2=fabs(e*q);
      if (2.0*p < (min1 < min2 ? min1 : min2)) {
        e=d;
        d=p/q;
      } else {
        d=xm;
        e=d;
      }
    } else {
      d=xm;
      e=d;
    }
    a=b; 
    fa=fb;
    if (fabs(d) > tol1)
      b += d;
    else
      b += (xm > 1) ? tol1 : - tol1;
    fb=fun(b)-r;
  }
  cerr<<"(brent:) -> Maximum number of iterations exceeded"<<endl;
  return 0.0;
}

// MR 6.2.2011: need a version of zbrent using function pointers
// exact copy of the existing zbrent except for the first argument
// should replace the version above after testing
// zbrent
 
// solve fun(x)-r = 0 for x 
// tol is (probably?) an absolute tolerance parameter ...

double zbrentAlt(TFunctor& fun, double r, double x1, double x2, double tol)
{
  int iter;
  double a=x1,b=x2,c=x2,d=x2-x1,e=x2-x1,min1,min2;
  double fa=fun(a)-r,fb=fun(b)-r,fc,p,q,r1,s,tol1,xm;
  double REALTINY=min(TINY,1e-12);
  tol = max(tol,REALTINY);

  // Check if there is a single zero in range
  if ( fa*fb > 0 ) { return 0.0; }

  // Start search
  fc=fb;
  for (iter=1;iter<max(1000,int(1.0/sqrt(tol)));iter++) {
    if ((fb > 0.0 && fc > 0.0) || (fb < 0.0 && fc < 0.0)) {
      c=a; 
      fc=fa;
      e=d=b-a;
    }
    if (abs(fc) < abs(fb)) {
      a=b;
      b=c;
      c=a;
      fa=fb;
      fb=fc;
      fc=fa;
    }
    tol1=2.0*REALTINY*fabs(b)+0.5*tol; 
    xm=0.5*(c-b);
    if (fabs(xm) <= tol1 || fb == 0.0) return b;
    if (fabs(e) >= tol1 && fabs(fa) > fabs(fb)) {
      s=fb/fa;
      if (a == c) {
        p=2.0*xm*s;
        q=1.0-s;
      } else {
        q=fa/fc;
        r1=fb/fc;
        p=s*(2.0*xm*q*(q-r1)-(b-a)*(r1-1.0));
        q=(q-1.0)*(r1-1.0)*(s-1.0);
      }
      if (p > 0.0) q = -q;
      p=fabs(p);
      min1=3.0*xm*q-fabs(tol1*q);
      min2=fabs(e*q);
      if (2.0*p < (min1 < min2 ? min1 : min2)) {
        e=d;
        d=p/q;
      } else {
        d=xm;
        e=d;
      }
    } else {
      d=xm;
      e=d;
    }
    a=b; 
    fa=fb;
    if (fabs(d) > tol1)
      b += d;
    else
      b += (xm > 1) ? tol1 : - tol1;
    fb=fun(b)-r;
  }
  cerr<<"(brent:) -> Maximum number of iterations exceeded"<<endl;
  return 0.0;
}

vector<int> nameVec2idVec(vector<string> nameVec) {
  vector<int> idVec;
  for (int ic=0;ic<int(nameVec.size());ic++) {
    if (nameVec[ic]=="q") idVec.push_back(1);
    else if (nameVec[ic]=="qbar") idVec.push_back(-1);
    else if (nameVec[ic]=="g") idVec.push_back(21);
    else if (nameVec[ic]=="gamma") idVec.push_back(22);
    else idVec.push_back(0);
  }
  return idVec;
}

}
